//===----------------------------------------------------------------------===//
//This software was produced with support from the Defense Advanced
//Research Projects Agency (DARPA) through AFRL Contract FA8650-09-C-1915. 
//Nothing in this work should be construed as reflecting the official policy or
//position of the Defense Department, the United States government, or
//Rice University.
//
//THIS SOFTWARE HAS BEEN APPROVED FOR PUBLIC RELEASE, UNLIMITED DISTRIBUTION.
//
//===----------------------------------------------------------------------===//


#include "AvailDefinitions.h"

#include "llvm/ADT/PostOrderIterator.h"
#include "llvm/ADT/DenseSet.h"

#include "llvm/CodeGen/MachineBasicBlock.h"
#include "llvm/CodeGen/MachineFunction.h"
#include "llvm/CodeGen/MachineInstr.h"
#include "llvm/Function.h" // Debug
#include "llvm/Support/Debug.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/Target/TargetRegisterInfo.h"

static bool DoesInstructionDefineAndUsesSameReg(MachineInstr *Inst) {
  unsigned NumOperands = Inst->getNumOperands();

  DenseSet<unsigned> Uses;
  for (unsigned I = 0; I != NumOperands; ++I) {
    MachineOperand &Oper = Inst->getOperand(I);
    // Ignore non uses.
    if (!Oper.isReg() || !Oper.getReg() || !Oper.isUse())
      continue;
    Uses.insert(Oper.getReg());
  }
  for (unsigned I = 0; I != NumOperands; ++I) {
    MachineOperand &Oper = Inst->getOperand(I);
    // Ignore non definitions.
    if (!Oper.isReg() || !Oper.getReg() || !Oper.isDef())
      continue;
    // Instruction has a register that is a use and def.
    if (Uses.count(Oper.getReg())) return true;
  }
  return false;
}

/// build - Builds AvailDefIns and AvailDefsOuts for the function.
void AvailableDefinitions::build() {
  TRI = MF->getTarget().getRegisterInfo();

  assert(TRI != NULL && "Register info must not be null");
  // Find which instructions could be killed by defining a register.
  buildPotentiallyKilledMap();
  // Build KilledDefs and DownwardExposedDefs sets.
  buildKillAndDownwardExposedSets();
  buildInAndOutSets();
}

std::set<MachineInstr *>
AvailableDefinitions::translateToSet(BitVector &Avail) {
  std::set<MachineInstr *> Set;
  for (unsigned Idx = 0, E = Avail.size(); Idx != E; ++Idx) {
    if (Avail.test(Idx))
      Set.insert(IndexToInstrMap[Idx]);
  }
  return Set;
}



std::set<MachineInstr *>
AvailableDefinitions::availableDefsAt(MachineInstr *AtInst) {
  typedef std::set<MachineInstr *> InstrSet;
  MachineBasicBlock *BB = AtInst->getParent();
  BitVector &AvailIn = AvailDefIns[BB];
  InstrSet CurrAvailInsts = translateToSet(AvailIn);
  for (MachineBasicBlock::iterator I = BB->begin(), E = BB->end(); I != E;
       ++I) {
    MachineInstr *Inst = &*I;
    if (Inst == AtInst) break;

    for (unsigned DefIdx = 0, DefEnd = Inst->getNumOperands();
         DefIdx != DefEnd; ++DefIdx) {
      MachineOperand &Oper = Inst->getOperand(DefIdx);
      if (!Oper.isReg() || !Oper.isDef() || Oper.getReg() == 0) continue;
      // Defs kill previous instructions.
      for (InstrSet::iterator II = CurrAvailInsts.begin();
           II != CurrAvailInsts.end();) {
        if ((*II)->readsRegister(Oper.getReg(), TRI) ||
            (*II)->modifiesRegister(Oper.getReg(), TRI)) {
          InstrSet::iterator Del = II;
          ++II;
          CurrAvailInsts.erase(Del);
        } else
          ++II;
      }
    }
    if (!DoesInstructionDefineAndUsesSameReg(Inst))
      CurrAvailInsts.insert(Inst);
  }
  return CurrAvailInsts;
}

void AvailableDefinitions::buildInAndOutSets() {
  // Initialize AvailDefsOuts to TOP (every definition is available).
  for (MachineFunction::iterator BB = MF->begin(), End = MF->end(); BB != End;
       ++BB)
    AvailDefOuts[BB].set();

  bool HasChangeOccured;
  ReversePostOrderTraversal<MachineFunction *> RPO(MF);
  do {
    HasChangeOccured = false;
    // Propagate information in reverse post order over the control flow graph.
    for (ReversePostOrderTraversal<MachineFunction *>::rpo_iterator RI =
         RPO.begin(), RE = RPO.end(); RI != RE; ++RI) {
      MachineBasicBlock *BB = *RI;
      // Build new AVAILDEFOUT[BB].
      BitVector Tmp = buildNewOut(BB);
      if (Tmp != AvailDefOuts[BB]) {
        AvailDefOuts[BB] = Tmp;
        HasChangeOccured = true;
      }
    }
  } while (HasChangeOccured);
}

BitVector AvailableDefinitions::buildNewOut(MachineBasicBlock *BB) {
  // AVAILDEFIN = intersection AVAILDEFOUT[p] : for all p in preds of BB
  if (BB->pred_empty())
    AvailDefIns[BB].reset();
  else {
    for (MachineBasicBlock::pred_iterator P = BB->pred_begin(), First = P,
         PE = BB->pred_end(); P != PE; ++P) {
      MachineBasicBlock *Pred = *P;
      if (P == First)
        AvailDefIns[BB] = AvailDefOuts[Pred];
      else
        AvailDefIns[BB] &= AvailDefOuts[Pred]; // Intersect.
    }
  }
  // AVAILDEFOUT = (AVAILDEFIN - KILLS) + DEDEFS
  BitVector NewDefOut(AvailDefIns[BB]);
  NewDefOut &= ~KilledDefs[BB];
  NewDefOut |= DownwardExposedDefs[BB];
  return NewDefOut;
}

// buildKillAndDownwardExposedSets - Builds the two sets by walking every basic
// block from end to beginning. First, the current definition d is added to the
// DownwardExposedDefs set if d is not already in the local  KilledDefs set.
// After that, the definitions killed by d (other instructions that contain the
// defs of d as - use and def - operands) are added to the KilledDefs set.
void AvailableDefinitions::buildKillAndDownwardExposedSets() {
  // Iterate over all basic blocks.
  for (MachineFunction::iterator MFI = MF->begin(), MFE = MF->end(); MFI != MFE;
       ++MFI) {
    MachineBasicBlock *BB = MFI;
    BitVector &LocalKilled = KilledDefs[BB];
    BitVector &LocalDEDefs = DownwardExposedDefs[BB];

    // Walk basic block from end to beginning.
    for (MachineBasicBlock::reverse_iterator BI = BB->rbegin(), BE = BB->rend();
         BI != BE; ++BI) {
      MachineInstr *Instr = &*BI;

      // Check that current instruction is not killed by an instruction we have
      // seen before (is in LocalKilled).
      // Note: An instruction can kill it self.
      //  x <- x + y
      // Because the instruction defines a register that it itself uses we must
      // assume that it is killed. The reason behind this is that  when we want
      // to copy the instruction further down the execution path x (the operand
      // is) will no longer have the same value.
      // x <- x + y
      // ...
      // ...
      // z <- z * x
      // If we copy the instruction (without deleting the original) to before z,
      // the result will be wrong.
      // x <- x + y
      // ...
      // ...
      // x <- x + y
      // z <- z * x
      unsigned SetIndex = InstrToIndexMap[Instr];
      if (!LocalKilled[SetIndex] && !DoesInstructionDefineAndUsesSameReg(Instr))
        LocalDEDefs.set(SetIndex);

      // Add the instructions that are killed by the def operands to the
      // LocalKilled set.
      for (unsigned I = 0, E = Instr->getNumOperands(); I != E; ++I) {
        MachineOperand &Op = Instr->getOperand(I);
        if (Op.isReg() && Op.getReg() != 0 && Op.isDef())
          LocalKilled |= PotKilled[Op.getReg()];
      }
    }
  }
}

void AvailableDefinitions::addInstrToPotKillOfReg(MachineInstr *Instr,
                                                  unsigned Reg) {
  BitVector &Set = PotKilled[Reg];
  if (Set.size() == 0)
    Set.resize(NumInstructions);
  Set.set(InstrToIndexMap[Instr]);
}

void AvailableDefinitions::addOperandsToKilledMap(MachineInstr *Instr) {
  // Add uses and defs.
  for (unsigned I = 0, End = Instr->getNumOperands(); I != End; ++I) {
    MachineOperand &Op = Instr->getOperand(I);
    // A zero register is not a register (see TargetRegisterInfo).
    if (Op.isReg() && Op.getReg() != 0) {
      unsigned Reg = Op.getReg();
      if (Op.isDef()  || Op.isUse()) {
        // Add all aliased registers for physical registers.
        if (TargetRegisterInfo::isPhysicalRegister(Reg))
          for (const unsigned *R = TRI->getOverlaps(Reg); R && *R; ++R)
            addInstrToPotKillOfReg(Instr, *R);
        else
          addInstrToPotKillOfReg(Instr, Reg); // VReg
      }
    }
  }
}

void AvailableDefinitions::reserveSpaceInSets() {
  for (MachineFunction::iterator BB = MF->begin(), End = MF->end(); BB != End;
       ++BB) {
    KilledDefs[BB].resize(NumInstructions);
    DownwardExposedDefs[BB].resize(NumInstructions);
    AvailDefIns[BB].resize(NumInstructions);
    AvailDefOuts[BB].resize(NumInstructions);
  }
}

/// buildPotentiallyKilledMap - Fills the PotentiallyKilledMap by adding an
/// entry to the InstrSet for each operand of each MachineInstr in the
/// function.
void AvailableDefinitions::buildPotentiallyKilledMap() {
  NumInstructions = 0;
  unsigned NumBBs = 0;
  // Iterate over basic blocks to count instructions.
  for (MachineFunction::iterator MFI = MF->begin(), MFE = MF->end(); MFI != MFE;
       ++MFI) {
    MachineBasicBlock *BB = MFI;
    ++NumBBs;
    // Iterate over instructions.
    for (MachineBasicBlock::iterator BI = BB->begin(), BE = BB->end(); BI != BE;
         ++BI) {
      InstrToIndexMap[BI] = NumInstructions;
      IndexToInstrMap[NumInstructions] = BI;
      ++NumInstructions;
    }
  }

  // Reserve space in bit vector sets.
  reserveSpaceInSets();

  // Iterate over basic blocks.
  for (MachineFunction::iterator MFI = MF->begin(), MFE = MF->end(); MFI != MFE;
       ++MFI) {
    MachineBasicBlock *BB = MFI;
    // Iterate over instructions.
    for (MachineBasicBlock::iterator BI = BB->begin(), BE = BB->end(); BI != BE;
         ++BI) {
      // Map all operands of the current instruction (defs and uses) to a
      // potential kill of the current instruction.
      addOperandsToKilledMap(BI);
    }
  }
}

#ifndef NDEBUG
/// dumpMachineInstrDefs - Dumps all the def operands of the machine
/// instruction.
void dumpMachineInstrDefs(raw_ostream &OS, MachineInstr *Instr) {
  for (unsigned I = 0, End = Instr->getNumOperands(); I != End; ++I) {
    MachineOperand &Oper = Instr->getOperand(I);
    if (Oper.isReg() && Oper.isDef()) {
      Oper.print(OS);
      OS << " ";
    }
  }
  if (Instr->getNumOperands() == 0)
    OS << Instr->getDesc().getName() << " ";
}

/// dumpLocalBBInfo - Dumps KilledDefs and DEDefs for every basic block.
void AvailableDefinitions::dumpLocalBBInfo(raw_ostream &OS) {
  MF->dump();
  for (MachineFunction::iterator MFI = MF->begin(), MFE = MF->end(); MFI != MFE;
       ++MFI) {
    MachineBasicBlock *BB = MFI;
    BitVector &LocalKilled = KilledDefs[BB];
    BitVector &LocalDEDef = DownwardExposedDefs[BB];
    BitVector &AvailIn = AvailDefIns[BB];
    BitVector &AvailOut = AvailDefOuts[BB];
    OS << "Local info for MBB " << BB->getName() << " " << BB->getNumber()
      << "\n";
    OS << "KilledDefs:\n";
    for (unsigned I = 0, E = NumInstructions; I != E; ++I)
      if (LocalKilled[I])
        dumpMachineInstrDefs(OS, IndexToInstrMap[I]);

    OS << "\n";
    OS << "DEDefs:\n";
    for (unsigned I = 0, E = NumInstructions; I != E; ++I)
      if (LocalDEDef[I])
        dumpMachineInstrDefs(OS, IndexToInstrMap[I]);

    OS << "\n";
    OS << "AvailIns:\n";
    for (unsigned I = 0, E = NumInstructions; I != E; ++I)
      if (AvailIn[I])
        dumpMachineInstrDefs(OS, IndexToInstrMap[I]);

    OS << "\n";
    OS << "AvailOuts:\n";
    for (unsigned I = 0, E = NumInstructions; I != E; ++I)
      if (AvailOut[I])
        dumpMachineInstrDefs(OS, IndexToInstrMap[I]);

    OS << "\n";

    for (MachineBasicBlock::iterator BI = BB->begin(), BE = BB->end(); BI != BE;
         ++BI) {
      std::set<MachineInstr *> AvailDefsAt = availableDefsAt(BI);
    }
  }
}

void AvailableDefinitions::dump() {
  dumpLocalBBInfo(dbgs());
}

void AvailableDefinitions::dumpAvailDefsAt(unsigned BBNum, unsigned InstrNum) {
  MachineBasicBlock *BB = MF->getBlockNumbered(BBNum);
  for (MachineBasicBlock::iterator II = BB->begin(), E = BB->end(); II != E;
       ++II) {
    if (InstrNum-- == 0) {
      std::set<MachineInstr *> AvailDefs = availableDefsAt(&*II);
      errs() << "Avail Defs:\n";
      for (std::set<MachineInstr *>::iterator AI = AvailDefs.begin(),
           AE = AvailDefs.end(); AI != AE; ++AI) {
        errs() << "  " << **AI;
      }
      return;
    }
  }
  errs() << "Could not find instruction\n";
}

AvailDefsIterator AvailableDefinitions::begin(MachineBasicBlock *BB) {
  std::set<MachineInstr *> Tmp = translateToSet(AvailDefIns[BB]);
  return AvailDefsIterator(BB, Tmp, false);
}

AvailDefsIterator AvailableDefinitions::end(MachineBasicBlock *BB) {
  std::set<MachineInstr *> EmptySet;
  return AvailDefsIterator(BB, EmptySet, true);
}

AvailDefsIterator::AvailDefsIterator(MachineBasicBlock *BB, InstrSet &Avail,
                                     bool IsEnd) : Block(BB) {
  TRI = BB->getParent()->getTarget().getRegisterInfo();
  CurAvailInsts = Avail;
  BlockIter = IsEnd ? BB->end() : BB->begin();
}

AvailDefsIterator &AvailDefsIterator::operator++() {
  assert(BlockIter != Block->end());
  MachineInstr *CurInst = &*BlockIter;
  // Look at all defs.
  for (unsigned DefIdx = 0, DefEnd = CurInst->getNumOperands();
       DefIdx != DefEnd; ++DefIdx) {
    MachineOperand &Oper = CurInst->getOperand(DefIdx);
    if (!Oper.isReg() || !Oper.isDef() || Oper.getReg() == 0) continue;
    // Defs kill previous instructions.
    for (InstrSet::iterator II = CurAvailInsts.begin();
         II != CurAvailInsts.end();) {
      if ((*II)->readsRegister(Oper.getReg(), TRI) ||
          (*II)->modifiesRegister(Oper.getReg(), TRI)) {
        InstrSet::iterator Del = II;
        ++II;
        CurAvailInsts.erase(Del);
      } else
        ++II;
    }
  }
  if (!DoesInstructionDefineAndUsesSameReg(CurInst))
    CurAvailInsts.insert(CurInst);
  // Advance to next instruction.
  ++BlockIter;
  return *this;
}

/// post-increment
AvailDefsIterator AvailDefsIterator::operator++(int) {
  AvailDefsIterator tmp = *this;
  ++*this;
  return tmp;
}

bool  AvailDefsIterator::operator==(AvailDefsIterator &Other) const {
  return Other.BlockIter == BlockIter;
}

bool AvailDefsIterator::operator!=(AvailDefsIterator &Other) const {
  return Other.BlockIter != BlockIter;
}

#endif
