//===----------------------------------------------------------------------===//
//This software was produced with support from the Defense Advanced
//Research Projects Agency (DARPA) through AFRL Contract FA8650-09-C-1915. 
//Nothing in this work should be construed as reflecting the official policy or
//position of the Defense Department, the United States government, or
//Rice University.
//
//THIS SOFTWARE HAS BEEN APPROVED FOR PUBLIC RELEASE, UNLIMITED DISTRIBUTION.
//
//===----------------------------------------------------------------------===//


#ifndef __AVAILABLEDEFINITIONS_H__
#define __AVAILABLEDEFINITIONS_H__

#include <map>
#include <set>
#include "llvm/ADT/DenseMap.h"
#include "llvm/ADT/BitVector.h"
#include "llvm/CodeGen/MachineBasicBlock.h"

namespace llvm {
  class MachineBasicBlock;
  class MachineFunction;
  class MachineInstr;
  class raw_ostream;
  class TargetRegisterInfo;
}

using namespace llvm;

/// Iterator for avail defs within a basic block.
class AvailDefsIterator {
  typedef std::set<MachineInstr *> InstrSet;

  MachineBasicBlock *Block;
  InstrSet CurAvailInsts;
  MachineBasicBlock::iterator BlockIter;
  const TargetRegisterInfo *TRI;

  AvailDefsIterator(MachineBasicBlock *BB, InstrSet &Avail, bool IsEnd);
public:
  const InstrSet &operator*() { return CurAvailInsts; }

  MachineInstr *getCurrentInstruction() { return BlockIter; }

  AvailDefsIterator &operator++();
  /// post-increment
  AvailDefsIterator operator++(int);
  bool operator==(AvailDefsIterator &Other) const;
  bool operator!=(AvailDefsIterator &Other) const;

  friend class AvailableDefinitions;
};

/// This class computes available definitions.
/// Rematerialization uses availabled definitions to identify which instruction
/// can be rematerialized at there use.
///
/// A definition d: x <- op1 + op2 is available at a program point p if
///   * d reaches p along all paths to p AND
///   * there is no redefinition of x between any path between d and p
///   * there is no definition of op1 or op2 along any path between d and p
///   * d those not defined and use the same register (see note below).
/// If a definition d is available at a point p then we can put a insert a copy
/// of d just before p and it will define the same value as the original d.
// Note: An instruction can kill it self.
//  x <- x + y
// Because the instruction defines a register that it itself uses we must
// assume that it is killed. The reason behind this is that  when we want
// to copy the instruction further down the execution path x (the operand
// is) will no longer have the same value.
// x <- x + y
// ...
// ...
// z <- z * x
// If we copy the instruction (without deleting the original) to before z,
// the result will be wrong.
// x <- x + y
// ...
// ...
// x <- x + y
// z <- z * x
/// We could relax the last constraint for available definitions whose original
/// instruction we can delete: Available definitions that are available at all
/// their reaching uses.
///
/// We view a machine instruction as a definition. Machine instructions can
/// define more than one register. We handle this by conservatively assuming
/// that if one of this registers is redefined the "whole" machine instruction
/// is redefined.
/// AVAILDEF analysis:
///
/// In the following we assume every definition is given a unique identifier.
/// bb1:                  bb2:
///   x <-  w + z          x <- w + z
///
/// =>
///
/// bb1:                  bb2:
///   x1: x <-  w + z          x2: x <- w + z
///
/// The lattice consists of elements of the powerset of all definitions.
///
///
///                         ANY_DEF = {x1, x2, x3, ... xn}
///                 {x1, ... xn-1}     ...       {x2 ... xn2}
///               ...
/// LATTICE =      {x1, x2}
///                    {x1} {x2}  ...
///                         NO_DEF
///
/// AVAILDEFIN(B) = INTERSECTIONOFALL(AVAILDEFOUT(P)) of all pred P of B
///
/// AVAILDEFOUT(B) = DEDEF + (AVAILDEFIN(B) - KILLDEF)
///
/// DEDEF = set of all definitions of xi: x <- op1 .. opn such that  op1 to opn
/// are not redefined on the path from x to the exit of the basic block and the
/// definition of x is the last one in the basic block.
///
/// KILLDEF =  set of definitions killed by an definition of x in the basic
/// block. Either because y <- op1 ...  and x defines (equals) op1 or because
/// x is assigned a new value (x2: x <- ... followed by x3: x <-  ..., here x3
/// kills x2 because it assigns to x).
///
///
class AvailableDefinitions {
  typedef std::map<MachineBasicBlock *, BitVector> BasicBlockToSetMap;
  // Maps MachineOperand register numbers to MachineInstrs that could be killed
  // by defining the MachineOperand in a basic block.
  typedef std::map<unsigned, BitVector> PotentiallyKilledMap;

  MachineFunction *MF;
  const TargetRegisterInfo *TRI;

  DenseMap<MachineInstr *, unsigned> InstrToIndexMap;
  DenseMap<unsigned, MachineInstr *> IndexToInstrMap;
  unsigned NumInstructions;

  PotentiallyKilledMap PotKilled;

  BasicBlockToSetMap KilledDefs; // Defs that are killed in a basic block.
  BasicBlockToSetMap DownwardExposedDefs; // Downward exposed defs.
  BasicBlockToSetMap AvailDefIns; // Available defs at the start of a BB.
  BasicBlockToSetMap AvailDefOuts; // Available defs at the end of a BB.

  /// build - Builds AvailDefIns and AvailDefsOuts for the function.
  void build();

  /// buildPotentiallyKilledMap - Fills the PotentiallyKilledMap by adding an
  /// entry to the InstrSet for each operand of each MachineInstr in the
  /// function.
  void buildPotentiallyKilledMap();

  /// buildKillAndDownwardExposedSets - Builds basic block local sets.
  void buildKillAndDownwardExposedSets();

  /// buildInAndOutSets - Finds fixpoint of AvailDefIns and
  /// AvailDefsOuts.
  void buildInAndOutSets();

  /// buildNewOut - Compute new AVAILDEFOUT[BB], taking the current
  /// AVAILDEFIN[p] for all predecessors p of BB into account.
  BitVector buildNewOut(MachineBasicBlock *BB);

  /// addOperandsToKilledMap -  Add entries for all operands of the instruction
  /// to the PotentiallyKilledMap.
  void addOperandsToKilledMap(MachineInstr *Instr);

  /// reserveSpaceInSets - Reserves space in bit vector sets.
  void reserveSpaceInSets();

  /// addInstrToPotKillOfReg - Adds the instruction to the potential kill set of
  /// the register.
  void addInstrToPotKillOfReg(MachineInstr *Instr, unsigned Reg);

  std::set<MachineInstr *> translateToSet(BitVector &V);

public:
  AvailableDefinitions(MachineFunction *F) : MF(F) {
    build();
  }

  /// availableDefsAt - Returns the available definitions at (before) the
  /// machine instruction.
  std::set<MachineInstr *> availableDefsAt(MachineInstr *MI);

  AvailDefsIterator begin(MachineBasicBlock *BB);
  AvailDefsIterator end(MachineBasicBlock *BB);

#ifndef NDEBUG
  /// Debugging facilities
  void dumpLocalBBInfo(raw_ostream &OS);
  void dump();
  void dumpAvailDefsAt(unsigned BBNum, unsigned InstrNum);
#endif
};

#endif
