//===--BuildInterferenceGraph.cpp   - Build an interference graph   -------===//
// Builds an interference graph
// Abstract base class which must be specialized by implementing the
// pure virtual member function: isCorrectRegisterType(unsigned)
//===----------------------------------------------------------------------===//
//===----------------------------------------------------------------------===//
//This software was produced with support from the Defense Advanced
//Research Projects Agency (DARPA) through AFRL Contract FA8650-09-C-1915. 
//Nothing in this work should be construed as reflecting the official policy or
//position of the Defense Department, the United States government, or
//Rice University.
//
//THIS SOFTWARE HAS BEEN APPROVED FOR PUBLIC RELEASE, UNLIMITED DISTRIBUTION.
//
//===----------------------------------------------------------------------===//

///
/// This builder uses the concept of representative physical registers (RPR)
/// Representative physical registers are registers representing an alias set
/// Thus, for instance, the register alias set {AL, AH, AX, EAX} is represented by one 
/// representative physical register. RPRs are used as nodes in the interference graph
/// since the interference graph must be able to express that if EAX is 
/// used (colored), then so are registers in its alias set, namely: AX, AL, and AH. 
/// Similarly, if AH is used (colored), so are EAX and AX.
///
/// This approach loses some granularity. For instance, if AL is colored, then 
/// AH can be assigned a different color. However, it is unclear how to encode this
/// elegantly into an interference graph. (Smith, Ramsey, Holloway, PLDI 2004, 
/// "A Generalized Algorithm for Graph-Coloring Register Allocation" might be 
/// one potential approach)
/// 

#define DEBUG_TYPE "regalloc"

#include "BuildLiveSets.h"

#include "llvm/ADT/DenseSet.h"
#include "llvm/CodeGen/MachineFunction.h"
#include "llvm/CodeGen/MachineInstr.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/CodeGen/MachineRegisterInfo.h"
#include "llvm/Target/TargetRegisterInfo.h"

#include "llvm/Support/CommandLine.h"
#include "llvm/Support/Debug.h"

#include "EquivalenceClasses.h"

#include "Common.h"
#include "BuildInterferenceGraph.h"

#include <algorithm>
#include <map>

using namespace llvm;
extern cl::opt<bool> EnableProfilingAlloc;
extern cl::opt<bool> PassiveSplitting;

cl::opt<bool> UseRepReg("RepReg", cl::init(false));
cl::opt<bool> PrintIGBuild ("printIGbuild", cl::init(false));
cl::opt<bool> CopyRegsDoNotInterfere("copy-regs-do-not-interfere", cl::init(true));

cl::opt<bool> IsRematDisabled("disable-simple-remat", cl::init(false),
                              cl::desc("Disable simple register rematerialization (false)"));

cl::opt<bool> IsArtificiallyIncreaseRegPressureEnabled(
 "artificially-increase-reg-pressure", cl::init(false),
 cl::desc("Increase register pressure by adding artifical interference edges"));

extern cl::opt<bool> DebugColoring;  


// ctor
BuildInterferenceGraph::BuildInterferenceGraph(MachineFunction* F,
                                               const TargetMachine &TM,
                                               std::set<const TargetRegisterClass*> &RegClassSet,
                                               BuildLiveSets *InitialLiveBuilder /*= NULL*/)
{
  LiveBuilder = InitialLiveBuilder;
  if (LiveBuilder) {
    LiveOut = LiveBuilder->GetLiveOutSet();
    LiveIn = LiveBuilder->GetLiveInSet();
    RefSet = LiveBuilder->GetRefSet();
  }
  CurrentGraph = NULL;
  MF = F;
  RegInfo = TM.getRegisterInfo();
  FilterRegisterClasses = RegClassSet;
  RegisterRepresentativeSet = 0;
  RepresentativeNum2RegistersSet.clear();
  ExcludedRepresentativeRegs.clear();
  
  Initialize();
}

void BuildInterferenceGraph::Initialize() {
  buildRegisterRepresentativeSet();
  if ( DebugColoring ) {
    printRegisterRepresentativeSet();
  }
}

// dtor
BuildInterferenceGraph::~BuildInterferenceGraph() {
  LiveOut.clear();
  LiveIn.clear();
  RefSet.clear();
  
  if (CurrentGraph != NULL) {
    delete CurrentGraph; CurrentGraph = 0;
  }

  if (RegisterRepresentativeSet != NULL) {
    delete[] RegisterRepresentativeSet; RegisterRepresentativeSet = 0;
  }

  RepresentativeNum2RegistersSet.clear();
  ExcludedRepresentativeRegs.clear();
}

bool BuildInterferenceGraph::isCorrectRegisterType(unsigned reg) const {
  const TargetRegisterClass* regClass;
  
  // Special case for register 0 aka NoRegister
  if (reg == 0) {
    return false;
  }
  
  if ( TargetRegisterInfo::isPhysicalRegister(reg) ) {
    regClass = getRegClass_(RegInfo, reg);
    if (regClass == NULL) return false;
    assert(regClass && "Did not get register class for physical register");
  }
  else {
    regClass = MF->getRegInfo().getRegClass(reg);
  }
  
  return isCorrectRegisterClass(regClass);
}

bool BuildInterferenceGraph::isCorrectRegisterClass(const TargetRegisterClass *RC) const {
  return ( FilterRegisterClasses.find(RC) != FilterRegisterClasses.end() );
}


/// Builds a mapping from register number -> representative number for physical
/// registers.  The representative number for a register alias set is the register
/// number of one of the registers in the set.
void BuildInterferenceGraph::buildRegisterRepresentativeSet() {
  
  int classCtr = 1;
  EquivalenceClasses13<int> RegisterEqSets;
  
  // Resize the RegisterRepresentativeSet
  if ( RegisterRepresentativeSet ) {
    delete [] RegisterRepresentativeSet; RegisterRepresentativeSet = 0;
  }
  RegisterRepresentativeSet = new unsigned[RegInfo->getNumRegs()];

  for (unsigned I = 0, E = RegInfo->getNumRegs(); I != E; ++I)
   RegisterRepresentativeSet[I] = 0;

  for (TargetRegisterInfo::regclass_iterator classit = RegInfo->regclass_begin();
       classit != RegInfo->regclass_end();
       ++classit, ++classCtr) {
    
    const TargetRegisterClass *RegClass = *classit;
    
    for (TargetRegisterClass::iterator rit = RegClass->begin();
         rit != RegClass->end();
         ++rit) {
      
      unsigned Reg = *rit;
      
      //
      // Register filter:
      //
      // If it's not the correct register type, do nothing

      if (! isCorrectRegisterType(Reg)) {
        continue;
      }
      
      unsigned SuperReg = getSuperReg(Reg);
      
      RegisterEqSets.addElement(SuperReg);
      
      // Give all its aliases the same representative number
      for (const unsigned* aliasIter = RegInfo->getAliasSet(Reg);
           aliasIter && *aliasIter; ++aliasIter) {
        RegisterEqSets.addElement(*aliasIter);
        RegisterEqSets.unionSetsWith(*aliasIter, SuperReg);
      }
    }
  }
  
  // Convert into RegisterRepresentativeSet
  std::set<int>& equivalenceSetReps = RegisterEqSets.getLeaderSet();
  
  for (std::set<int>::iterator repIt = equivalenceSetReps.begin();
       repIt != equivalenceSetReps.end(); ++repIt) {
    
    const std::set<int>& eqElements = RegisterEqSets.getEqClass(*repIt);
    
    // Assign the same rep. number to all the elements of the eq. set
    for (std::set<int>::iterator elemIt = eqElements.begin();
         elemIt != eqElements.end(); ++elemIt) {
      
      RegisterRepresentativeSet[*elemIt] = *repIt;
      RepresentativeNum2RegistersSet[*repIt].insert(*elemIt);
    }
    
  }
  for (unsigned I = 0, E = RegInfo->getNumRegs(); I != E; ++I)
    if (isCorrectRegisterType(I)) {
      if (RegisterRepresentativeSet[I] == 0)
        errs() << "Error with " << PrintReg(I) << "\n";
      assert(RegisterRepresentativeSet[I] != 0 &&
             "Representative register should exist");
    }
}

unsigned getRegNumberForName(const char *Name, const TargetRegisterInfo *TRI) {
  for (unsigned I = 1, E = TRI->getNumRegs(); I != E; ++I)
    if (std::string(TRI->getName(I)) == Name)
        return I;
  assert(false && "Unknow register name");
  return 0;
}

void BuildInterferenceGraph::buildExcludedRepresentativeRegs(std::set<int> &Colors, 
                                                             std::map<int, unsigned> &Color2RepRegMap,
                                                             RegisterClassInfo &RCI) {

  std::set<unsigned> ColorRepRegs;
  std::set<unsigned> ReservedRepRegs;
  std::set<int>::iterator ColorsIter;  
  std::set<int>::iterator ColorsIterEnd;
  
  ColorsIter    = Colors.begin();
  ColorsIterEnd = Colors.end();
  
  ColorRepRegs.clear();
  for (; ColorsIter != ColorsIterEnd; ++ColorsIter) {
    int Color = *ColorsIter;
    unsigned RepReg = Color2RepRegMap[Color];
    ColorRepRegs.insert(RepReg);
    if ( RCI.isReserved(RepReg) ) {
      ReservedRepRegs.insert(RepReg);
    }
  }

  // Populate ExcludedRepresentativeRegs[], where ExcludedRepresentativeRegs[RegClass]
  // contains the corresponding representative registers that are in the RegClass most
  // super register class but not in the register class.

  for (TargetRegisterInfo::regclass_iterator classit = RegInfo->regclass_begin();
       classit != RegInfo->regclass_end(); ++classit) {
    const TargetRegisterClass *RC = *classit;

    ExcludedRepresentativeRegs[RC].clear();
    
    //
    // Register filter:
    //
    // If it's not the correct register type, do nothing
    if ( ! isCorrectRegisterClass(RC) ) {
      continue;
    }
    
    // RC->SubClasses      - classes which contain a proper subset of RC's registers
    // RC->SuperClasses    - classes in which all of RC registers are a proper subset
    // RC->SubRegClasses   - classes which contain only the converted RC sub-registers,
    //                       converted to a sub-register (GR16 -> GR8)
    // RC->SuperRegClasses - not implemented?
    
    // ExcludedRepresentativeRegs[RC] = Registers(Colors) - AllocatedRegs
    
    // std::map<const TargetRegisterClass*, std::set<unsigned> > ExcludedRepresentativeRegs;
    // std::set<int> RAChaitinBriggs::Colors;
    
    ExcludedRepresentativeRegs[RC] = ColorRepRegs;
    
    Ks[RC] = ColorRepRegs.size();

    ArrayRef<unsigned> RawAlocationOrder = RC->getRawAllocationOrder(*MF);
    ArrayRef<unsigned>::iterator AllocOrderIter    = RawAlocationOrder.begin();
    ArrayRef<unsigned>::iterator AllocOrderIterEnd = RawAlocationOrder.end();

    for (; AllocOrderIter != AllocOrderIterEnd; ++AllocOrderIter) {
      unsigned Reg = *AllocOrderIter;
      
      if ( ! isCorrectRegisterType(Reg) ) {
        continue;
      }
      
      unsigned RepReg = getRepresentativeNumber(Reg);
      bool DoErase = true;
      if ( IsArtificiallyIncreaseRegPressureEnabled ) {
        // Increase pressure by adding more interference edges.
        std::string RCName(RC->getName());
        // RAX, RCX, RDX, RSI, RDI, R8, R9, R10, R11,
        // RBX, R14, R15, R12, R13,
        // function params:
        // RDI, RSI, RDX, RCX, R8 , R9
        if ((RCName.find_first_of("GR64") == 0 ||
             RCName.find_first_of("GR32") == 0 ||
             RCName.find_first_of("GR16") == 0 ||
             RCName.find_first_of("GR8")  == 0) 
            &&
            ((RepReg == getRegNumberForName("R10", RegInfo)) ||
            (RepReg == getRegNumberForName("R11", RegInfo)) ||
            (RepReg == getRegNumberForName("R12", RegInfo)) ||
            (RepReg == getRegNumberForName("R13", RegInfo)) ||
            (RepReg == getRegNumberForName("R14", RegInfo)) ||
            (RepReg == getRegNumberForName("R15", RegInfo))) )
          DoErase = false;
        if ((RCName.find_first_of("FR32") == 0 ||
             RCName.find_first_of("FR64") == 0) 
            &&
            ((RepReg == getRegNumberForName("XMM8", RegInfo)) ||
            (RepReg == getRegNumberForName("YMM8", RegInfo)) ||
            (RepReg == getRegNumberForName("XMM9", RegInfo)) ||
            (RepReg == getRegNumberForName("YMM9", RegInfo)) ||
            (RepReg == getRegNumberForName("XMM10", RegInfo)) ||
            (RepReg == getRegNumberForName("YMM10", RegInfo)) ||
            (RepReg == getRegNumberForName("XMM11", RegInfo)) ||
            (RepReg == getRegNumberForName("YMM11", RegInfo)) ||
            (RepReg == getRegNumberForName("XMM12", RegInfo)) ||
            (RepReg == getRegNumberForName("YMM12", RegInfo)) ||
            (RepReg == getRegNumberForName("XMM13", RegInfo)) ||
            (RepReg == getRegNumberForName("YMM13", RegInfo)) ||
            (RepReg == getRegNumberForName("XMM14", RegInfo)) ||
            (RepReg == getRegNumberForName("YMM14", RegInfo)) ||
            (RepReg == getRegNumberForName("XMM15", RegInfo)) ||
            (RepReg == getRegNumberForName("YMM15", RegInfo))) )
          DoErase = false;
      }

      // ExcludedRepresentativeRegs[RC] = Registers(Colors) - AllocatedRegs

      if ( DoErase ) {
        ExcludedRepresentativeRegs[RC].erase(RepReg);
      }

    }

    // ExcludedRepresentativeRegs[RC] = Registers(Colors) - AllocatedRegs - ReservedRegs

    for (std::set<unsigned>::iterator ReservedIter = ReservedRepRegs.begin(); 
         ReservedIter != ReservedRepRegs.end(); ++ReservedIter) {
      unsigned ReservedRepReg = *ReservedIter;

      if ( ExcludedRepresentativeRegs[RC].find(ReservedRepReg) == ExcludedRepresentativeRegs[RC].end() ) {
        ExcludedRepresentativeRegs[RC].insert(ReservedRepReg);
      }
    }

#if 1
    if ( DebugColoring ) {
      dbgs() << "\nExcluding the following registers for register class: "
      << RC->getName() << "\n\t{ ";
      
      for (std::set<unsigned>::iterator ExcludedRegsIter = ExcludedRepresentativeRegs[RC].begin(); 
           ExcludedRegsIter != ExcludedRepresentativeRegs[RC].end(); ++ExcludedRegsIter) {
        unsigned ExcludedReg = *ExcludedRegsIter;
        dbgs() << PrintReg(ExcludedReg, RegInfo) << ", ";
      }
      
      dbgs() << "}\n";
    }
#endif
  }
}

unsigned BuildInterferenceGraph::getSuperReg(unsigned Reg) {
  unsigned SuperReg, i;
 
  if ( TargetRegisterInfo::isPhysicalRegister(Reg) ) {
    const unsigned* Supers = RegInfo->getSuperRegisters(Reg);
    if ( (Supers == NULL) || (Supers[0] == 0) ) {
      SuperReg = Reg;
    }
    else {
      i = 0;
      while ( true ) {
        if ( Supers[i+1] == 0 ) {
          SuperReg = Supers[i];
          break;
        }
        i++;
      }
    }
  }
  else {
    SuperReg = Reg;
  }
  
  return SuperReg;
}


void BuildInterferenceGraph::printRegisterRepresentativeSet() {
  
  dbgs() << "----------- BEGIN REGISTER REPRESENTATIVE DUMP ------------\n";
  
  for (unsigned i = 0; i < RegInfo->getNumRegs(); ++i) {
    if (! isCorrectRegisterType(i))
      continue;
    
    dbgs() << "Reg " << RegInfo->getName(i) << " (" << i << ")  : " << RegisterRepresentativeSet[i]
    << ")\n";
  }
  
  dbgs() << "\nReverse Map:\n";
  
  for(std::map<unsigned, std::set<unsigned> >::iterator I = RepresentativeNum2RegistersSet.begin();
      I != RepresentativeNum2RegistersSet.end(); ++I) {
    
    dbgs() << "Rep Num: " << I->first << " : {";
    
    for (std::set<unsigned>::iterator II = I->second.begin(); II != I->second.end(); ++II)
      dbgs() << *II << " (" << RegInfo->getName(*II) << ") , ";
    
    dbgs() << "}\n";
    
  }
  
  dbgs() << "----------- END REGISTER REPRESENTATIVE DUMP ------------\n";
  
}

///
/// Iterators for RepresentativeSet
///
RepresentativeNumbersIterator BuildInterferenceGraph::getRepresentativeNumbersBegin() {
  return RepresentativeNum2RegistersSet.begin();
}

RepresentativeNumbersIterator BuildInterferenceGraph::getRepresentativeNumbersEnd() {
  return RepresentativeNum2RegistersSet.end();
}

///
/// Build the interference graph creating new live sets
///
bool
BuildInterferenceGraph::build(MachineFunction& F) {
  assert(RegisterRepresentativeSet && "RegisterRepresentativeSet is empty. Initialize not called?");

  if (CurrentGraph != NULL) {
    delete CurrentGraph; CurrentGraph = 0;
  }

  LiveOut = LiveBuilder->GetLiveOutSet();
  LiveIn  = LiveBuilder->GetLiveInSet();
  RefSet  = LiveBuilder->GetRefSet();

  RegInfo = F.getTarget().getRegisterInfo();
  MF = &F;

  // Create the interference graph for the machine function

  //   0           Not a register, can be used as a sentinel.
  //   [1;2^30)    Physical registers assigned by TableGen.
  //   [2^30;2^31) Stack slots. (Rarely used.)
  //   [2^31;2^32) Virtual registers assigned by MachineRegisterInfo.
  // use TargetRegisterInfo::virtReg2Index() to convert to a zero base index
  unsigned MaxPhysicalRegisterNum = F.getTarget().getRegisterInfo()->getNumRegs();
  unsigned MaxVirtualRegisterNum  = F.getRegInfo().getNumVirtRegs(); 

  CurrentGraph = new InterferenceGraph(&F, 
                                       RegInfo,
                                       this,
                                       MaxPhysicalRegisterNum,
                                       MaxVirtualRegisterNum);

  // Construct the interferences.
  buildInterferencesForFunction();

  // Create Adjacency lists from the Interference Matrix
  ///..    CurrentGraph->createNodesAndAdjacencyList()
  // BNW: buildInterferencesForFunction() creates the Adjacency lists aka Edge Sets

  return false;
}

/// copyLiveOutToCurrentLive - Copies the live out set entries to the
/// CurrentLive set if they are of the correct register type according to
/// Builder.
void copyLiveOutToCurrentLive(std::set<unsigned> &LiveOut,
                              DenseSet<unsigned> &CurrentLive,
                              BuildInterferenceGraph *Builder) {
  for (std::set<unsigned>::iterator LI = LiveOut.begin(), LE = LiveOut.end();
       LI != LE; ++LI)
    // Check that the register is one of the once we are looking at.
    if (Builder->isCorrectRegisterType(*LI))
      CurrentLive.insert(*LI);
}

void BuildInterferenceGraph::processRematInfo(InterferenceGraphNode *DefNode,
                                              MachineInstr *MI) {
  const TargetMachine* TM = &MF->getTarget();
  // Identify single definition live ranges that are rematerializable.
  if (!IsRematDisabled &&
      DefNode->isSingleDefUninitialized() &&
      TM->getInstrInfo()->isTriviallyReMaterializable(MI))
    DefNode->setSingleDef(MI);
  else
    DefNode->setSingleDef(NULL);
}

void BuildInterferenceGraph::debugDumpLiveOut(MachineBasicBlock *BB) {
  if ( PrintIGBuild ) {
    dbgs() << "\nSTARTING A NEW BLOCK(" << BB->getName() <<
      "): Current LiveOut is: {\n";
    DumpSet(LiveOut[BB], RegInfo);
  }
}

void BuildInterferenceGraph::debugEraseLiveSetEntry(unsigned LiveSetEntry) {
  if (PrintIGBuild) {
    if ( TargetRegisterInfo::isPhysicalRegister(LiveSetEntry) ) {
      dbgs() << "IG: Erasing " << PrintReg(LiveSetEntry) << "(" <<
        PrintReg(LiveSetEntry, RegInfo) << ")\n";
    } else {
      dbgs() << "IG: Erasing " << PrintReg(LiveSetEntry) << "\n";
    }
  }
}

void BuildInterferenceGraph::debugAddLiveSetEntry(unsigned LiveSetEntry) {
  if (PrintIGBuild) {
    if ( TargetRegisterInfo::isPhysicalRegister(LiveSetEntry) ) {
      dbgs() << "IG: Adding  " << PrintReg(LiveSetEntry) << "(" <<
        PrintReg(LiveSetEntry, RegInfo) << ")\n";
    } else {
      dbgs() << "IG: Adding  " << PrintReg(LiveSetEntry) << "\n";
    }
  }
}

void BuildInterferenceGraph::debugEndBasicBlock() {
  if (PrintIGBuild)
    dbgs() << "}\n";
}

/// buildInterferencesForFunction - To be used with new spill code
/// (SpillHelper).
void
BuildInterferenceGraph::buildInterferencesForFunction() {
  // Build interferences for each basic block.
  for (MachineFunction::iterator BB = MF->begin(), BE = MF->end(); BB != BE;
       ++BB) {
    debugDumpLiveOut(BB);

    DenseSet<unsigned> CurrentLive;

    // Fill current live information with live out info.
    copyLiveOutToCurrentLive(LiveOut[BB], CurrentLive, this);

    // Walk the basic block backwards (end to beginning).
    for (MachineBasicBlock::reverse_iterator RI = BB->rbegin(), RE = BB->rend();
         RI != RE; ++RI) {
      MachineInstr *CurInst = &*RI;

      // Add interferences for definitions with the members of the current live
      // set and remove the definitions from the current live set.
      for (unsigned OpIdx = 0, OpE = CurInst->getNumOperands(); OpIdx != OpE;
           ++OpIdx) {
        MachineOperand &Oper = CurInst->getOperand(OpIdx);

        // Ignore operands that are not registers or not definitions or of the
        // wrong register type.
        if (!Oper.isReg() || !Oper.getReg() || !Oper.isDef() ||
            !isCorrectRegisterType(Oper.getReg())) continue;

        unsigned DefReg = Oper.getReg();
        InterferenceGraphNode* DefNode =
          CurrentGraph->createInterferenceNode(DefReg);

        // Add rematerialization info.
        processRematInfo(DefNode, CurInst);

        // Check whether we are dealing with a copy instruction.
        unsigned CopyFromReg = 0;
        bool IsACopyInst = false;
        if (CopyRegsDoNotInterfere) {
          unsigned DestReg, SrcSubReg, DstSubReg;
          IsACopyInst = isMoveInstr(RegInfo, CurInst, CopyFromReg, DestReg,
                                    SrcSubReg, DstSubReg) &&
            (DestReg == DefReg);
        }

        // Add interferences between current live and the definition.
        for (DenseSet<unsigned>::iterator LI = CurrentLive.begin(),
             LE = CurrentLive.end(); LI != LE; ++LI) {
          unsigned LiveReg = *LI;
          assert(LiveReg != 0);
          if (!IsACopyInst || CopyFromReg != LiveReg)
            CurrentGraph->addInterference(LiveReg, DefReg);
        }

        // Remove the definition from current live.
        CurrentLive.erase(DefReg); debugEraseLiveSetEntry(DefReg);
      }

      // Add uses to current live set.
      for (unsigned OpIdx = 0, OpE = CurInst->getNumOperands(); OpIdx != OpE;
           ++OpIdx) {
        MachineOperand &Oper = CurInst->getOperand(OpIdx);

        // Ignore operands that are not registers or not uses or of the
        // wrong register type.
        if (!Oper.isReg() || !Oper.getReg() || !Oper.isUse() ||
            !isCorrectRegisterType(Oper.getReg())) continue;

        unsigned UseReg = Oper.getReg();

        // Add register to live set.
        CurrentLive.insert(UseReg); debugAddLiveSetEntry(UseReg);

        InterferenceGraphNode *UseNode =
          CurrentGraph->createInterferenceNode(UseReg);
        (void)UseNode; // Silence warning.
      }
    }
    debugEndBasicBlock();
  }

  // Add egdes between the vreg nodes and PR nodes that are not in the vreg's
  // register class. To handle vregs that can only be assigned a subset of the
  // representative physical registers.
  addExcludingEdges();
}

void BuildInterferenceGraph::addExcludingEdges() {
  // Need to add in interference which are derived from the virtual register's
  // register class.  for example on the X86-64 target, virtual registers in the
  // register class GR8_ABCD_H can only be assigned to 4 of the 15 GR8 registers.

  std::set<const TargetRegisterClass*>::iterator ClassesIter, ClassesIterEnd;
  std::set<unsigned>::iterator ExcludedRepRegsIter;
  std::set<unsigned>::iterator ExcludedRepRegsIterEnd;

  ClassesIter    = FilterRegisterClasses.begin();
  ClassesIterEnd = FilterRegisterClasses.end();

  for (; ClassesIter != ClassesIterEnd; ++ClassesIter) {
    const TargetRegisterClass *RC = *ClassesIter;
    if ( RC == 0 ) continue;

    if ( DebugColoring ) {
      dbgs() << "\nfutzing with Register Class: " << RC->getName() << "\n";
    }

    //VirtRegs = MRI->getRegClassVirtRegs(RC);
    //if ( (! VirtRegs.empty()) && (! ExcludedRepresentativeRegs[RC].empty()) ) {
    if ( (RC->getNumRegs() > 0) && (! ExcludedRepresentativeRegs[RC].empty()) ) {

      ExcludedRepRegsIter    = ExcludedRepresentativeRegs[RC].begin();
      ExcludedRepRegsIterEnd = ExcludedRepresentativeRegs[RC].end();

      for (; ExcludedRepRegsIter != ExcludedRepRegsIterEnd;
           ++ExcludedRepRegsIter) {
        unsigned ExcludedRepReg = *ExcludedRepRegsIter;
        unsigned VReg;

#if 0
        MachineRegisterInfo *MRI = &MF->getRegInfo();
        std::vector<unsigned>::iterator VRegIter, VRegIterEnd;
        std::vector<unsigned> VirtRegs;
        VirtRegs = MRI->getRegClassVirtRegs(RC);
        VRegIter    = VirtRegs.begin();
        VRegIterEnd = VirtRegs.end();

        for (; VRegIter != VRegIterEnd; ++VRegIter) {
          VReg = *VRegIter;

          CurrentGraph->addInterference(VReg, ExcludedRepReg);
          if ( DebugColoring ) {
            dbgs() << "Adding intereference for physical register " << PrintReg(VReg,RegInfo) 
            << " and excluded physical representative register " << PrintReg(ExcludedRepReg, RegInfo) 
            << "\n";
          }
        }
#endif

        // virtual registers only
        MachineRegisterInfo &MRI = MF->getRegInfo();
        unsigned i;
        const TargetRegisterClass *VRC;
        for (i=0; i<MRI.getNumVirtRegs(); i++) {
          VReg = TargetRegisterInfo::index2VirtReg(i);
          VRC = MRI.getRegClass(VReg);

          if ( VRC == RC ) {
            CurrentGraph->addInterference(VReg, ExcludedRepReg);
            if ( DebugColoring ) {
              dbgs() << "Adding intereference for virtual register " 
              << PrintReg(VReg)
              << "(" << MRI.getRegClass(VReg)->getName() << ")"
              << " and excluded physical representative register " 
              << PrintReg(ExcludedRepReg, RegInfo) 
              << "\n";
            }
          }
        }
      }
    }
  }
}

InterferenceGraph* BuildInterferenceGraph::getInterferenceGraph() {
  assert (CurrentGraph != NULL);
  return CurrentGraph;
}
