///==--BuildInterferenceGraph.h   - Build an interference graph     -------===//
/// Builds an interference graph
/// The constructor of this class takes a set of register classes. The 
/// graph will only be built for registers in those classes.
///
///==----------------------------------------------------------------------===//
//===----------------------------------------------------------------------===//
//This software was produced with support from the Defense Advanced
//Research Projects Agency (DARPA) through AFRL Contract FA8650-09-C-1915. 
//Nothing in this work should be construed as reflecting the official policy or
//position of the Defense Department, the United States government, or
//Rice University.
//
//THIS SOFTWARE HAS BEEN APPROVED FOR PUBLIC RELEASE, UNLIMITED DISTRIBUTION.
//
//===----------------------------------------------------------------------===//

#ifndef _BUILDINTERFERENCEGRAPH_H_
#define _BUILDINTERFERENCEGRAPH_H_

#include "llvm/CodeGen/MachineFunction.h"
#include "llvm/CodeGen/MachineLoopInfo.h"

#include "llvm/Target/TargetInstrInfo.h"

#include "RegisterClassInfo.h"

#include "BuildLiveSets.h"
#include "InterferenceGraph.h"

#include <map>
#include <set>

namespace llvm {
  
  class Tile;
  
  
  ///
  /// Builds an interference graph.
  ///
  /// This builder uses the concept of representative physical registers (RPR)
  /// Representative physical registers are registers representing an alias set
  /// Thus, for instance, the register alias set {AL, AH, AX, EAX} is 
  /// represented by one representative physical register. RPRs are used as 
  /// nodes in the interference graph since the interference graph must be able 
  /// to express that if EAX is used (colored), then so are registers in its 
  /// alias set, namely: AX, AL, and AH. Similarly, if AH is used (colored), 
  /// so are EAX and AX.
  ///
  /// This approach loses some granularity. For instance, if AL is colored, then 
  /// AH can be assigned a different color. However, it is unclear how to encode
  /// this elegantly into an interference graph.
  ///
  /// 
  class BuildInterferenceGraph {
    
  public:
    BuildLiveSets* LiveBuilder;

    // Map a subset register class to a set of registers contained in the super register class
    // but not in the subset.
    std::map<const TargetRegisterClass*, std::set<unsigned> > ExcludedRepresentativeRegs;

    // Map the number register in the super class to the subset register class. 
    // Edges from ExcludedRepresentativeRegs are added to the IG to make the K value work.
    std::map<const TargetRegisterClass*, unsigned > Ks;
    unsigned getK(const TargetRegisterClass *RC) {
      return Ks[RC];
    }
    
  protected:
    BlockDataMap LiveIn, LiveOut;
    BlockDataMap RefSet;            // CK allocator only.
    
    std::map<MachineBasicBlock*, std::map<unsigned, unsigned> > UseCounts;
    std::map<MachineBasicBlock*, std::map<unsigned, unsigned> > DefCounts;
 

    ///
    /// Set of TargetRegisterClasses. Only registers belonging to a class in this set
    /// will be considered by the builder. All other registers in the MachineFunction 
    /// will be ignored
    ///
    std::set<const TargetRegisterClass*> FilterRegisterClasses;
    std::set<unsigned> FilterRegisterClassIDs;
    
    InterferenceGraph* CurrentGraph;
    
    const TargetRegisterInfo* RegInfo;
    
    MachineFunction* MF;
    
    /// Map which establishes a representative physical register for all registers in 
    /// an alias set
    unsigned *RegisterRepresentativeSet;
    
    // Map from representative physical register number to all registers in the 
    // RPR set
    std::map<unsigned, std::set<unsigned> > RepresentativeNum2RegistersSet;
    
    ///
    /// buildRegisterRepresentativeSet()
    /// Builds a mapping from register number -> representative number for physical registers
    /// The representative number for a register alias set is the register number of one of the registers
    /// in the set
    ///
    void buildRegisterRepresentativeSet();
    
  public:
    
    BuildInterferenceGraph();
    
    virtual ~BuildInterferenceGraph();
    
    // Initialize -- used since we cannot call a virtual method in base constructor
    void Initialize();
    
    void buildExcludedRepresentativeRegs(std::set<int> &Colors, 
                                         std::map<int, unsigned> &Color2RepRegMap,
                                         RegisterClassInfo &RCI);
        
    void printRegisterRepresentativeSet();
    
    ///
    /// This function is invoked to only consider those registers that were
    /// passed in via the constructor
    ///
    virtual bool isCorrectRegisterType(unsigned) const;
    
    ///
    /// This function is invoked to only consider those register classess that were
    /// passed in via the constructor
    ///
    virtual bool isCorrectRegisterClass(const TargetRegisterClass *RC) const;
    
    // Constructor that takes the targetmachine and a set of TargetRegisterClasses.
    // The interference graph will be built for only those registers whose register
    // class belongs in this set.
    BuildInterferenceGraph(MachineFunction* F,
                           const TargetMachine &TM,
                           std::set<const TargetRegisterClass*> &RegClassSet,
                           BuildLiveSets *InitialLiveBuilder = NULL);
    BuildInterferenceGraph(const BuildInterferenceGraph&);
    
    RegisterSet& getRegistersOfRepresentative(unsigned RepReg) {
      assert(RepresentativeNum2RegistersSet.find(RepReg)
             != RepresentativeNum2RegistersSet.end());
      return RepresentativeNum2RegistersSet[RepReg];
    }

    unsigned getRepresentativeNumber(unsigned Reg) {
      if ( TargetRegisterInfo::isPhysicalRegister(Reg) ) {
        if ( RegisterRepresentativeSet[Reg] != 0 ) {
          return RegisterRepresentativeSet[Reg];
        }
        else {
          return Reg;
        }
      }
      else {
        return Reg;
      }
    }
    
    unsigned getSuperReg(unsigned Reg);

    ///
    /// Iterators for RepresentativeSet
    ///
    RepresentativeNumbersIterator getRepresentativeNumbersBegin(); 
    RepresentativeNumbersIterator getRepresentativeNumbersEnd(); 


    InterferenceGraph* getInterferenceGraph();
    BuildLiveSets *getLiveBuilder() { return LiveBuilder; }

    BlockDataMap& getLiveOutSet() { return LiveOut; }
    BlockDataMap& getLiveInSet() { return LiveIn; }
    BlockDataMap& getRefSet() { return RefSet; }


    ///
    /// Build the interference graph
    ///
    virtual bool build(MachineFunction &);

    void buildInterferencesForFunction();

    void processRematInfo(InterferenceGraphNode *DefNode, MachineInstr *MI);

    /// Some register classes only contain subset of  the representative
    /// registers. For example, GR64_ABCD contains a subset of the full set
    /// GR64. To deal with that we add the difference between the two as
    /// interference egdes to the virtual register not.
    void addExcludingEdges();

    void debugDumpLiveOut(MachineBasicBlock *BB);
    void debugEraseLiveSetEntry(unsigned LiveSetEntry);
    void debugAddLiveSetEntry(unsigned LiveSetEntry);
    void debugEndBasicBlock();
  };
}

#endif
