//===-- BuildLiveSets.cpp - Builds live sets                        -------===//
// Builds live sets
// 
// Anshuman Dasgupta and Jason Eckhardt, MSCP, Rice University
//
//===----------------------------------------------------------------------===//
//===----------------------------------------------------------------------===//
//This software was produced with support from the Defense Advanced
//Research Projects Agency (DARPA) through AFRL Contract FA8650-09-C-1915. 
//Nothing in this work should be construed as reflecting the official policy or
//position of the Defense Department, the United States government, or
//Rice University.
//
//THIS SOFTWARE HAS BEEN APPROVED FOR PUBLIC RELEASE, UNLIMITED DISTRIBUTION.
//
//===----------------------------------------------------------------------===//

#define DEBUG_TYPE "regalloc"

#include "llvm/Support/CommandLine.h"
#include "llvm/Support/Debug.h"

#include "llvm/Pass.h"

#include "llvm/CodeGen/MachineFunctionPass.h"
#include "llvm/CodeGen/MachineInstr.h"
#include "llvm/CodeGen/LiveVariables.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/Target/TargetRegisterInfo.h"

#include "llvm/ADT/PostOrderIterator.h"

#include "llvm/Transforms/Scalar.h"
#include "llvm/Support/CFG.h"
#include "llvm/CodeGen/Passes.h"
#include "llvm/BasicBlock.h"

#include <algorithm>

#include "Common.h"

#include "BuildLiveSets.h"
#include "ConstructRegisterClasses.h"

namespace llvm {
  
  cl::opt<bool> DebugBuildLive("debugbuildlive", cl::init(false));
  
  //  static RegisterAnalysis<BuildLiveSets> LS_Random1("build-live-sets", "Build Live Sets");
  
  BuildLiveSets::BuildLiveSets(MachineFunction& F) {
    MF = &F;
    TM  = &F.getTarget();
    TRI = TM->getRegisterInfo();
  }
  
  BuildLiveSets::~BuildLiveSets() {
    LiveIn.clear();
    LiveOut.clear();
    Killed.clear();
    UpwardUses.clear();
  }
  
  // Build local kill and Upward exposed uses sets
  // Kill is SSA-land is just def -- values cannot be redefined
  //    void BuildLiveSets::buildKilledAndUESets( ReversePostOrderTraversal<MachineFunction*>& RPOT) {
  
  bool BuildLiveSets::build(MachineFunction &F) {
    buildKilledAndUESets(F);
    
    buildLiveSets(F);

    if ( DebugBuildLive ) {
      Dump(F);
    }
    
    return false;
  }
  
  // Build local kill and Upward exposed uses sets
  void BuildLiveSets::buildKilledAndUESets(MachineFunction& F) {
    
    DEBUG(dbgs() << "Entered Killed and UE sets!!!\n");    
    
    // Should use ReversePostOrder here but GraphTraits<MachineFunction*> is 
    // not a valid specialization (unlike for Function* defined in CFG.h)    
    for (MachineFunction::iterator BlockIter = F.begin(); 
         BlockIter != F.end(); ++BlockIter) {
      
      buildBlockKilledAndUESets( BlockIter);
      
    }
  }
  
  void BuildLiveSets::buildBlockKilledAndUESets(MachineBasicBlock* block) {
    unsigned Register;

    // Block specific kill and UE sets
    ValueSetType& BlockKilled = Killed[ block  ];
    ValueSetType& BlockUpwardUses = UpwardUses[ block ];
    ValueSetType& BlockRefSet = RefSet[ block ];
    ValueSetType& BlockUsesSet = Uses[ block ];
    
    if ( DebugBuildLive ) {
      dbgs() << "\nBuilding KILL and UEVARS for basic block: " << block->getName() << "\n";
    }
    for (MachineBasicBlock::iterator InstIter = block->begin();
         InstIter != block->end();
         ++InstIter) {
      
      MachineInstr * MI = InstIter;
      if ( DebugBuildLive ) {
        dbgs() << "\tProcessing machine instruction: " << *MI;
      }
      
      const MCInstrDesc &MCID = TM->getInstrInfo()->get(MI->getOpcode());
      
      // First iterate over uses	  
      
      //
      // Implicit uses are physical registers
      //
      for (const unsigned *ImplicitUses = MCID.ImplicitUses;
           ImplicitUses != NULL && *ImplicitUses; ++ImplicitUses) {
        if (*ImplicitUses != 0) {
          
          // If it's not in the Killed (Def) set, then it's an UE Use
          if ( BlockKilled.find(*ImplicitUses) == BlockKilled.end()) {
            BlockUpwardUses.insert(*ImplicitUses);
            if ( DebugBuildLive ) {
              if ( TargetRegisterInfo::isPhysicalRegister(*ImplicitUses) ) {
                dbgs() << "\t\tNot killed so adding implicit use to UE: " << PrintReg(*ImplicitUses) 
                << "(" << PrintReg(*ImplicitUses, TRI) << ")\n";
              }
              else {
                dbgs() << "\t\tNot killed so adding implicit use to UE: " << PrintReg(*ImplicitUses) << "\n";
              }
            }
          }
          else {
            if ( DebugBuildLive ) {
              if ( TargetRegisterInfo::isPhysicalRegister(*ImplicitUses) ) {
                dbgs() << "\t\tKilled so not adding implicit use to UE: " << PrintReg(*ImplicitUses) 
                << "(" << PrintReg(*ImplicitUses, TRI) << ")\n";
              }
              else {
                dbgs() << "\t\tKilled so not adding implicit use to UE: " << PrintReg(*ImplicitUses) << "\n";
              }
            }
          }

          if ( TargetRegisterInfo::isPhysicalRegister(*ImplicitUses) ) {
            // Iterate through its alias set and mark them as used
            for (const unsigned* aliasIter = TRI->getAliasSet(*ImplicitUses);
                 aliasIter && *aliasIter; ++aliasIter) {
              // If it's not in the Killed (Def) set, then it's an UE Use
              if ( BlockKilled.find(*aliasIter) == BlockKilled.end()) {
                BlockUpwardUses.insert(*aliasIter);
                if ( DebugBuildLive ) {
                  dbgs() << "\t\tNot killed so adding implicit use to UE: " << PrintReg(*aliasIter) 
                  << "(" << PrintReg(*aliasIter, TRI) << ")\n";
                }
              }
              else {
                if ( DebugBuildLive ) {
                  dbgs() << "\t\tKilled so not adding implicit use to UE: " << PrintReg(*aliasIter) 
                  << "(" << PrintReg(*aliasIter, TRI) << ")\n";
                }
              }
              BlockRefSet.insert(*aliasIter);
            } //for
          } //if
          // Add to Refs.
          BlockRefSet.insert(*ImplicitUses);
          
          // Add to Uses
          BlockUsesSet.insert(*ImplicitUses);
          
          // increment UseCount
          ++UseCounts[block][*ImplicitUses];
        }
      }
      
      
      // 
      // Explicit uses
      //
      for (unsigned i = 0, e = MI->getNumOperands(); i != e; ++i) {
        
        // If it's a use (physical or virtual register)
        // and is not in the killed set then, its upward exposed
        MachineOperand Oper = MI->getOperand(i);
        if (Oper.isReg() && Oper.isUse() && (Oper.getReg() != 0)) {
          if (BlockKilled.find(Oper.getReg()) == BlockKilled.end()) {
            BlockUpwardUses.insert(Oper.getReg());
            if ( DebugBuildLive ) {
              if ( TargetRegisterInfo::isPhysicalRegister(Oper.getReg()) ) {
                dbgs() << "\t\tNot killed so adding explicit use to UE: " << PrintReg(Oper.getReg()) 
                << "(" << PrintReg(Oper.getReg(), TRI) << ")\n";
              }
              else {
                dbgs() << "\t\tNot killed so adding explicit use to UE: " << PrintReg(Oper.getReg()) << "\n";
              }
            }
          }
          else {
            if ( DebugBuildLive ) {
              if ( TargetRegisterInfo::isPhysicalRegister(Oper.getReg()) ) {
                dbgs() << "\t\tKilled so not adding explicit use to UE: " << PrintReg(Oper.getReg()) 
                << "(" << PrintReg(Oper.getReg(), TRI) << ")\n";
              }
              else {
                dbgs() << "\t\tKilled so not adding explicit use to UE: " << PrintReg(Oper.getReg()) << "\n";
              }
            }
          }
          
          if ( TargetRegisterInfo::isPhysicalRegister(Oper.getReg()) ) {
            // Iterate through its alias set and mark them as used
            for (const unsigned* aliasIter = TRI->getAliasSet(Oper.getReg());
                 aliasIter && *aliasIter; ++aliasIter) {
              if (BlockKilled.find(*aliasIter) == BlockKilled.end()) {
                BlockUpwardUses.insert(*aliasIter);
                if ( DebugBuildLive ) {
                  dbgs() << "\t\tNot killed so adding explicit use to UE: " << PrintReg(*aliasIter) 
                  << "(" << PrintReg(*aliasIter, TRI) << ")\n";
                }
              }
              else {
                if ( DebugBuildLive ) {
                  dbgs() << "\t\tKilled so not adding explicit use to UE: " << PrintReg(*aliasIter) 
                  << "(" << PrintReg(*aliasIter, TRI) << ")\n";
                }
              }
              BlockRefSet.insert(*aliasIter);
            } //for
          } //if
          
          // Add to Refs.
          BlockRefSet.insert(Oper.getReg());
          
          // Add to Uses
          BlockUsesSet.insert(Oper.getReg());
          
          // Increment UseCount
          ++UseCounts[block][Oper.getReg()];
          
        }
      }
      
      // Then iterate over defs
      
      //
      // Implicit Defs for physical registers
      //
      for (const unsigned *ImplicitDefs = MCID.ImplicitDefs;
           MCID.ImplicitDefs != NULL && *ImplicitDefs; ++ImplicitDefs) {
        Register = *ImplicitDefs;
        if (*ImplicitDefs != 0) {
          
          // Add to killed set
          BlockKilled.insert( *ImplicitDefs);
          if ( DebugBuildLive ) {
            dbgs() << "\t\tAdding implicit definition to KILL: " << PrintReg(*ImplicitDefs) 
            << "(" << PrintReg(*ImplicitDefs, TRI) << ")\n";
          }
          
          if ( TargetRegisterInfo::isPhysicalRegister(*ImplicitDefs) ) {
            
            // Kill all registers that are in its alias set 
            // and are at most as big as the register
            for (const unsigned* aliasIter = TRI->getAliasSet(*ImplicitDefs);
                 aliasIter && *aliasIter; ++aliasIter) {
              BlockKilled.insert( *aliasIter);
              if ( DebugBuildLive ) {
                dbgs() << "\t\tAdding implicit definition to KILL: " << PrintReg(*ImplicitDefs) 
                << "(" << PrintReg(*ImplicitDefs, TRI) << ")\n";
              }
              BlockRefSet.insert(*aliasIter);
            } //for
          } //if
          
          // Add to Refs.
          BlockRefSet.insert( *ImplicitDefs);
          
          // Increment DefCounts
          ++DefCounts[block][*ImplicitDefs];
          
        }
      }
      
      // 
      // Explicit defs
      //
      for (unsigned i = 0, e = MI->getNumOperands(); i != e; ++i) {
        
        // If it's a def (physical or virtual register)
        MachineOperand& Oper = MI->getOperand(i);
        if (Oper.isReg() && Oper.isDef() && (Oper.getReg() != 0)) {
          BlockKilled.insert(Oper.getReg());
          if ( DebugBuildLive ) {
            if ( TargetRegisterInfo::isPhysicalRegister(Oper.getReg()) ) {
              dbgs() << "\t\tAdding explicit definition to KILL: " << PrintReg(Oper.getReg()) 
              << "(" << PrintReg(Oper.getReg(), TRI) << ")\n";
            }
            else {
              dbgs() << "\t\tAdding explicit definition to KILL: " << PrintReg(Oper.getReg()) << "\n";
            }
          }
          
          // Kill all registers that are in its alias set 
          // and are at least as big as the register
          if ( TargetRegisterInfo::isPhysicalRegister(Oper.getReg()) ) {
            
            // Kill all registers that are in its alias set 
            // and are at most as big as the register
            for (const unsigned* aliasIter = TRI->getAliasSet(Oper.getReg());
                 aliasIter && *aliasIter; ++aliasIter) {
              Register = *aliasIter;
              BlockKilled.insert( *aliasIter);
              if ( DebugBuildLive ) {
                dbgs() << "\t\tAdding explicit definition to KILL: " << PrintReg(*aliasIter) 
                << "(" << PrintReg(*aliasIter, TRI) << ")\n";
              }
              BlockRefSet.insert(*aliasIter);
            } //for
          } //if
          
          // Add to Refs.
          BlockRefSet.insert(Oper.getReg());
          
          // Add to DefCounts
          ++DefCounts[block][Oper.getReg()];
        }
      }
    }
    
    /// Initialize the LiveIn set to the upward exposed sets
    LiveIn[block] = BlockUpwardUses;
    
    /// Initialize the Refs set.
    RefSet[block] = BlockRefSet;

    if ( DebugBuildLive ) {
      ValueSetType::iterator RegisterIter;

      if ( !Killed[block].empty() ) {
        dbgs() << "Killed[" << block->getName() << "] has " << Killed[block].size() << " entries.\n";
        DumpSet(Killed[block]);
      }
      else {
        dbgs() << "Killed[" << block->getName() << "] is empty\n";
      }
      
      // dump BlockDataMap UpwardUses
      if ( !UpwardUses[block].empty() ) {
        dbgs() << "UpwardUses[" << block->getName() << "] has " << UpwardUses[block].size() << " entries.\n";
        DumpSet(UpwardUses[block]);
      }
      else {
        dbgs() << "UpwardUses[" << block->getName() << "] is empty\n";
      }
    }
  }
  
  
  
  ///
  /// buildLiveSets -- iteratively build up LiveIn and LiveOut
  /// LiveIn = LiveOut - Killed + UE
  /// LiveOut = Union(LiveIn(s))
  ///
  /// Since we are in SSA land,  Killed = Def
  ///
  void BuildLiveSets::buildLiveSets(MachineFunction& F)
  {
    std::list<MachineBasicBlock*> worklist;
    
    // BNW: Dr. Tim says we should reverse the edges in the CFG, compute a post ordering,
    // and then reverse it.  Below, it appears that no attempt was made to optimize the
    // efficienvy of the Live In/Outcal calculation.

    for(MachineFunction::iterator FI = F.begin(), FE = F.end(); FI != FE; ++FI) {
      worklist.push_back(&(*FI));
    }
    
    buildLiveSetsWorklist(worklist);
  }
  
  
  
  void BuildLiveSets::buildLiveSetsWorklist(std::list<MachineBasicBlock* >& worklist) {
    
    // BNW: Not a bit vector in sight.  Thus, not built for speed.

    ValueSetType OldLiveIn, OldLiveOut;
    bool changed = false;
    
    std::set<MachineBasicBlock*> workset(worklist.begin(), worklist.end());
    
    for (std::list<MachineBasicBlock* >::iterator blockIter = worklist.begin();
         blockIter != worklist.end(); ++blockIter) {
      
      MachineBasicBlock* currentBlock = *blockIter;

      OldLiveOut = LiveOut[currentBlock];
      LiveOut[currentBlock].clear();
      
      // Remove the currentblock from workset. This is needed at the top of
      // the loop since currentblock can be in its own predecessor set
      // and in that case should not be in workset when the membership check is made

      workset.erase(currentBlock);
      
      
      // For all successors of the block s, LiveOut = Union(LiveIn(s))

      for (MachineBasicBlock::succ_iterator SI = currentBlock->succ_begin();
           SI != currentBlock->succ_end();
           ++SI) {
        MachineBasicBlock* CurrentSucc = *SI;
        LiveOut[currentBlock].insert(LiveIn[CurrentSucc].begin(), LiveIn[CurrentSucc].end());
      }
      
      if (OldLiveOut != LiveOut[currentBlock]) {
        changed = true;
      }
      
      
      // Now that we have the LiveOut set, calculate LiveIn
      
      if (changed) {
        
        // LiveOut - Killed
        ValueSetType DifferenceSet;
        set_difference(LiveOut[currentBlock].begin(), LiveOut[currentBlock].end(),
                       Killed[currentBlock].begin(), Killed[currentBlock].end(),
                       inserter(DifferenceSet, DifferenceSet.begin()));	
        
        // LiveIn = (LiveOut - Killed) + UE, initially LiveIn = UE
        LiveIn[currentBlock].insert(DifferenceSet.begin(), DifferenceSet.end());
        
        
        // Since live sets have changed, add predecessors to the end of the worklist,
        // if they are not currently in the worklist
        
        for (MachineBasicBlock::pred_iterator PI = currentBlock->pred_begin(), 
             PE = currentBlock->pred_end();
             PI != PE; ++PI) {
          if (workset.find(*PI) == workset.end()) {
            worklist.push_back(*PI);
            workset.insert(*PI);
          }
        }
        
        changed = false;
        
      }
      
    }
    
  }

#include <stdio.h>

  void BuildLiveSets::Dump(MachineFunction& MF) {
    MachineFunction::iterator MBBIter;
    MachineBasicBlock* MBB;
        
    const TargetRegisterInfo *TRI = MF.getTarget().getRegisterInfo();

    dbgs() << "Dumping LiveIn and LiveOut for Function: " << MF.getFunction()->getName() << "\n";
    
    dbgs() << "Target has " << TRI->getNumRegs() << " registers.\n"; 
    
    for (MBBIter = MF.begin(); MBBIter != MF.end(); ++MBBIter) {
      MBB = MBBIter;
      dbgs() << "\nBasic block starting with label: " << MBB->getName() << "\n";

      // dump BlockDataMap Killed
      if ( !Killed[MBB].empty() ) {
        dbgs() << "Killed[" << MBB->getName() << "] has " << Killed[MBB].size() << " entries.\n";
        DumpSet(Killed[MBB]);
      }
      else {
        dbgs() << "Killed is empty\n";
      }
      
      // dump BlockDataMap UpwardUses
      if ( !UpwardUses[MBB].empty() ) {
        dbgs() << "UpwardUses[" << MBB->getName() << "] has " << UpwardUses[MBB].size() << " entries.\n";
        DumpSet(UpwardUses[MBB]);
      }
      else {
        dbgs() << "UpwardUses is empty\n";
      }
      
      // dump BlockDataMap LiveIn
      if ( !LiveIn[MBB].empty() ) {
        dbgs() << "LiveIn[" << MBB->getName() << "] has " << LiveIn[MBB].size() << " entries.\n";
        DumpSet(LiveIn[MBB]);
      }
      else {
        dbgs() << "LiveIn is empty\n";
      }

      // dump BlockDataMap LiveOut
      if ( !LiveOut[MBB].empty() ) {
        dbgs() << "LiveOut[" << MBB->getName() << "] has " << LiveOut[MBB].size() << " entries.\n";
        DumpSet(LiveOut[MBB]);
      }
      else {
        dbgs() << "LiveOut is empty\n";
      }
    }
  }
  

  void DumpSet(ValueSetType &Set, const TargetRegisterInfo *TRI) {
    ValueSetType::iterator RegisterIter;
    unsigned Register;
    unsigned NumPhysRegs;
    unsigned RepReg;
    bool Found;
    
    // Dump the Physical Registers one Super Set per line
    
    NumPhysRegs = TRI->getNumRegs();
    for (RepReg=1; RepReg<=NumPhysRegs; ++RepReg) {
      Found = false;
      for (RegisterIter = Set.begin(); RegisterIter != Set.end(); ++RegisterIter) {
        Register = *RegisterIter;
        if ( TargetRegisterInfo::isPhysicalRegister(Register) ) {
          //if ( getRepresentativeNumber(Register, TRI) == RepReg ) {
            if ( !Found ) {
              Found = true;
              dbgs() << "\t";
            }
            dbgs() << PrintReg(Register) << "(" << PrintReg(Register, TRI) << ") ";
          //}
        }
      }
      if ( Found ) {
        dbgs() << "\n";
      }
    }
    
    // Now dump all of the virtual registers on a single line
    
    Found = false;
    for (RegisterIter = Set.begin(); RegisterIter != Set.end(); ++RegisterIter) {
      Register = *RegisterIter;
      if ( TargetRegisterInfo::isVirtualRegister(Register) ) {
        if ( !Found ) {
          Found = true;
          dbgs() << "\t";
        }
        dbgs() << PrintReg(Register) << " ";
      }
    }
    if ( Found ) {
      dbgs() << "\n";
    }
  }

} //namespace llvm

