//===--BuildLiveSets.h   Interface for live sets ----*- C++ -*-===//
//===----------------------------------------------------------------------===//
//
// This file defines the interface for live set creation
//
// Anshuman Dasgupta and Jason Eckhardt, MSCP, Rice University
//
//===----------------------------------------------------------------------===//
//===----------------------------------------------------------------------===//
//This software was produced with support from the Defense Advanced
//Research Projects Agency (DARPA) through AFRL Contract FA8650-09-C-1915. 
//Nothing in this work should be construed as reflecting the official policy or
//position of the Defense Department, the United States government, or
//Rice University.
//
//THIS SOFTWARE HAS BEEN APPROVED FOR PUBLIC RELEASE, UNLIMITED DISTRIBUTION.
//
//===----------------------------------------------------------------------===//

#ifndef BUILD_LIVE_SETS
#define BUILD_LIVE_SETS

#include "llvm/CodeGen/MachineFunction.h"

#include <set>
#include<list>
#include <map>

namespace llvm {

  typedef std::set<unsigned> ValueSetType;
  typedef std::map<MachineBasicBlock*, ValueSetType> BlockDataMap;
  
  extern void DumpSet(ValueSetType &Set, const TargetRegisterInfo *TRI);

  
  //////////////////////////////////////////////////////////////////////////////
  ///
  /// Build LiveIn, LiveOut, Killed, and Upward-exposed Uses sets for each 
  /// MachineBlock
  /// 
  //////////////////////////////////////////////////////////////////////////////

  class BuildLiveSets {

  private:
  

  public:

    ///
    /// buildLiveSets -- iteratively build up LiveIn and LiveOut
    /// LiveIn = LiveOut - Killed + UE
    /// LiveOut = Union(LiveIn(s))
    ///
    /// Since we are in SSA land,  Killed = Def
    ///
    BuildLiveSets(MachineFunction& F);
    virtual ~BuildLiveSets();

    
    /// 
    /// The main entry point for the live set builder
    /// Builds live information for the specified MachineFunction
    ///
    virtual bool build(MachineFunction &);

    BlockDataMap &GetLiveInSet() { return LiveIn;};
    BlockDataMap &GetLiveOutSet() { return LiveOut;};
    BlockDataMap &GetKilledSet() { return Killed;};

    BlockDataMap &GetRefSet() { return RefSet;};
    BlockDataMap &GetUsesSet() { return Uses; };

    std::map<MachineBasicBlock*, std::map<unsigned, unsigned> > &GetUseCounts() { return UseCounts; }
    std::map<MachineBasicBlock*, std::map<unsigned, unsigned> >  &GetDefCounts() { return DefCounts; }

    BlockDataMap& GetUpwardUses() { return UpwardUses; }
    
    void Dump(MachineFunction& MF);
    void DumpSet(ValueSetType &Set) { llvm::DumpSet(Set, TRI); }


  private:
    

    BlockDataMap LiveIn;
    BlockDataMap LiveOut;
    BlockDataMap RefSet;           // Hierarchical allocator only.
    BlockDataMap Uses;             // Chow allocator only

    std::map<MachineBasicBlock*, std::map<unsigned, unsigned> > UseCounts; // Chow allocator only
    std::map<MachineBasicBlock*, std::map<unsigned, unsigned> > DefCounts; // Chow allocator only

    BlockDataMap Killed;
    BlockDataMap UpwardUses;
    
    MachineFunction *MF;
    const TargetMachine* TM;
    const TargetRegisterInfo *TRI;

    void buildKilledAndUESets(MachineFunction& ); 
    void buildBlockKilledAndUESets(MachineBasicBlock* );
    void buildLiveSets(MachineFunction& );
    void buildLiveSetsWorklist(std::list<MachineBasicBlock* >& );


  };
  

  //  MachineFunctionPass* createLiveSetBuilder();
  
  
}
  


#endif
