//===--InterferenceGraph.cpp - Defines an interference graph         -------===//
// Defines an interference graph
// Anshuman Dasgupta and Jason Eckhardt, MSCP, Rice University
// 
//===----------------------------------------------------------------------===//
//===----------------------------------------------------------------------===//
//This software was produced with support from the Defense Advanced
//Research Projects Agency (DARPA) through AFRL Contract FA8650-09-C-1915. 
//Nothing in this work should be construed as reflecting the official policy or
//position of the Defense Department, the United States government, or
//Rice University.
//
//THIS SOFTWARE HAS BEEN APPROVED FOR PUBLIC RELEASE, UNLIMITED DISTRIBUTION.
//
//===----------------------------------------------------------------------===//

#include "llvm/CodeGen/MachineInstr.h"
#include "llvm/CodeGen/MachineFrameInfo.h"
#include "llvm/CodeGen/MachineRegisterInfo.h"

#include "llvm/Target/TargetMachine.h"
#include "llvm/Target/TargetRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"

#include <algorithm>

#include "llvm/Support/CommandLine.h"
#include "llvm/Support/Debug.h"
#include "llvm/Support/raw_ostream.h"

#include "InterferenceGraph.h"
#include "BuildInterferenceGraph.h"
#include "Common.h"

#include <list>

using namespace llvm;

extern cl::opt<bool> PrintSpillCosts;
extern cl::opt<bool> DebugCoalesce;  
extern cl::opt<bool> DebugColoring;  
extern cl::opt<bool> DebugSpills;  
extern cl::opt<SpillHeuristicName> SpillHeuristic;


unsigned InterferenceGraph::getRepresentativeNumber(unsigned Reg) {
  return Builder->getRepresentativeNumber(Reg);
}


//
// Interference Graph has two abstractions that require a "dense" or compact index.
// LLVM assigns register numbers based within certain ranges base on the type of
// register.  Need both the virtual and physical register numbers to be contiguous
// for the Interference Graph data structure.  The Interference Graph itself is a 
// triangular bit matrix, allocated as a one dimensional bit vector.
//

/*static*/ unsigned InterferenceGraph::getDenseIndex(unsigned RepReg) {
  // input parameter Reg can have the following values:
  //   0           Not a register, can be used as a sentinel.
  //   [1;2^30)    Physical registers assigned by TableGen.
  //   [2^30;2^31) Stack slots. (Rarely used.)
  //   [2^31;2^32) Virtual registers assigned by MachineRegisterInfo.
  // use TargetRegisterInfo::virtReg2Index() to convert to a zero base index
  if        ( TargetRegisterInfo::isPhysicalRegister(RepReg)  ) {
    return RepReg;
  } else if ( TargetRegisterInfo::isVirtualRegister(RepReg) ) {
    return ( InterferenceGraph::FirstDenseVirtualRegisterNum + 
             TargetRegisterInfo::virtReg2Index(RepReg) );
  }
  else {
    return 0;
  }
}


unsigned Reg2DenseIndexFunctor::operator()(unsigned RepReg) const { 
  return InterferenceGraph::getDenseIndex(RepReg);
}


/*static*/ unsigned InterferenceGraph::MaxPhysicalRegisterNum        = 0;
/*static*/ unsigned InterferenceGraph::MaxVirtualRegisterNum         = 0;
/*static*/ unsigned InterferenceGraph::FirstDensePhysicalRegisterNum = 0;
/*static*/ unsigned InterferenceGraph::FirstDenseVirtualRegisterNum  = 0;

/*static*/ unsigned InterferenceGraph::SimplifyIterations            = 0;


///
/// Many of these values should be calculated in BuildInterferenceGraph 
/// and passed into the ctor 
///
InterferenceGraph::InterferenceGraph(const MachineFunction* MachFn, 
                                     const TargetRegisterInfo* TRI,
                                     BuildInterferenceGraph *Builder,
                                     unsigned MaxPhysicalRegisterNum, 
                                     unsigned MaxVirtualRegisterNum)
: MF(MachFn), RegInfo(TRI), RegisterNumToNodeMap(NULL), Builder(Builder)
{
  MF = MachFn;
  RegInfo = TRI;
  
  //   0           Not a register, can be used as a sentinel.
  //   [1;2^30)    Physical registers assigned by TableGen.
  //   [2^30;2^31) Stack slots. (Rarely used.)
  //   [2^31;2^32) Virtual registers assigned by MachineRegisterInfo.
  MaxSizeOfIG = 1 + MaxPhysicalRegisterNum + MaxVirtualRegisterNum;
  InterferenceGraph::MaxPhysicalRegisterNum = MaxPhysicalRegisterNum;
  InterferenceGraph::MaxVirtualRegisterNum = MaxVirtualRegisterNum;
  InterferenceGraph::FirstDensePhysicalRegisterNum = 1;
  InterferenceGraph::FirstDenseVirtualRegisterNum = InterferenceGraph::FirstDensePhysicalRegisterNum + 
                                                    InterferenceGraph::MaxPhysicalRegisterNum;
  
  InterferenceGraph::RegClass2KMap.clear();

  // Allocate a triangular bitmatrix.
  InterferenceMatrix.resize(MaxSizeOfIG * MaxSizeOfIG / 2);
  
  RemovedNodes.resize(MaxSizeOfIG);
  
  // Set 0 (NULL) as the "not-found" value for the DenseMap
  //RegisterNumToNodeMap((const llvm::InterferenceGraphNode*) NULL);
  
  // All bits in a std::bitset are initially zero -- hence no need to initialize
}


// dtor
InterferenceGraph::~InterferenceGraph() {
  
  for (std::set<InterferenceGraphNode*>::iterator deleteIter = AllNodes.begin(); 
       deleteIter != AllNodes.end();
       ++deleteIter) {
    
    delete (*deleteIter);
  }
  
}

void InterferenceGraph::GrowMatrix(unsigned Size) {
  InterferenceMatrix.resize(Size);

  RemovedNodes.resize(Size);
}

///
/// Add a node to the graph only if it does not already exist
///
InterferenceGraphNode* InterferenceGraph::createInterferenceNode(unsigned reg, 
                                                                 MachineOperand* moper) {
  
  // Convert to representative number
  unsigned RepReg = getRepresentativeNumber(reg);
  unsigned DenseReg = getDenseIndex(RepReg);
  
  // If the node number is greater than the maximum number of nodes,
  // we have to expand our interference matrix
  if (DenseReg >= MaxSizeOfIG) {
    unsigned newsize;
    
    newsize = DenseReg + 50;
    
    GrowMatrix(newsize);
    MaxSizeOfIG = newsize;
  }
  
  InterferenceGraphNode* NewNode = NULL;
  
  if ( getNodeOrNull(RepReg) == getIGNullVal() )
  {
    
    // Cannot have register 0 since that will be the "not-found" value 
    // for the RegisterNumToNodeMap DenseMap
    assert(RepReg != 0);
    
    NewNode = new InterferenceGraphNode(RepReg);
    RegisterNumToNodeMap[RepReg] = NewNode;
    AllNodes.insert(NewNode);
    if ( TargetRegisterInfo::isPhysicalRegister(reg) ) {
      PhysicalRegNodes.insert(NewNode);
    }
  }
  else {
    NewNode = RegisterNumToNodeMap[RepReg];
  }
  
  NewNode->setRegOperand(moper);
  
  //assert(AllNodes.size() == RegisterNumToNodeMap.size() && 
  //       "AllNodes and RegisterNumToNodesMap are not the same size");
  return NewNode;
}

///
/// addInterference -- add an interference between reg1 and reg2
///
void InterferenceGraph::addInterference(unsigned reg1, unsigned reg2) {
  
  // Convert to representative numbers
  unsigned RepReg1 = getRepresentativeNumber(reg1);
  unsigned RepReg2 = getRepresentativeNumber(reg2);
  
  /// If we want only update the Interference Matrix, uncomment this
  /// and the return uncommented below
  ///..    MatrixValidRows.insert(RepReg1);
  ///..    MatrixValidRows.insert(RepReg2);
  
  InterferenceGraphNode *reg1Node, *reg2Node;
    
  if ( getNodeOrNull(RepReg1) == getIGNullVal() ) {
    reg1Node = createInterferenceNode(RepReg1);
  }
  else {
    reg1Node = RegisterNumToNodeMap[RepReg1];
  }
  
  if ( getNodeOrNull(RepReg2) == getIGNullVal() ) {
    reg2Node = createInterferenceNode(RepReg2);
  }
  else {
    reg2Node = RegisterNumToNodeMap[RepReg2];
  }
  
  unsigned CoalescedIndex1 = reg1Node->index;
  unsigned CoalescedIndex2 = reg2Node->index;
  
  // Dont add self interferences
  // CHECK
  if (CoalescedIndex1 == CoalescedIndex2) {
    return;
  }
  InterferenceMatrix.set(getIdx(CoalescedIndex1, CoalescedIndex2));
  
  /// If we want only update the Interference Matrix, uncomment this next line: 
  /// return
  
  reg1Node->addInterference(reg2Node);
  reg2Node->addInterference(reg1Node);
  
}

///
/// Predicate: Check if two nodes interfere
///
bool InterferenceGraph::doesInterfere(unsigned reg1, unsigned reg2) {
  
  // Convert to representative numbers
  unsigned RepReg1 = getRepresentativeNumber(reg1);
  unsigned RepReg2 = getRepresentativeNumber(reg2);
  
  
  InterferenceGraphNode* reg1Node = RegisterNumToNodeMap[RepReg1];
  InterferenceGraphNode* reg2Node = RegisterNumToNodeMap[RepReg2];
  
  assert(reg1Node && reg2Node 
         && "Register numbers not found in map");
  
  unsigned idx = getIdx(reg1Node->index, reg2Node->index);
  return InterferenceMatrix.test(idx);
}

///
/// Iterate over a node's adjacency list
///
InterferenceGraphNodeIterator
InterferenceGraph::getNeighborsBegin(unsigned reg) {
  
  // Convert to representative numbers
  unsigned RepReg = getRepresentativeNumber(reg);
  
  InterferenceGraphNode* regNode = RegisterNumToNodeMap[RepReg];
  
  assert( regNode
         && "Register number not found in map");
  
  return regNode->getNeighborsBegin();
}


///
/// Iterate over a node's adjacency list
///
InterferenceGraphNodeIterator 
InterferenceGraph::getNeighborsEnd(unsigned reg) {
  
  // Convert to representative numbers
  unsigned RepReg = getRepresentativeNumber(reg);
  
  InterferenceGraphNode* regNode = RegisterNumToNodeMap[RepReg];
  
  assert(regNode
	       && "Register number not found in map");
  
  return regNode->getNeighborsEnd();
}


///
/// Iterate over the physical register nodes  
///
InterferenceGraphNodePhysRegIterator InterferenceGraph::getPhysRegBegin() {
  return PhysicalRegNodes.begin();
}

///
/// Iterate over the physical register nodes  
///
InterferenceGraphNodePhysRegIterator InterferenceGraph::getPhysRegEnd() {
  return PhysicalRegNodes.end();
}


///
/// Calculate spill costs
///
/// At the end of this, AllSortedNodes should consist of only virtual register
/// numbers that can potentially be spilt (i.e. it should not contain any 
/// register that has infinite spill cost). The nodes should be ordered in 
/// ascending order of spill preference -- the last node in AllSortedNodes 
/// will be the first to be spilled.
/// 
void InterferenceGraph::calcSpillCosts(const std::set<unsigned>& InfiniteSpillCostRegs) {
  
  std::set<unsigned> TempSet, DiffSet;
  
  // Take out the infinite spill registers -- ugly, but we dont have access to the
  // nodes while creating InfiniteSpillCostRegs, hence the conversion to register
  // numbers is required
  for (std::set<InterferenceGraphNode*>::iterator i = AllNodes.begin();
       i != AllNodes.end(); ++i) {
    if (TargetRegisterInfo::isVirtualRegister((*i)->index)) {
      TempSet.insert((*i)->index);
    }
  }
  
  set_difference( TempSet.begin(), TempSet.end(),
                 InfiniteSpillCostRegs.begin(), InfiniteSpillCostRegs.end(),
                 inserter(DiffSet, DiffSet.begin()));
  
  // Next, sort the nodes in ascending order of spill preference
  // Again, can we minimize the conversion from nodes -> register number and back?
  std::vector<InterferenceGraphNode*> sortVec;
  for (std::set<unsigned>::iterator ii = DiffSet.begin(); ii != DiffSet.end();
       ++ii) {
    // uncomment sortVec.push_back(getNode(*ii));
    AllSortedNodes.push_back(getNode(*ii));
  }
  
  
  if (SpillHeuristic == callahan) {
    std::sort(AllSortedNodes.begin(), AllSortedNodes.end(), CompareInterferenceGraphNodeCK());
  }
  else if (SpillHeuristic == chaitin) {
    std::sort(AllSortedNodes.begin(), AllSortedNodes.end(), CompareInterferenceGraphNodeChaitin());
  }
  else if (SpillHeuristic == chaitin_biasrefs) {
    std::sort(AllSortedNodes.begin(), AllSortedNodes.end(), CompareInterferenceGraphNodeChaitinBiasRefs());
  }
  else if (SpillHeuristic == refcount) {
    std::sort(AllSortedNodes.begin(), AllSortedNodes.end(), CompareInterferenceGraphNodeRefs());
  }
  else if (SpillHeuristic == pure_refcount) {
    std::sort(AllSortedNodes.begin(), AllSortedNodes.end(), CompareInterferenceGraphNodePureRefs());
  }
  else {
    assert (false && "No spill heuristic");
  }
  
  
  if ( PrintSpillCosts ) {
    dbgs() << "Node spill costs are:\n";
    if ( InfiniteSpillCostRegs.size() > 0 ) {
      for (std::set<unsigned>::iterator d = InfiniteSpillCostRegs.begin();
           d != InfiniteSpillCostRegs.end(); ++d) {
        dbgs() << PrintReg(*d) << ":  SpillCost = infinite.\n";
      }
    }
    if ( AllSortedNodes.size() > 0 ) {
      for (std::vector<InterferenceGraphNode*>::iterator d = AllSortedNodes.begin();
           d != AllSortedNodes.end(); ++d) {
        dbgs() << PrintReg((*d)->index) << ":  SpillCost = " << (*d)->Refs
        << "  Degree = " << (*d)->getDegree() 
        << "  TotalCost = " << (long long) (pow((*d)->Refs,1)/(*d)->getDegree() * 1000)
        << "\n";
      }
    }
    dbgs() << "\n";
  }
  
}

///
/// Returns a (potentially non-unique) virtual register node with 
/// the maximum degree in the graph
/// Used for spilling
/// Post: Removes the node from AllSortedNodes but NOT from AllNodes
InterferenceGraphNode* InterferenceGraph::
getVirtualRegisterNodeWithMostDegree(const std::set<unsigned>& InfiniteSpillCostRegs) {    
  
  if (AllSortedNodes.empty()) { 
    return 0;
  }
  
  InterferenceGraphNode* returnedNode = AllSortedNodes.back();
  
  AllSortedNodes.pop_back();
  
  return returnedNode;
}

///
/// removePRNodes -- removes all nodes containing physical registers
///
NodeSetType InterferenceGraph::removePRNodes() {
  
  NodeSetType TemporarilyRemovedNodes;
  
  std::set<InterferenceGraphNode*,  ComparePhysicalRegNodes> copyOfPhysicalRegNodes = PhysicalRegNodes;
  
  for (InterferenceGraphNodePhysRegIterator pIter = copyOfPhysicalRegNodes.begin();
       pIter != copyOfPhysicalRegNodes.end(); ++pIter) {
    
    removeNode(*pIter);
    TemporarilyRemovedNodes.insert(*pIter);
  }
  
  return TemporarilyRemovedNodes;
}

/// 
/// For a register class, returns the number of unique RPRs in that
/// class. For a register r belonging to class c, this is the 
/// quantity which determines if the corresponding IG node is 
/// trivially colorable
///
/*static*/ std::map<const TargetRegisterClass*, unsigned> InterferenceGraph::RegClass2KMap;
unsigned InterferenceGraph::getUniqueRPRsInRegClass(const TargetRegisterClass* TRC) {
  
  // The degree which determines whether a node is trivially colorable or not
  // is equal to the number of unique RPRs in its RegClass. Pre-calculate this
  
  if ( InterferenceGraph::RegClass2KMap.find(TRC) == InterferenceGraph::RegClass2KMap.end() || 
       InterferenceGraph::RegClass2KMap[TRC] == 0 ) {
    ArrayRef<unsigned> RawAlocationOrder = TRC->getRawAllocationOrder(*MF);
    ArrayRef<unsigned>::iterator RI = RawAlocationOrder.begin();
    ArrayRef<unsigned>::iterator RE = RawAlocationOrder.end();

    InterferenceGraph::RegClass2KMap[TRC] = 0;
    for (; RI != RE; ++RI) {
      ++InterferenceGraph::RegClass2KMap[TRC];
    }    
  }
  
  return InterferenceGraph::RegClass2KMap[TRC];
}

/// removeTriviallyColorableNodes -- removes nodes with degree less than or 
/// equal to the number of registers in their class as long as they are not marked
/// for ForcedSpilling (i.e. their ForceSpill flag is clear)
/// Returns a set of the removed nodes
///
///
/// Alternatively, for efficiency, we could use the strategy in the LLVM code in IGNode.h --
/// have a flag associated with each InterferenceGraphNode which is set on removal
/// The iterators and predicates have to then check for the flag and pretend
/// the node does not exist if the flag is set
///
unsigned
InterferenceGraph::removeTriviallyColorableNodes(MachineFunction *MF, 
                                                 std::set<unsigned> *InfiniteSpillCostRegisters,
                                                 std::stack<InterferenceGraphNode*> *ColoringStack)
{
  std::list<InterferenceGraphNode*> NodesToBeRemoved;
  unsigned K;
  InterferenceGraphNodeIterator AllNodesIter;
  InterferenceGraphNodeIterator AllNodesIterEnd;
  InterferenceGraphNodeIterator NeighborIter;
  InterferenceGraphNodeIterator NeighborIterEnd;
  InterferenceGraphNodeIterator CurrentNeighborIter;
  InterferenceGraphNodeIterator CurrentNeighborIterEnd;
  InterferenceGraphNode *CurrentNode;
  InterferenceGraphNode *CurrentNeighborNode;
  InterferenceGraphNode *NeighborNode;
  const TargetRegisterClass *RC;
  unsigned InitDegree;
  unsigned NumRemovedNeighbors;
  unsigned NumRemovedNode;
  unsigned NumPushedNodes;
  unsigned Degree;
  bool RemoveThisNode;
  bool IsInAllNodes;
  
  NodesToBeRemoved.clear();
  NumRemovedNode = 0;
  NumPushedNodes = 0;
  
  // move the trivially colored nodes from AllNodes to NodesToBeRemoved -- O(n)
  
  AllNodesIter    = AllNodes.begin();
  AllNodesIterEnd = AllNodes.end();
  while ( AllNodesIter != AllNodesIterEnd ) {
    CurrentNode = *AllNodesIter;

    RemoveThisNode = false;

    // Consider only virtual registers that are not pre-colored or that are not marked for spilling
    // Precoloring currently is never done

    if ( TargetRegisterInfo::isVirtualRegister(CurrentNode->index) &&
         CurrentNode->PreColorIndex == InterferenceGraphNode::NO_PRECOLOR_INDEX &&
         !CurrentNode->ForceSpill ) {
      RC = MF->getRegInfo().getRegClass(CurrentNode->index);
      K = Builder->getK(RC);

      InitDegree = CurrentNode->getDegree();
      NumRemovedNeighbors = 0;
      
      NeighborIter    = CurrentNode->getNeighborsBegin();
      NeighborIterEnd = CurrentNode->getNeighborsEnd();

      for (; NeighborIter != NeighborIterEnd; ++NeighborIter) {
        NeighborNode = *NeighborIter;
        if ( isRemovedNode(NeighborNode->index) ) {
          ++NumRemovedNeighbors;
        }
      }
      
      Degree = InitDegree - NumRemovedNeighbors;
      
      if ( Degree < K ) {
        RemoveThisNode = true;
      }
    }

    ++AllNodesIter;
    if ( RemoveThisNode ) {
      NodesToBeRemoved.push_front(CurrentNode);
      AllNodes.erase(prior(AllNodesIter));
      NumRemovedNode++;
      
      if ( DebugColoring ) {
        dbgs() << PrintReg(CurrentNode->index) << " is trivially colorable (primary effect)\n";
      }
    }
  }
  
  // move nodes from NodesToBeRemoved to ColoringStack, 
  // adding trivially colored nodes as we go.
  // O(n) for primary effect nodes and O(n**2) for secondary effect nodes???
  
  while ( ! NodesToBeRemoved.empty() ) {
    CurrentNode = NodesToBeRemoved.back(); NodesToBeRemoved.pop_back();
        
    // remove CurrentNode from the Interference Graph,
    // including edges

    removeNode(CurrentNode);
    
    // check to see if any of the virtual node neighbors are now colorable and
    // add the colorable neighbors to NodesToBeRemoved (secondary effect)
    
    InitDegree = CurrentNode->getDegree();
    NumRemovedNeighbors = 0;
    
    CurrentNeighborIter    = CurrentNode->getNeighborsBegin();
    CurrentNeighborIterEnd = CurrentNode->getNeighborsEnd();
    
    for (; CurrentNeighborIter != CurrentNeighborIterEnd; ++CurrentNeighborIter) {
      CurrentNeighborNode = *CurrentNeighborIter;
      
      AllNodesIter = AllNodes.find(CurrentNeighborNode); // regrettably O(log n)
      IsInAllNodes = ( AllNodesIter != AllNodes.end() );
      
      if ( TargetRegisterInfo::isVirtualRegister(CurrentNeighborNode->index) && IsInAllNodes ) {
        RC = MF->getRegInfo().getRegClass(CurrentNeighborNode->index);
        K = Builder->getK(RC);
        
        InitDegree = CurrentNeighborNode->getDegree();
        NumRemovedNeighbors = 0;
        
        NeighborIter    = CurrentNeighborNode->getNeighborsBegin();
        NeighborIterEnd = CurrentNeighborNode->getNeighborsEnd();
        
        for (; NeighborIter != NeighborIterEnd; ++NeighborIter) {
          NeighborNode = *NeighborIter;
          if ( isRemovedNode(NeighborNode->index) ) {
            ++NumRemovedNeighbors;
          }
        }
        
        Degree = InitDegree - NumRemovedNeighbors;
        
        if ( Degree < K ) {
          AllNodes.erase(AllNodes.find(CurrentNeighborNode)); 
          NodesToBeRemoved.push_front(CurrentNeighborNode);
          NumRemovedNode++;
          
          if ( DebugColoring ) {
            dbgs() << PrintReg(CurrentNeighborNode->index) << " is trivially colorable (secondary effect)\n";
          }
        }
      }
    }

    // push node onto the Coloring Stack
    
    ColoringStack->push(CurrentNode);
    NumPushedNodes++;

  }
  
  assert((NumRemovedNode == NumPushedNodes) && 
         "The number of nodes removed from IG does not match "
         "the number of nodes pushed onto the coloring stack.");

  return NumPushedNodes;
}


//
// Practically identical to removeTriviallyColorableNodes()
// TODO: should collapse into one method
//
std::set<InterferenceGraphNode*, CompareIGNodes_PartnerNodes> 
InterferenceGraph::getTriviallyColorableNodes(MachineFunction* MF) {
  
  std::set<InterferenceGraphNode*, CompareIGNodes_PartnerNodes> TemporarilyRemovedNodes;
  // Cannot iterate over AllNodes, as it will be modified in the call to removeNode()
  // in the loop -- hence iterate over a copy of AllNodes
  std::set<InterferenceGraphNode*, CompareIGNodes_PartnerNodes> copyOfAllNodes;

  copyOfAllNodes.insert(AllNodes.begin(), AllNodes.end());

  for ( std::set<InterferenceGraphNode*, CompareIGNodes_PartnerNodes>::iterator DegreeIter = 
        copyOfAllNodes.begin(); 
        DegreeIter != copyOfAllNodes.end(); ++DegreeIter) {

    InterferenceGraphNode* currentNode = *DegreeIter;

    // Consider only virtual registers
    if (TargetRegisterInfo::isVirtualRegister(currentNode->index) &&
        currentNode->PreColorIndex == InterferenceGraphNode::NO_PRECOLOR_INDEX &&
        !currentNode->ForceSpill) {

      // Check if the node has less than k neighbors
      unsigned k = Builder->getK(MF->getRegInfo().getRegClass(currentNode->index));

      // Calculate the degree of a node, but ensure that removed neighbors are not counted
      unsigned InitDegree = currentNode->getDegree();
      unsigned RemovedNeighbors = 0;
      for (InterferenceGraphNodeIterator NI = currentNode->getNeighborsBegin(); 
           NI != currentNode->getNeighborsEnd(); ++NI) {
        InterferenceGraphNode *NeighborNode = *NI;
        if (isRemovedNode(NeighborNode->index)) {
          ++RemovedNeighbors;
        }
      }

      unsigned Degree = InitDegree - RemovedNeighbors;

      if (Degree < k) {

        // Insert into returned set
        TemporarilyRemovedNodes.insert(currentNode);
      } // (degree <= k)
    }
  }
  return TemporarilyRemovedNodes;
}

unsigned InterferenceGraph::numberOfTriviallyColorableNodes() {
  
  unsigned numTrivNodes = 0;
  
  for ( std::set<InterferenceGraphNode*>::iterator DegreeIter = 
       AllNodes.begin(); 
       DegreeIter != AllNodes.end(); ++DegreeIter) {
    
    InterferenceGraphNode* currentNode = *DegreeIter;
    // Consider only virtual registers
    if (TargetRegisterInfo::isVirtualRegister(currentNode->index)) {
      
      // Check if the node has less than k neighbors
      unsigned k = Builder->getK(MF->getRegInfo().getRegClass(currentNode->index));
      
      if (currentNode->getDegree() < k) {  
        ++numTrivNodes;
      }
    }
  }
  
  return numTrivNodes;
}


///
/// Predicate: Checks if a node has been removed from graph via removeNode()
///
bool InterferenceGraph::isRemovedNode(unsigned nodeIndex) {
  return (RemovedNodes[getDenseIndex(nodeIndex)]);
}

///
/// removeNode -- remove a node from the graph
///
void InterferenceGraph::removeNode(InterferenceGraphNode* removedNode) {
  
  
  /// InterferenceMatrix is lazily updated because one
  /// should not be looking at matrix entries for 
  /// removed nodes at any rate
  
  RemovedNodes[getDenseIndex(removedNode->index)] = true;
  
  
  // Remove this node's entry from the interference matrix
  //    InterferenceMatrix.erase(removedNode->index);
  
  // Remove from RegisterNumToNodeMap
  RegisterNumToNodeMap.remove(removedNode->index);
  
  // Remove node from AllNodes
  AllNodes.erase(removedNode);
  
  // They can be of different sizes due to coalescing
  //    assert(SortedNodes.size() == RegisterNumToNodeMap.size() && 
  //	   "SortedNodes and RegisterNumToNodesMap are not the same size");
  
  // Remove node from PhysicalRegNodes
  if ( TargetRegisterInfo::isPhysicalRegister(removedNode->index) ) {
    PhysicalRegNodes.erase(removedNode);
  }
}

/// Adds a node temporarily removed from the graph back to the graph Special
/// care is taken while adding the node since some of its erstwhile neighbors
/// could have been removed.  Updates InterferenceMatrix O(n) times in the
/// worst case
///
/// See comments in removeTriviallyColorableNodes for possible more efficient
/// implementation
///
/// Pre: The node's adjacency list is intact
///
void InterferenceGraph::addNodeBackToGraph(InterferenceGraphNode* addedNode ) {
  
  RemovedNodes[getDenseIndex(addedNode->index)] = false;
  
  // Add to RegisterNumToNodeMap
  RegisterNumToNodeMap[addedNode->index] = addedNode;
  
  // Add to AllNodes
  AllNodes.insert(addedNode);
  
  // Add node to PhysicalRegNodes
  if ( TargetRegisterInfo::isPhysicalRegister(addedNode->index) ) {
    PhysicalRegNodes.insert(addedNode);
  }
  
}


///
/// isAPhysicalReg -- predicate to check if it's a physical register
///
bool InterferenceGraph::isAPhysicalReg(InterferenceGraphNode* node) {
  return (PhysicalRegNodes.find(node) != PhysicalRegNodes.end());
}


/// 
/// Predicate to check whether the 2 nodes, if coalesced, will spill
///
bool InterferenceGraph::willSpillCoalescedNodes(unsigned nodeIndex1, unsigned nodeIndex2) {
  
  InterferenceGraphNode* node1 = RegisterNumToNodeMap[nodeIndex1];
  InterferenceGraphNode* node2 = RegisterNumToNodeMap[nodeIndex2];
  
  
  assert(TargetRegisterInfo::isVirtualRegister(nodeIndex1) &&
         TargetRegisterInfo::isVirtualRegister(nodeIndex2) &&
         "Can only coalesce virtual registers");
  
  assert(node1 && node2 && 
         "Cannot find a node in the graph");
  
  std::set<InterferenceGraphNode*> combinedInterfs;
  combinedInterfs.insert(node1->getNeighborsBegin(), node1->getNeighborsEnd());
  
  unsigned n1 = combinedInterfs.size();
  
  combinedInterfs.insert(node2->getNeighborsBegin(), node2->getNeighborsEnd());
  
  unsigned n2 = combinedInterfs.size() - n1;
  
  // Count the number of nodes that have not been removed

  unsigned totalInterfs = 0;
  for (std::set<InterferenceGraphNode*>::iterator i = combinedInterfs.begin(); i != combinedInterfs.end();
       ++i) {
    InterferenceGraphNode *Node = *i;

    if ( ! RemovedNodes[getDenseIndex(Node->index)] ) {
      ++totalInterfs;
    }
  }
  
  // Both nodes have the same register class
  // Check if the node has less than k neighbors
  unsigned k;
  if ( TargetRegisterInfo::isPhysicalRegister(node1->index) ) 
    k = Builder->getK(getRegClass_(MF->getTarget().getRegisterInfo(), node1->index));
  else 
    k = Builder->getK(MF->getRegInfo().getRegClass(node1->index));
  
  if ( DebugCoalesce ) {
    dbgs() << "Node1: (" << node1->index << "," << PrintReg(node1->index) << ") :" << n1
    << "\t " << "Node2: (" << node2->index << "," << PrintReg(node2->index) << ") :" << n2
    << "\t " << "total interf = " << totalInterfs 
    << "\t " << "k = " << k
    << "\n" << "willSpillCoalescedNodes() returns : " << bool(totalInterfs >= k)
    << "\n";
  }
  
  return ( totalInterfs >= k );  
}

void InterferenceGraph::coalesceNodes(unsigned nodeIndex1, unsigned nodeIndex2) {
  
  assert(TargetRegisterInfo::isVirtualRegister(nodeIndex1) &&
         TargetRegisterInfo::isVirtualRegister(nodeIndex2) &&
         "Can only coalesce virtual registers");
  
  InterferenceGraphNode* node1 = RegisterNumToNodeMap[nodeIndex1];
  InterferenceGraphNode* node2 = RegisterNumToNodeMap[nodeIndex2];
  
  // Update the node indices to reflect the fact that the nodes might have 
  // been previously coalesced
  nodeIndex1 = node1->index;
  nodeIndex2 = node2->index;

  // If they are the same register or have been coalesced to the same register, then done
  if (nodeIndex1 == nodeIndex2) {
    return;
  }
  
  assert( !doesInterfere(nodeIndex1, nodeIndex2) && 
         "Cannot coalesce nodes that interfere");
  
  // Delete node2 from AllNodes
  AllNodes.erase(node2);
  
  unsigned RepRegNode2 = getRepresentativeNumber(nodeIndex2);
  
  //    RegisterNumToNodeMap.erase(RepRegNode2);
  
  // Insert node1 in node2's place in RegisterNumToNodeMap
  RegisterNumToNodeMap[RepRegNode2] = node1;
  
  // Change over the coalesced nodes to point to node1
  // i.e. path compression
  for (CoalescedNodesIterator ci = node2->getCoalescedNodesBegin(); ci != node2->getCoalescedNodesEnd();
       ++ci) {
    RegisterNumToNodeMap[*ci] = node1;
  }
  
  
  // Change the interference matrix for the new coalesced node
  ///    InterferenceMatrix[node1->index].insert(InterferenceMatrix[node2->index].begin(), InterferenceMatrix[node2->index].end());
  
  // Iterate over the second node and add all members of its 
  // adjacency list to the first node
  for (InterferenceGraphNodeIterator ai = node2->getNeighborsBegin(); ai != node2->getNeighborsEnd();
       ++ai) { 
    InterferenceGraphNode *NeighborNode = *ai;
    
    
    // TODO: could do this in a insertion via iterators: faster?
    node1->addInterference(NeighborNode);
    
    
    // Remove node2 from adjacency list of its neigbors and
    // add node1 to the adjacency list of node2's neighbors
    
    assert((NeighborNode->getES().find(node2) != NeighborNode->getES().end()) && 
           "inconsistent nodes");
    
    NeighborNode->removeInterference(node2);
    NeighborNode->addInterference(node1);
    
    
    // Update Interference Matrix
    InterferenceMatrix.set(getIdx(node1->index, NeighborNode->index));
    InterferenceMatrix.reset(getIdx(node2->index, NeighborNode->index));

  }
  
  node1->addCoalescedNode(node2->index);
  
  // Coalesce nodes that have been coalesced into node2 into node1
  // and update registernumtonodemap
  for (CoalescedNodesIterator coIter = node2->getCoalescedNodesBegin(); 
       coIter != node2->getCoalescedNodesEnd(); 
       ++coIter) {
    node1->addCoalescedNode(*coIter);
    RegisterNumToNodeMap[*coIter] = node1;
  }

  // Update SingleDef field. If node1 was not SingleDef, merging the copy will
  // not make it SingleDef.
  MachineInstr *Node1SingleDef = node1->getSingleDef();
  MachineInstr *Node2SingleDef = node2->getSingleDef();
  if (!(Node1SingleDef && Node2SingleDef && Node2SingleDef->isCopy() &&
        Node2SingleDef->getOperand(0).getSubReg() ==
        Node2SingleDef->getOperand(1).getSubReg()))
    node1->setSingleDef(NULL);


  // node1 inherits node2 biases ...
  for (InterferenceGraphNodeIterator PI = node2->getPartnerNodesBegin(); PI != node2->getPartnerNodesEnd(); 
       ++PI) {
    InterferenceGraphNode *PartnerNode = *PI;
    node1->addPartnerNode(PartnerNode);
    
    // fix the other direction.  partner must now reference node1 instead of node2
    PartnerNode->replacePartner(node1, node2);
  }
  
  // node2 does not exist in the graph...
  //    InterferenceMatrix.erase(node2->index);
  // UNCOMMENT...
  delete node2; node2 = 0;
}

///
/// dump() -- print the interference graph
///
void InterferenceGraph::dump(const TargetRegisterInfo *RegInfo,
                             MachineFunction * MF)
{
  dbgs() << "\nDumping Interference Graph\n";
  
  std::set<InterferenceGraphNode*>::iterator miter;
  for (miter = AllNodes.begin(); miter != AllNodes.end(); ++miter) {
    InterferenceGraphNode *IGNode = *miter;
    
    IGNode->dump(RegInfo, MF, this);
  }

  dbgs() << "\n";
  
}

void InterferenceGraph::dumpSN() {
  std::set<InterferenceGraphNode*>::iterator i;
  for (i = AllNodes.begin(); i != AllNodes.end(); ++i) {
    dbgs() << (*i)->index << ", ";      
  }
  dbgs() << "\n";
}


// this should eventually go into InterferenceGraphNode.cpp
// which currently does not exist.  sue me.  i'm lazy.

///
/// dump() -- print the interference graph node 
///
void InterferenceGraphNode::dump(const TargetRegisterInfo *RegInfo,
                                 MachineFunction * MF, 
                                 InterferenceGraph *IG, 
                                 bool deep /*= false*/)
{
  unsigned NumRemovedNeighbors;
  unsigned K;
  const TargetRegisterClass *RC;
  
  if ( TargetRegisterInfo::isPhysicalRegister(index) ) {
    dbgs() << PrintReg(index) << "(" << PrintReg(index, RegInfo) << ") : \n";
  }
  else {
    const TargetRegisterClass* regClass = MF->getRegInfo().getRegClass(index);
    dbgs() << PrintReg(index) << "(" << regClass->getName() << ") : \n";
  }
  
  if ( TargetRegisterInfo::isPhysicalRegister(index) ) 
    RC = getRegClass_(RegInfo, index);
  else
    RC = MF->getRegInfo().getRegClass(index);
  K = IG->Builder->getK(RC);

  NumRemovedNeighbors = 0;
  for (InterferenceGraphNodeIterator siter = getNeighborsBegin(); 
       siter != getNeighborsEnd(); ++siter) {
    InterferenceGraphNode *NeighborNode = *siter;
    unsigned NeighborReg = NeighborNode->index;
    
    if ( IG->isRemovedNode(NeighborReg) ) {
      NumRemovedNeighbors++;
    }
  }

  dbgs() << "\tK is " << K << " (" << RC->getName() << ")\n";
  dbgs() << "\tNumber of neighbors is " << getDegree() << "\n";
  dbgs() << "\tNumber of removed neighbors is " << NumRemovedNeighbors << "\n";
  dbgs() << "\tCurrent degree is " << int(getDegree() - NumRemovedNeighbors) << "\n";
  dbgs() << "\tTrivially colorable? ";
  if ( getDegree() - NumRemovedNeighbors < K ) dbgs() << "true\n";
  else                                         dbgs() << "false\n";
  dbgs() << "\tColor is " << Color << "\n";
  dbgs() << "\tNeighbors : ";

  for (InterferenceGraphNodeIterator siter = getNeighborsBegin(); 
       siter != getNeighborsEnd(); ++siter) {
    InterferenceGraphNode *NeighborNode = *siter;
    unsigned NeighborReg = NeighborNode->index;
    
    if ( TargetRegisterInfo::isPhysicalRegister(NeighborReg) ) {
      // It's a physical register
      dbgs() << PrintReg(NeighborReg) << "(" << PrintReg(NeighborReg, RegInfo) << "), ";
    }
    else {
      // It's a virtual register
      dbgs() << PrintReg(NeighborReg) << ", ";
    }
  }
  dbgs() << "\n";

  dbgs() << "\t" << getNumPartners() << " Partners : ";

  for (InterferenceGraphNodeIterator siter = getPartnerNodesBegin(); 
       siter != getPartnerNodesEnd(); ++siter) {
    InterferenceGraphNode *PartnerNode = *siter;
    unsigned PartnerReg = PartnerNode->index;
    
    if ( TargetRegisterInfo::isPhysicalRegister(PartnerReg) ) {
      // It's a physical register
      dbgs() << PrintReg(PartnerReg) << "(" << PrintReg(PartnerReg, RegInfo) << "), ";
    }
    else {
      // It's a virtual register
      dbgs() << PrintReg(PartnerReg) << ", ";
    }
  }
  dbgs() << "\n";
  
  if ( deep ) {
    // not removed virtual regs
    for (InterferenceGraphNodeIterator siter = getNeighborsBegin(); 
         siter != getNeighborsEnd(); ++siter) {
      InterferenceGraphNode *NeighborNode = *siter;
      unsigned NeighborReg = NeighborNode->index;
      int NeighborColor = NeighborNode->Color;
      
      if ( TargetRegisterInfo::isPhysicalRegister(NeighborReg) ) continue;
      if ( IG->isRemovedNode(NeighborReg) ) continue;
      
      dbgs() << "\tNeighbor node " << PrintReg(NeighborReg);
      dbgs() << " has color " << NeighborColor << "\n";
    }
    // removed virtual regs
    for (InterferenceGraphNodeIterator siter = getNeighborsBegin(); 
         siter != getNeighborsEnd(); ++siter) {
      InterferenceGraphNode *NeighborNode = *siter;
      unsigned NeighborReg = NeighborNode->index;
      
      if ( IG->isRemovedNode(NeighborReg) ) {
        dbgs() << "\tNeighbor node removed(" << PrintReg(NeighborReg) << ")\n";
      }
    }
    // physical regs
    for (InterferenceGraphNodeIterator siter = getNeighborsBegin(); 
         siter != getNeighborsEnd(); ++siter) {
      InterferenceGraphNode *NeighborNode = *siter;
      unsigned NeighborReg = NeighborNode->index;
      
      if ( TargetRegisterInfo::isPhysicalRegister(NeighborReg) ) {
        dbgs() << "\tNeighbor node " << PrintReg(NeighborReg) 
        << "(" << PrintReg(NeighborReg, RegInfo) << ")\n";
      }
    }
  }

  dbgs() << "\n";
}


