//===--InterferenceGraph.h   - Defines an interference graph         -------===//
// Defines an interference graph
// Anshuman Dasgupta, MSCP, Rice University
// 
//===----------------------------------------------------------------------===//
//===----------------------------------------------------------------------===//
//This software was produced with support from the Defense Advanced
//Research Projects Agency (DARPA) through AFRL Contract FA8650-09-C-1915. 
//Nothing in this work should be construed as reflecting the official policy or
//position of the Defense Department, the United States government, or
//Rice University.
//
//THIS SOFTWARE HAS BEEN APPROVED FOR PUBLIC RELEASE, UNLIMITED DISTRIBUTION.
//
//===----------------------------------------------------------------------===//

#ifndef _INTERFERENCE_GRAPH_H_
#define _INTERFERENCE_GRAPH_H_

#include "InterferenceGraphNode.h"

#include "Shared.h" // from the MSCP ILOC compiler

#include "llvm/ADT/BitVector.h"
#include "llvm/Target/TargetRegisterInfo.h"

#include "llvm/CodeGen/MachineFunction.h"

#include "GrowableDenseMap.h"

#include "Common.h"


#include <set>
#include <vector>
#include <map>
#include <stack>

///
/// The nodes in this graph are created thus:
///    For physical registers: There is one node per set of aliased registers (one for AX, AL, and EAX)
///    For virtual registers: one node per virtual register
///
namespace llvm {
  
  typedef std::map<unsigned, std::set<unsigned> >::iterator RepresentativeNumbersIterator;
  
  typedef std::set<InterferenceGraphNode*> NodeSetType;
  typedef std::set<InterferenceGraphNode*>::iterator NodeSetTypeIterator;
  typedef std::set<unsigned> RegisterSet;
  
  enum SpillHeuristicName { chaitin, refcount, pure_refcount, callahan, chaitin_biasrefs };
  
  struct Reg2DenseIndexFunctor : std::unary_function<unsigned, unsigned> {
    unsigned operator()(unsigned RepReg) const;
  };
  
  class BuildInterferenceGraph;

  //////////////////////////////////////////////////////////////////////////////
  /// The interference graph class used by BuildInterferenceGraph
  /// An effort has been made to make this relatively architecture-
  /// independent
  /// The two major data structures used to represent the graph 
  /// are an InterferenceMatrix structure and a set of nodes.
  /// 
  ///
  /// The graph uses the concept of a Representative Physical Registers (RPRs).
  /// RPRs represent physical registers that are aliased to each other (partial 
  /// or full aliases). There is one Interference Graph node per RPR
  ///
  /// Each node in this graph either represents a virtual register or a RPR. 
  /// Nodes are identified by their virtual register number or their RPR number
  //////////////////////////////////////////////////////////////////////////////
  class InterferenceGraph {
    
  private:
    
    const MachineFunction* MF;
    
    const TargetRegisterInfo* RegInfo;
    
    typedef std::set<unsigned> UnsignedSetType;
    
    unsigned MaxSizeOfIG;
    
    /// Maps from TargetRegisterClass* -> unsigned inde
    /// Passed down from BuildInterference Graph
    std::map<const TargetRegisterClass*, unsigned> TRC2IntMap;
    
    // Map from representative physical register number to all registers in the 
    // RPR set. Passed down from BuildInterference Graph
    std::map<unsigned, std::set<unsigned> > RepresentativeNum2RegistersSet;
    
    
    /// Holds a map from Register Representative numbers -> Interference Graph Nodes
    GrowableDenseMap<InterferenceGraphNode*, Reg2DenseIndexFunctor> RegisterNumToNodeMap;
    
    // Representation of graph
    // Is symmetric (can optimize)
    // Change -- InterferenceMatrix lookup whould be ~linear
    BitVector InterferenceMatrix;

    // Rows of the interference matrix that are valid
    std::set<unsigned> MatrixValidRows;
    
    /// All nodes 
    std::set<InterferenceGraphNode*> AllNodes;
    
    std::vector<InterferenceGraphNode*> AllSortedNodes;
    
    /// 
    /// Set of physical register nodes in the graph
    std::set<InterferenceGraphNode*,  ComparePhysicalRegNodes> PhysicalRegNodes;
    
    ///
    /// Set of register no longer in the graph
    BitVector RemovedNodes;
    
    void buildRegisterRepresentativeSet();
    void buildTRCToIntMap();
    
    
    ///
    /// Predicate that returns true if two nodes can occupy common registers
    /// register set -- for instance integer and FP valued nodes cannot 
    /// share the same register
    ///
    /// For the x86, Register classes are
    ///  Register class: 0
    /// AX,CX,DX,BX,SI,DI,BP,SP,
    ///  Register class: 1
    /// EAX,ECX,EDX,EBX,ESI,EDI,EBP,ESP,
    ///   Register class: 2
    /// AL,CL,DL,BL,AH,CH,DH,BH,
    ///  Register class: 3
    /// FP0,FP1,FP2,FP3,FP4,FP5,FP6,
    ///   Register class: 4
    /// ST(0),ST(1),ST(2),ST(3),ST(4),ST(5),ST(6),ST(7),
    ///
    /// So, classes 0, 1, and 2 can share a physical register
    ///
    bool canSharePhysicalRegister(unsigned reg1, unsigned reg2);
    
  public:
    
    static unsigned MaxPhysicalRegisterNum;
    static unsigned MaxVirtualRegisterNum;
    static unsigned FirstDensePhysicalRegisterNum;
    static unsigned FirstDenseVirtualRegisterNum;
    
    static unsigned SimplifyIterations;
    
    BuildInterferenceGraph *Builder;
    
    ///
    /// Many of these values should be calculated in BuildInterferenceGraph 
    /// and passed into the ctor 
    ///
    InterferenceGraph(const MachineFunction*, 
                      const TargetRegisterInfo*,
                      BuildInterferenceGraph*,
                      unsigned, 
                      unsigned);
    
    ~InterferenceGraph();
    
    
    void GrowMatrix(unsigned);
    
    ///
    /// Checks if the Interference Graph is empty
    ///
    bool isEmpty() { return AllNodes.empty(); }
    
    ///
    /// Total size of the Interference Graph
    ///
    unsigned size() { return AllNodes.size(); }
    
    
    ///
    /// Add a node to the graph only if it does not already exist
    ///
    InterferenceGraphNode* createInterferenceNode(unsigned reg, MachineOperand* moper = NULL);
    
    ///
    /// Predicate: Check if two nodes interfere
    ///
    bool doesInterfere(unsigned reg1, unsigned reg2);
    
    
    /// 
    /// For a register class, returns the number of unique RPRs in that
    /// class. For a register r belonging to class c, this is the 
    /// quantity which determines if the corresponding IG node is 
    /// trivially colorable
    ///
    static std::map<const TargetRegisterClass*, unsigned> RegClass2KMap;
    unsigned getUniqueRPRsInRegClass(const TargetRegisterClass*);
    
    
    ///
    /// Iterate over a node's adjacency list
    ///
    InterferenceGraphNodeIterator getNeighborsBegin(unsigned reg);
    InterferenceGraphNodeIterator getNeighborsEnd(unsigned reg);
    
    ///
    /// Iterate over the physical register nodes in the graph
    ///
    InterferenceGraphNodePhysRegIterator getPhysRegBegin();
    InterferenceGraphNodePhysRegIterator getPhysRegEnd();
    
    void addInterference(unsigned reg1, unsigned reg2);
    

    ///
    /// removePRNodes -- removes all nodes containing physical registers
    ///
    NodeSetType removePRNodes();
    
    
    
    ///  removeTriviallyColorableNodes -- removes nodes with degree less than or 
    /// equal to specified number
    /// Returns a set of the removed nodes
    unsigned removeTriviallyColorableNodes(MachineFunction *MF, 
                                           std::set<unsigned> *InfiniteSpillCostRegisters,
                                           std::stack<InterferenceGraphNode*> *ColoringStack);
    
    ///  getTriviallyColorableNodes -- returns nodes with degree less than or 
    /// equal to specified number
    /// Practically identical to removeTriviallyColorableNodes() TODO: should collapse into one method
    ///
    std::set<InterferenceGraphNode*, CompareIGNodes_PartnerNodes> getTriviallyColorableNodes(MachineFunction* );
    
    
    ///
    /// Get number of trivially colorable nodes
    ///
    unsigned numberOfTriviallyColorableNodes();
     
    
    ///
    /// Predicate: Checks if a node has been removed from graph via removeNode()
    ///
    bool isRemovedNode(unsigned);
    
    
    ///
    /// removeNode -- remove a node from the graph
    ///
    void removeNode(InterferenceGraphNode*);
    
    
    /// addNodeBackToGraph -- adds a node temporarily removed from the graph back to the graph
    /// Special care is taken while adding the node since some of its erstwhile neighbors could 
    /// have been removed.
    /// Updates InterferenceMatrix O(n) times in the worst case
    void addNodeBackToGraph(InterferenceGraphNode* );
    
    
    /// isAPhysicalReg -- predicate to check if it's a physical register
    bool isAPhysicalReg(InterferenceGraphNode* );
    
    /// 
    /// Predicate to check whether the 2 nodes, if coalesced, will spill
    ///
    bool willSpillCoalescedNodes(unsigned, unsigned);
    
    ///
    /// coalesceNodes -- coalesces two nodes if they do not interfere
    ///
    void coalesceNodes(unsigned, unsigned);
    
    ///
    /// Iterate over all graph nodes
    ///
    InterferenceGraphNodeIterator begin() { return AllNodes.begin();}
    InterferenceGraphNodeIterator end() { return AllNodes.end();}
    
    void calcSpillCosts(const std::set<unsigned>& InfiniteSpillCostRegs);
    
    ///
    /// Returns a (potentially non-unique) virtual register node with 
    /// the maximum degree in the graph
    /// Used for spilling
    ///
    InterferenceGraphNode* getVirtualRegisterNodeWithMostDegree(const std::set<unsigned>& );
    
    ///
    /// Print the graph
    ///
    void dump(const TargetRegisterInfo *TRI, MachineFunction *MF);
    
    ///
    /// Returns the node for the specified register if it exists
    /// Else, returns NULL
    ///
    InterferenceGraphNode* getNodeIfInGraph(unsigned i) { 
      unsigned RepReg;
      RepReg = getRepresentativeNumber(i);
      if (getNodeOrNull(RepReg) == getIGNullVal()) {
        return NULL;
      }
      // check if it's been coalesced, if so return NULL
      else if ( RegisterNumToNodeMap[RepReg]->index != RepReg) {
        return NULL;
      }
      else {
        return RegisterNumToNodeMap[RepReg];
      }
    }
    
    InterferenceGraphNode* getNode(unsigned i) {
      unsigned RepReg;
      RepReg = getRepresentativeNumber(i);
      assert(RegisterNumToNodeMap[RepReg]); 
      return RegisterNumToNodeMap[RepReg];
    }
    
    InterferenceGraphNode* getIGNullVal() {
      return RegisterNumToNodeMap.nullVal_;
    }
    
    InterferenceGraphNode* getNodeOrNull(unsigned i) {
      unsigned RepReg, DenseReg;
      RepReg = getRepresentativeNumber(i);
      DenseReg = getDenseIndex(RepReg);
      if ( DenseReg < RegisterNumToNodeMap.size() ) {
        return RegisterNumToNodeMap[RepReg];
      }
      else {
        return getIGNullVal();
      }
    }
    
    static unsigned getDenseIndex(unsigned RepReg);

    unsigned getRepresentativeNumber(unsigned Reg);
    
    // Compute the index into the bit matrix for entry (i,j).
    unsigned int getIdx(unsigned ii, unsigned jj) {
      unsigned i = InterferenceGraph::getDenseIndex(ii);
      unsigned j = InterferenceGraph::getDenseIndex(jj);
      unsigned int x = i > j ? i : j;
      unsigned int y = i < j ? i : j;
      return ((x * x - x) / 2) + y;
    }
    
    // Compute the index into the containment bit matrix for entry (i,j).
    unsigned int getCMIdx(unsigned ii, unsigned jj) {
      unsigned i = InterferenceGraph::getDenseIndex(ii);
      unsigned j = InterferenceGraph::getDenseIndex(jj);
      return i * i  + j;
    }
    
    void dumpSN();
    
    
  };
  
  
}

#endif
