//===--InterferenceGraphNode.h   - Defines an interference graph node-------===//
// Defines an interference graph node
// 
//===----------------------------------------------------------------------===//
//===----------------------------------------------------------------------===//
//This software was produced with support from the Defense Advanced
//Research Projects Agency (DARPA) through AFRL Contract FA8650-09-C-1915. 
//Nothing in this work should be construed as reflecting the official policy or
//position of the Defense Department, the United States government, or
//Rice University.
//
//THIS SOFTWARE HAS BEEN APPROVED FOR PUBLIC RELEASE, UNLIMITED DISTRIBUTION.
//
//===----------------------------------------------------------------------===//

#ifndef _INTERFERENCEGRAPHNODE_H_
#define _INTERFERENCEGRAPHNODE_H_

#include <set>
#include "llvm/CodeGen/MachineInstr.h"
#include "llvm/ADT/PointerIntPair.h"
#include "llvm/Support/raw_ostream.h"

namespace llvm {

class InterferenceGraph;

class InterferenceGraphNode;

typedef std::set<InterferenceGraphNode*>::iterator InterferenceGraphNodeIterator;
//typedef std::set<unsigned>::iterator InterferenceGraphNodeIndexIterator;

typedef std::set<unsigned>::iterator CoalescedNodesIterator;
  
class InterferenceGraphNode {


 private:
  
  /// Adjacency list
  std::set<InterferenceGraphNode*> EdgeSet;
  
  MachineOperand* regOperand;


  // Indices of nodes coalesced into this one
  std::set<unsigned> coalescedNodes;

  // Nodes that are connected by copies to this node
  std::set<InterferenceGraphNode*> partnerNodes;

  // Contains the single definition for this live range or NULL.
  // If the bool associated with SingleDef is true, than no def has been
  // encountered yet.
  PointerIntPair<MachineInstr *, 1, bool> SingleDef;
 public:

  static const int NO_COLOR = -1;
  static const unsigned NO_PRECOLOR_INDEX = ~0;

  // If this flag is set, then colorNode() will not color the node and 
  // thus force it to be spilled
  bool ForceSpill;
  
  // Removed from the IG and placed on the Color stack, optimitically hoping
  // that the node can be colored and not spilled.
  bool OptimisticallyPushed;

  // index contains the representative register number 
  // which is equal to the virtual reg. number if the 
  // register is a VR 
  // If it is a PR, then the reg. number is equal to the 
  // representative class number
  unsigned index;

  // Number of references -- used in Spill cost calculation
  double Refs;
  // Number of rematerializations.
  double Remats;

  // Split cost for passive splitting.
  double SplitCost;

  int Color;

  // Number of times this value is used
  int NumUses;

  // This is used to pre-color VRs.  It must be a physical RPR.
  unsigned PreColorIndex;

  // CK Weight -- holds CK spill cost
  float CKWeight;

  // ctor
  InterferenceGraphNode(unsigned i): regOperand(NULL), SingleDef(NULL, true),
				     ForceSpill(false), OptimisticallyPushed(false),
				     index(i), Refs(0), Remats(0), SplitCost(0), 
				     Color(NO_COLOR), NumUses(0), 
				     PreColorIndex(NO_PRECOLOR_INDEX),
             CKWeight(-55555.0) {
  }

  /// Returns true if not definition has been encountered yet.
  bool isSingleDefUninitialized() {
    return SingleDef.getInt();
  }

  MachineInstr *getSingleDef() {
    return SingleDef.getPointer();
  }

  void setSingleDef(MachineInstr *I) {
    SingleDef.setInt(false);
    SingleDef.setPointer(I);
  }

  void setRegOperand(MachineOperand* m) {
    regOperand = m;
  }

  inline MachineOperand* getRegOperand() {
    assert( (regOperand != NULL) && "machine Operand not set for this register");
    return regOperand;
  }
    
  inline unsigned getDegree() { return EdgeSet.size(); }
  
  inline void removeInterference(InterferenceGraphNode *node) { EdgeSet.erase(node); }
  inline void addInterference(InterferenceGraphNode *node) { 
    assert(node->index != index && "No Self interferences allowed");
    EdgeSet.insert(node);}
  inline InterferenceGraphNodeIterator getNeighborsBegin() { return EdgeSet.begin();}
  inline InterferenceGraphNodeIterator getNeighborsEnd()   { return EdgeSet.end();}
  inline unsigned getNumNeighbors() { return EdgeSet.size(); };

  inline void addCoalescedNode(unsigned n) { coalescedNodes.insert(n); }
  inline CoalescedNodesIterator getCoalescedNodesBegin() { return coalescedNodes.begin(); }
  inline CoalescedNodesIterator getCoalescedNodesEnd()   { return coalescedNodes.end(); }
  inline unsigned getNumCoalesced() { return coalescedNodes.size(); };

  inline void addPartnerNode(InterferenceGraphNode* n) { partnerNodes.insert(n);}
  inline InterferenceGraphNodeIterator getPartnerNodesBegin() { return partnerNodes.begin(); };
  inline InterferenceGraphNodeIterator getPartnerNodesEnd()   { return partnerNodes.end(); };
  inline unsigned getNumPartners() { return partnerNodes.size(); };

  inline void replacePartner(InterferenceGraphNode *NewPartner, InterferenceGraphNode *OldPartner) {
    InterferenceGraphNodeIterator OldIter = partnerNodes.find(OldPartner);
    assert( ( OldIter != partnerNodes.end() ) && "Failing to update a partner node." );
    partnerNodes.erase(OldIter);
    partnerNodes.insert(NewPartner);
  };

  /// Clears the adjacency list
  inline void clearNeighborList() { EdgeSet.clear();}
  
  inline void deleteNeighbor(InterferenceGraphNode *node) { EdgeSet.erase(node); }

  inline std::set<InterferenceGraphNode*>& getES() { return EdgeSet; }

  ///
  /// Print the graph node
  ///
  void dump(const TargetRegisterInfo *TRI, MachineFunction *MF, InterferenceGraph *IG, bool deep = false);

};

 
struct ComparePhysicalRegNodes {
  bool operator() (InterferenceGraphNode* n1, InterferenceGraphNode* n2) {
    return (n1->index < n2->index);
  }
};

typedef std::set<InterferenceGraphNode*, ComparePhysicalRegNodes>::iterator InterferenceGraphNodePhysRegIterator;


struct CompareInterferenceGraphNode2 {
  bool operator() (InterferenceGraphNode* n1, InterferenceGraphNode* n2) {
    
    return (  n1->getDegree() < n2->getDegree() );
  }

};


struct CompareInterferenceGraphNodeCK {
  bool operator() (InterferenceGraphNode* n1, InterferenceGraphNode* n2) {    
    return (n1->CKWeight > n2->CKWeight);
  }

};

struct CompareInterferenceGraphNodeChaitin {
  bool operator() (InterferenceGraphNode* n1, InterferenceGraphNode* n2) {

    unsigned N1Degree = n1->getDegree();
    unsigned N2Degree = n2->getDegree();

    double power = 1;  // was 2

    // Changed to long long instead of double because if the allocator is
    // compiled with -O3, std::sort crashes if this comparator is used
    long long val1 = N1Degree == 0 ? 1 :
      (long long) (pow(n1->Refs,power)/N1Degree * 1000);
    long long val2 = N2Degree == 0 ? 1 :
      (long long) (pow(n2->Refs,power)/N2Degree * 1000);

    // Spills are twice as expensive as rematerializations.
    val1 *= 2;
    val2 *= 2;

    // Compute rematerialization costs.
    if (n1->getSingleDef()) {
      long long RematVal1 = N1Degree == 0 ? 0 :
        (long long) (pow(n1->Remats,power)/N1Degree * 1000);
      if (RematVal1 <= val1)
        val1 = RematVal1;
      else assert(false && "Should not get here!"); // If we get here we need to change the spilling code. It relies on only testing getSingleDef.
    }
    if (n2->getSingleDef()) {
      long long RematVal2 = N2Degree == 0 ? 0 :
        (long long) (pow(n2->Remats,power)/N2Degree * 1000);
      if (RematVal2 <= val2)
        val2 = RematVal2;
      else assert(false && "Should not get here!"); // If we get here we need to change the spilling code. It relies on only testing getSingleDef.
    }

    return (val1 > val2);
  }

};

 struct CompareInterferenceGraphNodePureRefs {

   bool operator() (InterferenceGraphNode* n1, InterferenceGraphNode* n2) {
     return (n1->Refs > n2->Refs);
   }

 };



 struct CompareInterferenceGraphNodeRefs {

   bool operator() (InterferenceGraphNode* n1, InterferenceGraphNode* n2) {
    
     /*double power = 1.5;
     
     double val1 = (pow(n1->Refs,power)/n1->getDegree());
     double val2 = (pow(n2->Refs, power)/n2->getDegree()) ;
     */

     const double LowSpillCost = 3;
     const double CutOffRefs = 5;
     const double CutOffDegree = 30;
     const double bias = 50;

     // If the node is less than cutoffrefs, then we want to give it a close to infinite spill 
     // cost, i.e. we want it at the beginning of the sorted list
     
     // If the node has high degree, then we want it at the end of the sorted list

     // Else we want the ones with the most references to be at the beginning of the list

     double val1, val2;

     if (n1->getDegree() > CutOffDegree) {
       val1 = (n1->getDegree() + bias) * 10;
     }

     else if (n1->Refs < CutOffRefs) {
       val1 = LowSpillCost;
     }
     else {
       val1 = 3 + (1/n1->Refs);
     }
          
     if (n2->getDegree() > CutOffDegree) {
       val2 = (n2->getDegree() + bias) * 10;
     }
     else if (n2->Refs < CutOffRefs) {
       val2 = LowSpillCost;
     }
     else {
       val2 = 3 + 1/n2->Refs;
     }
         
     

     return (val1 < val2);

  }

 
 };




/*
struct CompareDegrees {
  bool operator() (unsigned n1, unsigned n2) {
    
    return (DegreeVec[n1] > DegreeVec[n2]);
  }

};
*/

 
struct CompareInterferenceGraphNode {
  bool operator() (InterferenceGraphNode* n1, InterferenceGraphNode* n2) {
    
    if (n1->index == n2->index) {
      return false;
    }
    else if (n1->getDegree() == n2->getDegree()) {
      return (n1->index < n2->index);
    }
    else {
      return (n1->getDegree() > n2->getDegree());
    } 
    
  }

};


 struct CompareIGNodes_PartnerNodes {
   bool operator() (InterferenceGraphNode* n1, InterferenceGraphNode* n2) {
     

     int n1Parts = n1->getNumPartners();
     int n2Parts = n2->getNumPartners();

     if (n1Parts == n2Parts) {
       return (n1->index < n2->index);
     }
     else {
       return (n1Parts < n2Parts);
     }
     


   }
 };




 struct CompareInterferenceGraphNodeChaitinBiasRefs {
   bool operator() (InterferenceGraphNode* n1, InterferenceGraphNode* n2) {
     
     double power = 2;  // was 2

     // Changed to long long instead of double because if the allocator is
     // compiled with -O3, std::sort crashes if this comparator is used
     long long val1 = (long long) (pow(n1->Refs,power)/n1->getDegree() * 1000);
     long long val2 = (long long) (pow(n2->Refs,power)/n2->getDegree() * 1000);
     
     return (val1 > val2);


   }

   

 };


}
#endif
