//===--ChaitinBriggsAllocator.cpp - Creates a Chatin Briggs allocator-------===//
// Builds a Chaitin-Briggs register allocator according to:
// "Improvements to Graph Coloring Register Allocation" -- Briggs, Cooper,
//  Torczon
//
//
//===----------------------------------------------------------------------===//
//===----------------------------------------------------------------------===//
//This software was produced with support from the Defense Advanced
//Research Projects Agency (DARPA) through AFRL Contract FA8650-09-C-1915. 
//Nothing in this work should be construed as reflecting the official policy or
//position of the Defense Department, the United States government, or
//Rice University.
//
//THIS SOFTWARE HAS BEEN APPROVED FOR PUBLIC RELEASE, UNLIMITED DISTRIBUTION.
//
//===----------------------------------------------------------------------===//

#define DEBUG_TYPE "regalloc" // -debug-only=regalloc

#include "llvm/ADT/Statistic.h"
#include "StatisticStringMap.h"

#include "llvm/Support/CommandLine.h"
#include "llvm/Support/Debug.h"

#include "llvm/Analysis/AliasAnalysis.h"

#include "llvm/CodeGen/Passes.h"
#include "llvm/CodeGen/RegAllocRegistry.h"
#include "llvm/CodeGen/CalcSpillWeights.h"
#include "llvm/CodeGen/LiveIntervalAnalysis.h"
#include "llvm/CodeGen/LiveStackAnalysis.h"

#include "llvm/CodeGen/MachineFunctionPass.h"
#include "llvm/CodeGen/MachineRegisterInfo.h"
#include "llvm/CodeGen/MachineFrameInfo.h"
#include "llvm/CodeGen/MachineInstr.h"
#include "llvm/CodeGen/MachineLoopInfo.h"

#include "llvm/Target/TargetMachine.h"
#include "llvm/Target/TargetOptions.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetRegisterInfo.h"
#include "llvm/Target/TargetOpcodes.h"

#include "VirtRegMap.h"
#include "RegisterClassInfo.h"

#include "InterferenceGraph.h"
#include "BuildInterferenceGraph.h"
#include "BuildLiveSets.h"
#include "ConstructRegisterClasses.h"
#include "Common.h"
#include "Rematerialization.h"

#include <stack>


using namespace llvm;


// Statistics for this pass ...

STATISTIC(DidRun,                         "For how many functions did Chaitin-Briggs actually run?");
STATISTIC(RegsSpilled,                    "Total Number of registers spilled");
STATISTIC(NumRematerializableRegs,        "Total Number of registers can (will) be rematerialized");
STATISTIC(NumRematerializedInsts,         "Total Number of rematerialized spill instructions");
STATISTIC(NumAvailDefRematerializedInsts,         "Total Number of rematerialized spill instructions (avail def)");
STATISTIC(LoadsAdded,                     "Total Number of loads added");
STATISTIC(StoresAdded,                    "Total Number of stores added");
STATISTIC(OperandsFolded,                 "Total Number of operands folded");
STATISTIC(InterRegCopiesBeforeAllocation, "Total number of inter register copies before allocation");
STATISTIC(InterRegCopiesAfterAllocation,  "Total number of inter register copies after allocation");

StatisticStringMap ProcRegsSpilled("ra-ChaitinBriggs", "Summary of Number of registers spilled");
StatisticStringMap ProcLoadsAdded( "ra-ChaitinBriggs", "Summary of Number of loads added");
StatisticStringMap ProcStoresAdded( "ra-ChaitinBriggs", "Summary of Number of stores added");

StatisticStringMap  BlockLoadsAdded( "ra-ChaitinBriggs", "Summary of Number of loads added per block");
StatisticStringMap  BlockStoresAdded( "ra-ChaitinBriggs", "Summary of Number of stores added per block");
StatisticStringMap  BlockOperandsFolded( "ra-ChaitinBriggs", "Summary of Number of operands folded per block");
StatisticStringMap  ProcOperandsFolded( "ra-ChaitinBriggs", "Summary of Number of operands folded");


// Command Line Options for the Register allocator
cl::opt<bool> NoFuse("nofuse", cl::init(false),
                     cl::desc("Don't fold spilled memory operands into instructions (false)"));

cl::opt<bool> SmartFolds("smart-folds", cl::init(true),
                         cl::desc("Enable smart folding of mem. ops into instructions"));

cl::opt<bool> ConservativeCoalescing("conservative-coalescing", cl::init(false),
                         cl::desc("Turn on conservative coalescing for Chaitin-Briggs allocator (false)"));

cl::opt<bool> SSACoalescing("ssac", cl::init(false),
                            cl::desc("Turn on SSA coalescing"));

cl::opt<bool> BiasedColoring("biased-coloring", cl::init(true),
                             cl::desc("Turns on biased coloring (true)"));

cl::opt<bool> BiasedTim("biased-tim", cl::init(false),
                             cl::desc("Tim's biasing -- reduce spilling (false)"));

cl::opt<bool> BiasedBrian("biased-brian", cl::init(false),
                        cl::desc("Brian's biasing -- reduce copies (false)"));

cl::opt<bool> IsAvailDefBasedRematerializationEnabled("enable-avail-def-remat",
                                                      cl::init(false),
              cl::desc("Enables available defintions based rematerialization"));

cl::opt<int> LW_CmdLine("loopweight",
                        cl::desc("This weight raised to the loop depth is added to the spill "
                                 "cost of a node"), cl::init(10));

cl::opt<SpillHeuristicName>
SpillHeuristic ("spillheuristic",
                cl::desc("Spill heuristic to use for the Chaitin-Briggs allocator"),
                cl::values(clEnumVal(chaitin,          "Regular Chaitin Briggs"),
                           clEnumVal(refcount,         "Ref count heuristic"),
                           clEnumVal(pure_refcount,    "Pure Ref count heuristic"),
                           clEnumVal(chaitin_biasrefs, "Chaitin with biased refs"),
                           clEnumValEnd),
                cl::init(chaitin));

cl::opt<bool> CalcIRCopies("calc-ir-copies", cl::desc("Calculate Inter-register copies"),
                           cl::init(true));
// Hidden Command Line Options for the Register allocator

cl::opt<bool> DebugSpills("debugspills", cl::init(false));
cl::opt<bool> DebugCoalesce("debugcoalesce", cl::init(false));
cl::opt<bool> DebugColoring("debugcoloring", cl::init(false));
cl::opt<bool> PrintSpillCosts("printspillcosts", cl::init(false));
cl::opt<bool> PrintLiveInonExit("print-livein-on-exit", cl::init(false));

// createChaitinBriggsnRegisterAllocator() must be declare in : 
//     lib/CodeGen/Passes.h
// and defined below
//
// also for link purposes, createChaitinBriggsnRegisterAllocator() needs to be called in :
//     include/llvm/CodeGen/LinkAllCodegenComponents.h

static RegisterRegAlloc
chaitinbriggsRegAlloc("chaitinbriggs", "chaitin briggs register allocator",
                      createChaitinBriggsnRegisterAllocator);

namespace {

  //
  // RAChaitinBriggs
  //

  struct RAChaitinBriggs : public MachineFunctionPass {
  public:
    static char ID;
    
    RAChaitinBriggs();
    
    virtual const char* getPassName() const {
      return "Chaitin Briggs Register Allocator";
    }
    
    virtual void getAnalysisUsage(AnalysisUsage &AU) const;
    
    bool runOnMachineFunction(MachineFunction &F);
    
  private:
    
    InterferenceGraph* FunctionGraph;

    Rematerialization Remat;
    
    BuildLiveSets *InitialLiveBuilder;
    
    BuildInterferenceGraph* Builder;
    
    MachineFunction* MF;
    
    const TargetRegisterInfo* TRI;

    MachineRegisterInfo *MRI;
    
    RegisterClassInfo RCI;

    static unsigned AllocationIterations;
    
    // Set of all available colors
    std::set<int> Colors;
    
    // Set of colors for a particular reg class (this should ultimately supercede Colors)
    std::map<const TargetRegisterClass*, std::set<int> > ColorsByClass;
    
    // Map from representative physical register -> color
    std::map<unsigned, int> PhysReg2ColorMap;
    
    // Map from color -> representative physical register
    std::map<int, unsigned> Color2PhysRegMap;
    
    // Map to give us an order in which to use colors so that more restrictive
    // registers are tried first.
    std::map<const TargetRegisterClass *, std::list<unsigned> > RegClassToAllocOrderMap;
    
    // Stack for graph coloring
    std::stack<InterferenceGraphNode*> ColoringStack;
    
    // Map from virtual register to the allocated physical register
    std::map<unsigned, unsigned> Virt2PhysRegMap;
    
    std::set<unsigned> InfiniteSpillCostRegisters;

    ///
    /// Main, driver function for register allocation and assignment
    ///
    void allocateAndAssign(MachineFunction& F,
                           std::vector<std::pair<unsigned, unsigned> >& SSACopyRegs);

    ///
    /// Some allocators need to exclude certain basic blocks from
    /// consideration during register assignment.  This returns 
    /// true if the given basic block is to be ignored.
    ///
    virtual bool blockNotInCurrentRegion(MachineBasicBlock&) { return false; }
    
    ///
    /// Some allocators need to post-process the graph after coloring.
    ///
    virtual void summarizeInterferenceGraph(BuildInterferenceGraph*) { return; }
    
    void colorPhysRegs();
    
    struct ColorStats_T { unsigned NumBiasesAssigned; ColorStats_T() : NumBiasesAssigned(0) {}; };
    bool colorNode(InterferenceGraphNode* node, ColorStats_T *ColorStats = 0);
    
    unsigned getNumColorsInClass(const TargetRegisterClass* RC);
    
    bool colorUsingAvailableColors(InterferenceGraphNode* node,
                                   std::set<int> AvailableColors);
    
    int getColorOfPhysicalReg(unsigned reg);
    
    unsigned getPhysicalRegForColor(int color);
    
        
    ///
    /// Insert spill code for the specified virtual registers
    ///
    void spillVRs(std::set<unsigned>& VRsToSpill);
    
    ///
    /// Allocate a physical register for each vrtual register
    ///
    void mapVirtToPhysRegs();
    
    ///
    /// Register assignment -- assign physical registers to virtual registers
    ///
    void changeVirtRegs();
    
    ///
    /// Mark interference nodes for spilling
    ///
    void spillNodes(std::set<InterferenceGraphNode*>& spillNodes);
    
    bool copyCoalesce(bool ConservativeCoalescing = false);
        
    bool rewriteCoalescedRegs();
    
    bool coalesceSSACopies(std::vector<std::pair<unsigned, unsigned> >& SSACopyRegs);
    
    void resetAllStructures();
    
    void CalcInterRegCopies(Statistic& StatsCounter);
    
    void SetupColors();
    
    ///
    /// The build-coalesce loop.
    ///
    void doBuildAndCoalesceLoop(MachineFunction& F, 
                                std::vector<std::pair<unsigned, unsigned> >& SSACopyRegs);

    ///
    /// Map virtual regs to physical, and do assignment.
    /// Some allocators override this and do it selectively.
    ///
    void mapAndChangeVirtRegs();

    void releaseMemory();
    void SanityCheck();

    void populateMBBLiveIns(MachineFunction &MF);
    void recomputeKilledOperands();
    
    void pushNodeOptimistically(InterferenceGraphNode *IGNode);

  };
  
}


// LLVM uses the address of RAChaitinBriggs::ID 
// as the pass's unque id (and not the actual value thereof)
char RAChaitinBriggs::ID = 0;

// 
// Below is magic ...  Copied from RegAllocBasic.cpp
//

RAChaitinBriggs::RAChaitinBriggs() : MachineFunctionPass(ID) {
  initializeLiveIntervalsPass(*PassRegistry::getPassRegistry());
  initializeSlotIndexesPass(*PassRegistry::getPassRegistry());
  initializeStrongPHIEliminationPass(*PassRegistry::getPassRegistry());
  initializeCalculateSpillWeightsPass(*PassRegistry::getPassRegistry());
  initializeLiveStacksPass(*PassRegistry::getPassRegistry());
  initializeMachineDominatorTreePass(*PassRegistry::getPassRegistry());
  initializeMachineLoopInfoPass(*PassRegistry::getPassRegistry());
  initializeVirtRegMapPass(*PassRegistry::getPassRegistry());
  initializeRenderMachineFunctionPass(*PassRegistry::getPassRegistry());
}

void RAChaitinBriggs::getAnalysisUsage(AnalysisUsage &AU) const {
  AU.setPreservesCFG();
  AU.addRequired<AliasAnalysis>();
  //AU.addPreserved<AliasAnalysis>();
  AU.addRequired<LiveIntervals>();
  //AU.addPreserved<SlotIndexes>();
  if (StrongPHIElim)
    AU.addRequiredID(StrongPHIEliminationID);
  AU.addRequired<CalculateSpillWeights>();
  AU.addRequired<LiveStacks>();
  //AU.addPreserved<LiveStacks>();
  AU.addRequiredID(MachineDominatorsID);
  //AU.addPreservedID(MachineDominatorsID);
  AU.addRequired<MachineLoopInfo>();
  //AU.addPreserved<MachineLoopInfo>();
  AU.addRequired<VirtRegMap>();
  //AU.addPreserved<VirtRegMap>();
  MachineFunctionPass::getAnalysisUsage(AU);
}



// Defines void initializeRAChaitinBriggsPass(PassRegistry&).
//
// initializeRAChaitinBriggsPass() must be declared in include/llvm/InitializePasses.h 
//
// initializeRAChaitinBriggsPass() must called from llvm::initializeCodeGen() within
//     lib/CodeGen/CodeGen.cpp

INITIALIZE_PASS_BEGIN(
                      RAChaitinBriggs                     /* C++ class name */,
                      "chaitinbriggs-regalloc"            /* Command line switch? */, 
                      "Chaitin Briggs Register Allocator" /* Command line description */, 
                      false                               /* Only looks at CFG */,
                      false                               /* Analysis Pass */
                      )
//INITIALIZE_PASS_DEPENDENCY(LoopInfo)
INITIALIZE_PASS_END(
                    RAChaitinBriggs                     /* C++ class name */,
                    "chaitinbriggs-regalloc"            /* Command line switch? */, 
                    "Chaitin Briggs Register Allocator" /* Command line description */, 
                    false                               /* Only looks at CFG */,
                    false                               /* Analysis Pass */
                    )

FunctionPass* llvm::createChaitinBriggsnRegisterAllocator() {
  return new RAChaitinBriggs();
}


/*static*/ unsigned RAChaitinBriggs::AllocationIterations = 0;

///
/// colorPhysRegs -- Colors all the physical registers in the interference graph
///
void RAChaitinBriggs::colorPhysRegs() {
  
  for (InterferenceGraphNodePhysRegIterator PRIter = FunctionGraph->getPhysRegBegin();
       PRIter != FunctionGraph->getPhysRegEnd();
       ++PRIter) {
    if (! colorNode(*PRIter)) {
      DEBUG(dbgs() << "Cannot color a physical register!\n";);
      exit(-1);
    }
  }
}


///
/// colorNode -- colors the specified node, giving it a color distinct from
/// its neighbors
///
bool RAChaitinBriggs::colorNode(InterferenceGraphNode* node, ColorStats_T *ColorStats /* = 0 */) {
  
  // Check if the node has been marked for spilling
  if (node->ForceSpill) {
    return false;
  }
  
  if (DebugColoring) {
    if ( TargetRegisterInfo::isPhysicalRegister(node->index) ) {
      dbgs() << "Coloring node: " << PrintReg(node->index) << "(" << PrintReg(node->index, TRI) << ")\n";
    }
    else {
      const TargetRegisterClass *RC;
      RC = MRI->getRegClass(node->index);
      dbgs() << "Coloring node: " << PrintReg(node->index) << " class: " << RC->getName() << "\n";
    }

    unsigned NumRemovedNeighbors = 0;
    for (InterferenceGraphNodeIterator siter = node->getNeighborsBegin(); 
         siter != node->getNeighborsEnd(); ++siter) {
      InterferenceGraphNode *NeighborNode = *siter;
      unsigned NeighborReg = NeighborNode->index;
      
      if ( FunctionGraph->isRemovedNode(NeighborReg) ) {
        NumRemovedNeighbors++;
      }
    }
    
    unsigned InitialDegree = node->getNumNeighbors();
    unsigned CurrentDegree = InitialDegree - NumRemovedNeighbors;
    dbgs() << "\tInitial degree is " << InitialDegree << ".\n";
    dbgs() << "\tCurrent degree is " << CurrentDegree << ".\n";
    
    if ( node->OptimisticallyPushed ) {
      dbgs() << "\tOptimisitically Pushed.\n";
    }
  }
  
  std::set<int> NeighborColors;
  std::set<int> AvailableColors;
  std::set<InterferenceGraphNode*> NeighborPartnersWithBias;
  std::set<InterferenceGraphNode*> OptimisticallyPushedNeighbors;

  for (InterferenceGraphNodeIterator NI = node->getNeighborsBegin();
       NI != node->getNeighborsEnd(); ++NI) {
    InterferenceGraphNode *NeighborNode = *NI;
    unsigned NeighborReg = NeighborNode->index;
    int NeighborColor = NeighborNode->Color;
    
    if ( DebugColoring ) {
      dbgs() << "\tNeighbor node ";
      if ( ! FunctionGraph->isRemovedNode(NeighborReg) ) {
        dbgs() << PrintReg(NeighborReg); 
        if ( TargetRegisterInfo::isPhysicalRegister(NeighborNode->index) ) {
          dbgs() << "(" << PrintReg(NeighborReg, TRI) << ")";
        }
        dbgs() << " has color ";
        dbgs() << NeighborColor;
        dbgs() << " (";
        if ( NeighborColor != InterferenceGraphNode::NO_COLOR ) {
          dbgs() << PrintReg(getPhysicalRegForColor(NeighborColor), TRI);
        }
        else {
          dbgs() << "None";
        }
        dbgs() << ")";
      }
      else {
        dbgs() << "removed(" << PrintReg(NeighborReg) << ")";
      }
    }

    if ( ! FunctionGraph->isRemovedNode(NeighborReg) ) {
      
      // Two cases:
      //   i)  NeighborNode is a PR and was never removed
      //   ii) NeighborNode is a VR and is no longer removed from the graph => VR is colored
      
      if ( NeighborColor != InterferenceGraphNode::NO_COLOR ) {
        NeighborColors.insert(NeighborColor);
      }
      else {
        if ( TargetRegisterInfo::isVirtualRegister(NeighborReg) ) {
          // Neighbor is an uncolored virtual register
          for (InterferenceGraphNodeIterator NPI = NeighborNode->getPartnerNodesBegin(); 
               NPI != NeighborNode->getPartnerNodesEnd(); 
               ++NPI) {
            InterferenceGraphNode *NeighborPartnerNode = *NPI;
            int NeighborPartnerColor = NeighborPartnerNode->Color;

            if ( NeighborPartnerColor != InterferenceGraphNode::NO_COLOR ) {
              // Partner of the neighbor is either a physical register or
              // or a colored virtual register
              NeighborPartnersWithBias.insert(NeighborPartnerNode);
              if ( DebugColoring ) {
                dbgs() << ", biased(" << NeighborPartnerColor << " " 
                << PrintReg(getPhysicalRegForColor(NeighborPartnerColor), TRI) << ")";
              }
            }
          }
        }
      }
    }
    else {
      if ( TargetRegisterInfo::isVirtualRegister(NeighborReg) ) {
        for (InterferenceGraphNodeIterator NPI = NeighborNode->getPartnerNodesBegin(); 
             NPI != NeighborNode->getPartnerNodesEnd(); 
             ++NPI) {
          InterferenceGraphNode *NeighborPartnerNode = *NPI;
          int NeighborPartnerColor = NeighborPartnerNode->Color;
          
          if ( NeighborPartnerColor != InterferenceGraphNode::NO_COLOR ) {
            // Partner of the neighbor is either a physical register or
            // or a colored virtual register
            NeighborPartnersWithBias.insert(NeighborPartnerNode);
            if ( DebugColoring ) {
              dbgs() << ", biased(" << NeighborPartnerColor << " " 
              << PrintReg(getPhysicalRegForColor(NeighborPartnerColor), TRI) << ")";
            }
          }
        }
      }
    }

    if ( NeighborNode->OptimisticallyPushed ) {
      OptimisticallyPushedNeighbors.insert(NeighborNode);
      if ( DebugColoring ) {
        dbgs() << ", optimistically pushed";
      }
    }
    
    if ( DebugColoring ) {
      dbgs() << "\n";
    }
    
  }
  
  // AvailableColors = Colors - NeighborColors

  set_difference(Colors.begin(), Colors.end(),
                 NeighborColors.begin(), NeighborColors.end(),
                 inserter(AvailableColors, AvailableColors.begin()));
  
  // If no colors are left, then this node cannot be colored
  if ( AvailableColors.empty() ) {
    if ( DebugColoring ) {
      dbgs() << "NO AVAILABLE COLOR. Node will be spillled.";
      if ( ! node->OptimisticallyPushed )  {
        dbgs() << ", Node was NOT optimistically colored.";
      }
      dbgs() << "\n";
    }

    // only optimistically pushed nodes should be spilled
    if ( ! node->OptimisticallyPushed )  {
      assert(false && "colorNode() did not find a color but should have. "
             "Spilling node was NOT optimistically pushed.");
    }

    return false;
  }
  
  // If it's a physical register, then check if the corresponding color
  // is available
  // If not, then probably a virtual register was colored before a physical register!!
  if ( FunctionGraph->isAPhysicalReg(node) ) {
    
    // Get the corresponding color
    int correspColor = getColorOfPhysicalReg(node->index);
    
    assert ( (AvailableColors.find(correspColor) != AvailableColors.end()) &&
            "Color for physical register not available!");
    
    node->Color = correspColor;
    return true;
  }
  
  // For virtual registers that are pre-colored (not currently done).
  else if ( node->PreColorIndex != InterferenceGraphNode::NO_PRECOLOR_INDEX ) {
    
    // Get the corresponding color (the index must be a physical RepReg).
    int correspColor = getColorOfPhysicalReg(node->PreColorIndex);
    
    assert ( (AvailableColors.find(correspColor) != AvailableColors.end()) &&
            "Color for pre-colored virtual register not available!");
    
    node->Color = correspColor;
    return true;
    
  }
  
  // For virtual registers
  // Get PRs than can be used for this particular VR. Check if a color is free
  else {
    
    assert(TargetRegisterInfo::isVirtualRegister(node->index) &&
           "Virtual register expected");

    if ( DebugColoring ) {
      dbgs() << "\tAvailable colors are: { ";
      for (std::set<int>::iterator a = AvailableColors.begin(); a != AvailableColors.end(); ++a) {
        dbgs() << *a
        << " ("
        << TRI->getName(getPhysicalRegForColor(*a)) << "), ";
      }
      dbgs() << "}\n";
    }
    
    const TargetRegisterClass *RegClass = MF->getRegInfo().getRegClass(node->index);
    
    // Try biased coloring first
    
    if ( BiasedColoring ) {
      
      if (DebugColoring) {
        dbgs() << "\t" << node->getNumPartners() << " Partners.\n";
      }

      // if the given node has colored bias, use that color and return.

      for (InterferenceGraphNodeIterator PI = node->getPartnerNodesBegin(); PI != node->getPartnerNodesEnd();
           ++PI) {
        InterferenceGraphNode *PartnerNode = *PI;
        
        if (DebugColoring) {        
          dbgs() << "\tPartner node: " << PrintReg(PartnerNode->index) 
                << " with assigned color: "
                << PartnerNode->Color << " (";
          if ( PartnerNode->Color != -1 ) {
            // dbgs() << TRI->getName(getPhysicalRegForColor((PartnerNode->Color))); 
            dbgs() << PrintReg(getPhysicalRegForColor((PartnerNode->Color)), TRI); 
          }
          else {
            dbgs() << "None";
          }
          dbgs() << ")\n";
        }
        
        if ( PartnerNode->Color != InterferenceGraphNode::NO_COLOR) {
          if (AvailableColors.find(PartnerNode->Color) != AvailableColors.end()) {
            
            if ( DebugColoring ) {
              dbgs() << "\tChoosing biased color node: " << PartnerNode->Color << " " 
              << PrintReg(getPhysicalRegForColor(PartnerNode->Color), TRI) 
              << " (" << PrintReg(PartnerNode->index, TRI)<< ")";
              if ( node->OptimisticallyPushed ) {
                dbgs() << " OPTIMISTICALLY PUSHED, GOT LUCKY";
              }
              dbgs() << "\n";
            }
            
            node->Color = PartnerNode->Color;
            if ( ColorStats != 0 ) {
              ColorStats->NumBiasesAssigned++;
            }
            return true;
          }
        }
      }
      
      if ( BiasedBrian ) {
        // Tim Harvey's Limited look ahead as interpreted by Brian West:
        // Targeted case: VR x VR x PR
        // Summary: help optimistically colored neighbor reduce its degree and not spill
        //
        // For all of the uncolored neighbors that are optimistically pushed on the stack, collect the colors 
        // from their colored neignbors.  Pick a color from this set and an uncolored neighbor's 
        // degree will be reduced which may prevent a spill.
        
        if ( OptimisticallyPushedNeighbors.size() > 0 ) {
          unsigned int NumColors = PhysReg2ColorMap.size();
          std::vector<int> ColorCount;
          std::list<unsigned int> ColorsToTry;
          
          ColorCount.clear();
          ColorCount.resize(NumColors);
          
          ColorsToTry.clear();
          
          for (InterferenceGraphNodeIterator NI = OptimisticallyPushedNeighbors.begin();
               NI != OptimisticallyPushedNeighbors.end(); ++NI) {
            InterferenceGraphNode *OptimisticallyPushedNeighbor = *NI;
            std::set<int> NeighborNeighborColors;
            
            NeighborNeighborColors.clear();
            for (InterferenceGraphNodeIterator NNI = OptimisticallyPushedNeighbor->getNeighborsBegin();
                 NNI !=  OptimisticallyPushedNeighbor->getNeighborsEnd(); ++NNI) {
              InterferenceGraphNode *NeighbordNeighborNode = *NNI;
              int NeighborNeighborColor = NeighbordNeighborNode->Color;
              
              if ( NeighborNeighborColor != InterferenceGraphNode::NO_COLOR ) {
                NeighborNeighborColors.insert(NeighborNeighborColor);
              }
            }
            
            // foreach Color in NeighborNeighborColors, ColorCount[Color]++

            for (std::set<int>::iterator NNCI = NeighborNeighborColors.begin();
                 NNCI != NeighborNeighborColors.end(); ++NNCI) {
              unsigned int Color = *NNCI;
              
              ColorCount[Color]++;
            }
#if 1
            if ( DebugColoring ) {
              if ( NeighborNeighborColors.size() > 0 ) {
                dbgs() << "\t\tOptimistically pushed neighbor " << PrintReg(OptimisticallyPushedNeighbor->index) 
                << " has colored neighbors which use the " << NeighborNeighborColors.size() << " colors: {";
                for (std::set<int>::iterator NNCI = NeighborNeighborColors.begin();
                     NNCI != NeighborNeighborColors.end(); ++NNCI) {
                  unsigned int Color = *NNCI;
                  dbgs() << Color
                  << " (" << TRI->getName(getPhysicalRegForColor(Color)) << "), ";
                }
                dbgs() << "}\n";
              }
            }
#endif
            // ColorCount[Color] == 2 =>
            // for the given node, it has 2 optimistically colored neighbors 
            // which have colored neighbors with color Color
          }
          
          // foreach nonzero ColorCount[Color], add to ColorsToTry in reverse sorted order
          
          std::list<unsigned>::iterator RI = RegClassToAllocOrderMap[RegClass].begin();
          std::list<unsigned>::iterator RE = RegClassToAllocOrderMap[RegClass].end();
          
          for (; RI != RE; ++RI) {
            
            unsigned RepReg = Builder->getRepresentativeNumber(*RI);
            unsigned NewColor = getColorOfPhysicalReg(RepReg);
            int Count = ColorCount[NewColor];
            
            if ( Count > 0 ) {
              std::list<unsigned int>::iterator CTTI;
              for (CTTI = ColorsToTry.begin(); CTTI != ColorsToTry.end(); ++CTTI) {
                int OldColor = *CTTI;
                
                if ( ColorCount[NewColor] > ColorCount[OldColor] ) {
                  ColorsToTry.insert(CTTI, NewColor); // insert NewColor before OldColor
                  break;
                }
              }
              if ( CTTI == ColorsToTry.end() ) {
                ColorsToTry.insert(ColorsToTry.end(), NewColor); // insert NewColor at end
              }
            }
          }
          
          if ( ColorsToTry.size() > 0 ) {
            if ( DebugColoring ) {
              dbgs() << "\tReduced the degree of optimistically pushed neighbor(s) by choosing from: { ";
              for (std::list<unsigned int>::iterator CTTI = ColorsToTry.begin(); 
                   CTTI != ColorsToTry.end(); ++CTTI) {
                int Color = *CTTI;
                if ( AvailableColors.find(Color) != AvailableColors.end() ) {
                  dbgs() << Color << " ("
                  << PrintReg(getPhysicalRegForColor(Color), TRI) << ") x"
                  << ColorCount[Color] << ", ";
                }
              }
              dbgs() << "}\n";
            }
            
            for (std::list<unsigned int>::iterator CTTI = ColorsToTry.begin(); 
                 CTTI != ColorsToTry.end(); ++CTTI) {
              int Color = *CTTI;
              
              if ( AvailableColors.find(Color) != AvailableColors.end() ) {
                if ( DebugColoring ) {
                  dbgs() << "\tReduced the degree of optimistically pushed neighbor(s) by choosing "
                  << Color << " ("
                  << PrintReg(getPhysicalRegForColor(Color), TRI) << "), helps " 
                  << ColorCount[Color] << " neighbors.";
                  if ( node->OptimisticallyPushed ) {
                    dbgs() << " OPTIMISTICALLY PUSHED, GOT LUCKY";
                  }
                  dbgs() << "\n";
                }
                //break;
                node->Color = Color;
                return true;
              }
            }
          }
        }
      }
      
      // Biased coloring failed.  Reduce the available colors to facilate a neighbor or
      // partner getting its biased color.  Two methods are used.  The first came from Preston's
      // thesis.  The second came from Brian West inability to read said thesis :)
      
      {
        // Preston Briggs's Limited look ahead:
        // Targeted case: VR1 <- VR2 x PR
        // Summary: help VR2 get its biased color
        //
        // Effectively want to remove partner VR2's colored interferences (neighbors) 
        // from VR1's avalable colors. Later when VR2 is colored, the bias coloring logic
        // has a chance to pick VR1's color and this copy can be coalesced.
        
        if ( AvailableColors.size() > 1 ) {
          std::set<int> PartnerNeighborColors;
          std::set<int> PartnerAvailableColors;
          std::set<int> NewAvailableColors;
          bool UpdateAvailableColors;
          
          UpdateAvailableColors = false;
          
          for (InterferenceGraphNodeIterator PI = node->getPartnerNodesBegin();
               PI != node->getPartnerNodesEnd(); ++PI) {
            InterferenceGraphNode *PartherNode = *PI;
            
            for (InterferenceGraphNodeIterator PNI = PartherNode->getNeighborsBegin();
                 PNI != PartherNode->getNeighborsEnd(); ++PNI) {
              InterferenceGraphNode *NeighborNode = *PNI;
              int NeighborColor = NeighborNode->Color;
              
              if ( NeighborColor != InterferenceGraphNode::NO_COLOR ) {
                PartnerNeighborColors.insert(NeighborColor);
              }
            }
          }
          
          if ( PartnerNeighborColors.size() > 0 ) {
            if ( DebugColoring ) {
              dbgs() << "\tPartners neighbors colors are: { ";
              for (std::set<int>::iterator ColorIter = PartnerNeighborColors.begin();
                   ColorIter != PartnerNeighborColors.end(); ++ColorIter) {
                int PartnerNeighborColor = *ColorIter;
                dbgs() << PartnerNeighborColor
                << " (" << TRI->getName(getPhysicalRegForColor(PartnerNeighborColor)) << "), ";
              }
              dbgs() << "}\n";
            }
            
            for (std::set<int>::iterator ColorIter = PartnerNeighborColors.begin();
                 ColorIter != PartnerNeighborColors.end(); ++ColorIter) {
              int PartnerNeighborColor = *ColorIter;
              
              if ( AvailableColors.find(PartnerNeighborColor) != AvailableColors.end() ) {
                AvailableColors.erase(AvailableColors.find(PartnerNeighborColor));
                UpdateAvailableColors = true;
                if ( AvailableColors.size() == 1 ) break;
              }
            }            
            
            if ( UpdateAvailableColors && DebugColoring ) {
              dbgs() << "\tAvailable colors after Preston's biasing are: { ";
              for (std::set<int>::iterator ColorIter = AvailableColors.begin(); 
                   ColorIter != AvailableColors.end(); ++ColorIter) {
                int AvailableColor = *ColorIter;
                dbgs() << AvailableColor
                << " (" << TRI->getName(getPhysicalRegForColor(AvailableColor)) << "), ";
              }
              dbgs() << "}\n";
            }
          }
        }
      }
      
      if ( BiasedBrian ) {
        // Brian West's Limited look ahead:
        // Targeted case: VR1 x VR2 <- PR
        // Summary: help neighbor get its biased color
        //
        // If the given node VR1 has neighbors with a colored bias, exclude that color from
        // AvailableColors if possible.  This may allow the neighbor VR2 to be colored with
        // its biased color.
        
        if ( AvailableColors.size() > 1 && NeighborPartnersWithBias.size() > 0 ) {
          InterferenceGraphNodeIterator NeighborPartnersWithBiasIter;
          InterferenceGraphNodeIterator NeighborPartnersWithBiasIterEnd;
          bool UpdateAvailableColors;
          
          UpdateAvailableColors = false;
          
          NeighborPartnersWithBiasIter    = NeighborPartnersWithBias.begin();
          NeighborPartnersWithBiasIterEnd = NeighborPartnersWithBias.end();
          for (; NeighborPartnersWithBiasIter != NeighborPartnersWithBiasIterEnd; 
               ++NeighborPartnersWithBiasIter) {
            InterferenceGraphNode *NeighborPartnerWithBias = *NeighborPartnersWithBiasIter;
            int NeighborPartnerColor = NeighborPartnerWithBias->Color;
            
            if ( AvailableColors.find(NeighborPartnerColor) != AvailableColors.end() ) {
              AvailableColors.erase(AvailableColors.find(NeighborPartnerColor));
              UpdateAvailableColors = true;
              if ( AvailableColors.size() == 1 ) break;
            }
          }
          
          if ( UpdateAvailableColors && DebugColoring ) {
            dbgs() << "\tAvailable colors after Brian's biasing are: { ";
            for (std::set<int>::iterator ColorIter = AvailableColors.begin(); 
                 ColorIter != AvailableColors.end(); ++ColorIter) {
              int AvailableColor = *ColorIter;
              dbgs() << AvailableColor
              << " (" << TRI->getName(getPhysicalRegForColor(AvailableColor)) << "), ";
            }
            dbgs() << "}\n";
          }
        }
      }
      
    }
    
    // Biased coloring failed
    
    std::list<unsigned>::iterator RI = RegClassToAllocOrderMap[RegClass].begin();
    std::list<unsigned>::iterator RE = RegClassToAllocOrderMap[RegClass].end();
    unsigned RepReg;
    unsigned ColorForReg;
    for (; RI != RE; ++RI) {
      
      RepReg = Builder->getRepresentativeNumber(*RI);
      ColorForReg = getColorOfPhysicalReg(RepReg);
      
      if ( AvailableColors.find(ColorForReg) != AvailableColors.end() ) {
        
        // Found an available color
        node->Color = ColorForReg;
        
        if ( DebugColoring ) {
          dbgs() << "\tFound color "
          << node->Color << " "
          << PrintReg(getPhysicalRegForColor(node->Color), TRI)
          << "(" << PrintReg(RepReg) << ")";
          if ( node->OptimisticallyPushed ) {
            dbgs() << " OPTIMISTICALLY PUSHED, GOT LUCKY";
          }
          dbgs() << "\n";
        }
        
        return true;
      }
    } // for

    if ( DebugColoring ) {
      RI = RegClassToAllocOrderMap[RegClass].begin();
      dbgs() << "Allocation order for register class: " << RegClass->getName() << "\n{ ";
      for (; RI != RE; ++RI) {        
        unsigned RepReg = Builder->getRepresentativeNumber(*RI);

        dbgs() << PrintReg(RepReg, TRI) << ", ";
      }
      dbgs() << "}\n";
    }
  }
  
  // No color was found
  if ( DebugColoring ) {
    dbgs() << "NO COLOR was found. Node will be spillled.\n";
  }
#if 1
  if ( ! node->OptimisticallyPushed )  {
    assert(false && "colorNode() did not find a color but should have.");
  }
#endif

  return false;
}


unsigned RAChaitinBriggs::getNumColorsInClass(const TargetRegisterClass* RC) {
  assert ( ColorsByClass.find(RC) != ColorsByClass.end() && "Could not find register class");  
  return ColorsByClass[RC].size();
}


bool RAChaitinBriggs::colorUsingAvailableColors(InterferenceGraphNode* node,
                                                std::set<int> AvailableColors) {
  return true;
}


int RAChaitinBriggs::getColorOfPhysicalReg(unsigned reg) {
  assert ((PhysReg2ColorMap.find(reg) != PhysReg2ColorMap.end()) &&
          "Color not found for physical register");
  return PhysReg2ColorMap[reg];
}


unsigned RAChaitinBriggs::getPhysicalRegForColor(int color) {
  assert( (Color2PhysRegMap.find(color) != Color2PhysRegMap.end()) &&
         "Physical register not found for color");
  assert(Color2PhysRegMap[color != 0]);
  return Color2PhysRegMap[color];
}


/// SpillHelper - Computes spill cost or emits spill code depending on
/// the constructor used.
class SpillHelper {
  Rematerialization &Remat;
  DenseMap<MachineOperand *, MachineOperand *> PreviousMention; // By BB.
  SmallVector<MachineOperand *, 32> Spills; // By BB.
  DenseSet<MachineOperand *> NoDeferredLoadsPending; // By BB.

  DenseMap<unsigned, int> StackSlotForVirtReg; // Whole Function.

  std::set<unsigned> *InfiniteSpillCostRegs; // Used for spill cost computation.
  std::set<unsigned> *VRsToSpill; // Used for emitting spill code.

  bool DoComputeCost;

  MachineBasicBlock *Block;

  const TargetRegisterInfo *TRI;
  MachineFunction *MF;
  const TargetMachine *TM;
  BuildInterferenceGraph *Builder;
  MachineLoopInfo *LoopInfo;
  bool CallsKillRegs;

  /// buildPreviousMentions - Builds a map of machine operand to its previous
  /// mention in another instruction.
  void buildPreviousMentions();

  // spill - Spill the machine register. If DoComputeCost is true computes the
  // spill cost for that register occurence, else stores the operand to be
  // spilled.
  void spill(MachineOperand &Oper, DenseSet<unsigned> &SpillRegisters);

  void emitSpills();

  void emitSpillCode(MachineOperand &Oper,
                     SmallVector<MachineOperand*, 32>::iterator SpillsIter,
                     SmallVector<MachineOperand*, 32>::iterator SpillsIterEnd);

  int getStackSpaceFor(unsigned Reg, const llvm::TargetRegisterClass* RC);
  bool shouldIgnoreRegister(unsigned Reg);
  bool shouldFoldOperand(MachineInstr* MI, unsigned Oper, bool OperIsDef);
  MachineInstr *isRematerializable(unsigned Reg);

  /// analyseDeferredSpills - Fills NoDeferredLoadsPending set with loads that
  /// service only one instruction.
  void analyseDeferredSpills();
public:
  // Constructor to emit spill code.
  SpillHelper(Rematerialization &R,
              std::set<unsigned> *VRsToSpill,
              const TargetRegisterInfo *RegInfo,
              BuildInterferenceGraph *Builder);

  // Constructor to compute spill cost.
  SpillHelper (Rematerialization &R,
               std::set<unsigned> *InfiniteCostRegs,
               const TargetRegisterInfo *RegInfo,
               BuildInterferenceGraph *Builder,
               MachineFunction *F,
               MachineLoopInfo *LI);

  void initializeInfiniteSpillCostRegisters(MachineFunction *MF);
  void spillRegisters();
  void dumpInfiniteCost(const char *description = 0);
  void startProcessingBlock(MachineBasicBlock *BB);
};

MachineInstr *SpillHelper::isRematerializable(unsigned Reg) {
  // Catch noreg registers.
  assert(Reg != 0 && "We need a valid register number");

  // Get graph node.
  InterferenceGraph *Graph = Builder->getInterferenceGraph();
  InterferenceGraphNode *Node = Graph->getNodeIfInGraph(Reg);
  assert(Node != NULL);

  if (!Node) return NULL;

  // Check whether there is a single definition.
  return Node->getSingleDef();
}

/// shouldFoldOperand - Check if register operand should be folded.
///
/// Given a register R and whether we are checking for a use or a def of R,
/// iterate through each use or def since a def or a use corresponds to a line
/// of potential spill code.
/// If we can fold all uses or (exclusive or) defs of R  into the instruction,
/// then return true. Else, we have to spill and folding the memory operand will
/// prove to be redundant, thus return false.
bool SpillHelper::shouldFoldOperand(MachineInstr* MI, unsigned Oper,
                                    bool OperIsDef) {
  if (!SmartFolds)
    return false;

  MachineInstr* FoldedInst = NULL;
  MachineInstr* TestInstr = MF->CloneMachineInstr(MI);


  MachineOperand &Op = TestInstr->getOperand(Oper);
  unsigned TheReg = Op.getReg();

  // Don't fold Operands that are rematerializable.
  if (TheReg && (isRematerializable(TheReg) ||
                 (!OperIsDef && Remat.isRematerializable(Op)))) {
    MF->DeleteMachineInstr(TestInstr);
    return false;
  }

  for (unsigned I = Oper; I != TestInstr->getNumOperands(); ++I) {
    if (TestInstr->getOperand(I).isReg() &&
        (TestInstr->getOperand(I).isDef() == OperIsDef) &&
        TestInstr->getOperand(I).getReg() == TheReg) {
      // Dummy frame index
      int FrameIndex = 0;

      SmallVector<unsigned, 1> Ops; Ops.push_back(I);
      if (TestInstr->getParent() != NULL
          && (FoldedInst = TM->getInstrInfo()->foldMemoryOperand(TestInstr, Ops,
                                                                 FrameIndex))) {
        // Delete TestInstr.
        MF->DeleteMachineInstr(TestInstr);
        TestInstr = FoldedInst;

        // Restart the loop.
        I = -1;
      } else {
        if (TestInstr)
          MF->DeleteMachineInstr(TestInstr);

        return false;
      }
    } // if
  } // for

  if (TestInstr)
    MF->DeleteMachineInstr(TestInstr);

  return true;
}

void SpillHelper::dumpInfiniteCost(const char *description /*= 0*/) {
  dbgs() << "Infinited Registers: " << (description?description:"") << "\n\t{ ";
  for (std::set<unsigned>::iterator SI = InfiniteSpillCostRegs->begin(),
       SE = InfiniteSpillCostRegs->end(); SI != SE; ++SI) {
    dbgs() << PrintReg(*SI) << ", ";
  }
  dbgs() << "}\n";
}

// Constructor to emit spill code.
SpillHelper::SpillHelper(Rematerialization &R,
                         std::set<unsigned> *SpillVRs,
                         const TargetRegisterInfo *RegInfo,
                         BuildInterferenceGraph *IGBuilder) :
  Remat(R), InfiniteSpillCostRegs(NULL), VRsToSpill(SpillVRs),
  DoComputeCost(false), Block(NULL), TRI(RegInfo), MF(NULL), TM(NULL),
  Builder(IGBuilder), LoopInfo(NULL), CallsKillRegs(false) {
}

// Constructor to compute spill cost.
SpillHelper::SpillHelper (Rematerialization &R,
                          std::set<unsigned> *InfiniteCostRegs,
                          const TargetRegisterInfo *RegInfo,
                          BuildInterferenceGraph *IGBuilder,
                          MachineFunction *F,
                          MachineLoopInfo *LI) :
  Remat(R), InfiniteSpillCostRegs(InfiniteCostRegs), VRsToSpill(NULL),
  DoComputeCost(true), Block(NULL), TRI(RegInfo), MF(F), TM(NULL),
  Builder(IGBuilder), LoopInfo(LI), CallsKillRegs(false) {
    unsigned i;
    const unsigned *CalleeSavedRegs;

    CallsKillRegs = true;

    CalleeSavedRegs = RegInfo->getCalleeSavedRegs(MF);
    for (i=0; CalleeSavedRegs[i] != 0; i++) {
      unsigned CalleeSavedReg = CalleeSavedRegs[i];
      if ( Builder->isCorrectRegisterType(CalleeSavedReg) ) {
        CallsKillRegs = false;
        break;
      }
    }
    
    initializeInfiniteSpillCostRegisters(MF);
}


void SpillHelper::initializeInfiniteSpillCostRegisters(MachineFunction *MF) {
  if ( DoComputeCost ) {
    MachineFunction::iterator BlockIter;
    MachineFunction::iterator BlockIterEnd;
    
    BlockIter = MF->begin();
    BlockIterEnd = MF->end();
    
    for (; BlockIter != BlockIterEnd; ++BlockIter) {
      MachineBasicBlock *MB = BlockIter;
      
      MachineBasicBlock::iterator InstIter;
      MachineBasicBlock::iterator InstIterEnd;
      
      InstIter = MB->begin();
      InstIterEnd = MB->end();
      
      for (; InstIter != InstIterEnd; ++InstIter) {
        MachineInstr *MI = InstIter;
#if 0
        if ( DebugSpills ) {
          dbgs() << "BB#" << MB->getNumber() << ":";
          MI->dump();
        }
#endif
        for (unsigned I = 0, E = MI->getNumOperands(); I != E; ++I) {
          MachineOperand &Oper = MI->getOperand(I);
          unsigned Reg;
          
          // if operand is a register ...
          if ( Oper.isReg()  ) {
            Reg = Oper.getReg();
            
            // and the register is not the zero special case
            // and is a virtual register that we are currently processing ...
            if ( Reg != 0 &&
                 !shouldIgnoreRegister(Oper.getReg()) ) {
#if 0
              if ( DebugSpills ) {
                dbgs() << "BB#" << MB->getNumber() << ": Adding to Infinite Spill Cost Registers: " 
                << PrintReg(Reg) << "\n";
              }
#endif
              InfiniteSpillCostRegs->insert(Reg);
            }
          }
        }
      }
    }
  }
}

void SpillHelper::startProcessingBlock(MachineBasicBlock *BB) {
  // Initialize basic block local members.
  PreviousMention.clear();
  Spills.clear();
  NoDeferredLoadsPending.clear();
  Block = BB;

  // Initialize on first basic block.
  if (MF == NULL) {
    MF = Block->getParent();
  }
  if (TM == NULL) {
    TM = &MF->getTarget();
  }

  // Build previous mention list.
  buildPreviousMentions();
}

void SpillHelper::buildPreviousMentions() {
  // Maps a register number to the previous operand that mentioned it.
  DenseMap<unsigned, MachineOperand *> RegToPrevOperand;

  // Iterate over all instructions.
  for (MachineBasicBlock::iterator BI = Block->begin(), BE = Block->end();
       BI != BE; ++BI) {
    MachineInstr *MI = &*BI;

    // Iterate over all operands.
    for (unsigned I = 0, E = MI->getNumOperands(); I != E; ++I) {
      MachineOperand &Oper = MI->getOperand(I);

      // Only look at registers.
      if (!Oper.isReg() || Oper.getReg() == 0 ||
          shouldIgnoreRegister(Oper.getReg()))
        continue;

      unsigned Reg = Oper.getReg();
      MachineOperand *Prev = NULL;

      // Lookup previous instruction.
      if (RegToPrevOperand.find(Reg) != RegToPrevOperand.end())
        Prev = RegToPrevOperand[Reg];

      RegToPrevOperand[Reg] = &Oper;

      // Map only to a different instruction.
      while (Prev && Prev->getParent() == Oper.getParent())
        Prev = PreviousMention[Prev];

      PreviousMention[&Oper] = Prev;

    }
  }
}

static std::set<unsigned> getUses(MachineInstr *Instr) {
  std::set<unsigned> Uses;
  for (unsigned I = 0, E = Instr->getNumOperands(); I != E; ++I) {
    MachineOperand &Oper = Instr->getOperand(I);
    if (Oper.isReg() && Oper.getReg() && Oper.isUse())
      Uses.insert(Oper.getReg());
  }
  return Uses;
}

static std::set<unsigned> getDefs(MachineInstr *Instr) {
  std::set<unsigned> Defs;
  for (unsigned I = 0, E = Instr->getNumOperands(); I != E; ++I) {
    MachineOperand &Oper = Instr->getOperand(I);
    if (Oper.isReg() && Oper.getReg() && Oper.isDef())
      Defs.insert(Oper.getReg());
  }
  return Defs;
}

bool SpillHelper::shouldIgnoreRegister(unsigned Reg) {
  // Ignore physical registers.
  if (TargetRegisterInfo::isPhysicalRegister(Reg))
    return true;

  // Ignore registers of that are not of the current register class being
  // allocated.
  if (!Builder->isCorrectRegisterType(Reg))
    return true;

  // Ignore removed nodes when computing the spill cost.
  if (DoComputeCost &&
      Builder->getInterferenceGraph()->isRemovedNode(Reg))
    return true;

  // During spill code insertion only handle nodes that were choosen to be
  // spilled.
  if (!DoComputeCost && !VRsToSpill->count(Reg))
    return true;

  return false;
}

/// spillRegisters - Spills registers whose cost is not infinite. Whether a use
/// has an infinite spill cost is determined by the following algorithm.
///
///        as inst_runner runs backwards through the code,
///        current = inst_runner
///        cuses = uses(current) /* uses() returns the uses in an inst */
///        for each c in cuses:
///            prev = prev(c) /* prev() returns the previous instruction that
///                              mentions c */
///            puses = uses(prev)
///            pdefs = defs(prev) /* defs() returns the defs in an inst */
///            if c is not in pdefs /* c is in puses only */
///               if pdefs is not a subset of cuses
///                 c is not infinite
///            for each i between current and prev
///                idefs = defs(i)
///                iuses = uses(i)
///                if idefs is not in cuses
///                  c is not infinite
///                if c is in pdefs
///                  if iuses is not in pdefs
///                    c is not infinite
///                else
///                  c is not infinite
///
/// Slightly modified version of spill algorithm from the ILOC compiler,
/// courtesy Tim Harvey. Changed so that the ultimate 'else' clause always
/// marks as not infinite.
void SpillHelper::spillRegisters() {
#if 1
  if ( DebugSpills && InfiniteSpillCostRegs ) {
    //const BasicBlock *BB = Block->getBasicBlock();
    dbgs() << "\nSpillHelper::spillRegisters(): BB#" << Block->getNumber() << ":\n";
    dumpInfiniteCost("values before processing basic bock");
  }
#endif
  
  // Process instruction from the end of the block to the beginning.
  for (MachineBasicBlock::reverse_iterator RI = Block->rbegin(),
       RE = Block->rend(); RI != RE; ++RI) {
    MachineInstr *Curr = &*RI;
    
    // Registers in the current instruction that are already spilled, and that
    // will be ignored if they are encountered again.
    DenseSet<unsigned> SpilledRegisters;

    // Process definitions before uses.
    for (unsigned I = 0, E = Curr->getNumOperands(); I != E; ++I) {
      MachineOperand &CurOper = Curr->getOperand(I);

      // Ignore non-register operands, operands of the wrong register type,
      // physical registers, registers we already processed, and registers that
      // should not be spilled.
      if (!CurOper.isReg() || CurOper.getReg() == 0 ||
          SpilledRegisters.count(CurOper.getReg()) ||
          shouldIgnoreRegister(CurOper.getReg()))
        continue;

      // Spill definition.
      if (CurOper.isDef()) {
        spill(CurOper, SpilledRegisters);
      }
    }

    // We do need to spill definitions and uses.
    SpilledRegisters.clear();

    // For each virtual register use in the current instruction.
    for (unsigned I = 0, E = Curr->getNumOperands(); I != E; ++I) {
      MachineOperand &CurOper = Curr->getOperand(I);

      // Ignore non-register operands, definitions, operands of the wrong
      // register type, physical registers, registers we already processed, and
      // registers that should not be spilled.
      if (!CurOper.isReg() || CurOper.getReg() == 0 || CurOper.isDef() ||
          SpilledRegisters.count(CurOper.getReg()) ||
          shouldIgnoreRegister(CurOper.getReg()))
        continue;

      // Get previous instruction that mentions the register.
      MachineOperand *PrevReg = PreviousMention[&CurOper];

      // Spill upward exposed uses.
      if (!PrevReg) {
        spill(CurOper, SpilledRegisters);
        continue;
      }

      assert(PrevReg->getParent() != CurOper.getParent() &&
             PrevReg->getReg() == CurOper.getReg() &&
             "Must be a different instruction");

      // Get uses and definitions of the previous instruction that mentions the
      // current register.
      MachineInstr *PrevInstr = PrevReg->getParent();
      std::set<unsigned> PrevUses = getUses(PrevInstr);
      std::set<unsigned> PrevDefs = getDefs(PrevInstr);
      std::set<unsigned> CurUses = getUses(Curr);
      unsigned CurReg = CurOper.getReg();

      // CurReg is not in PrevDefs.
      if (!PrevDefs.count(CurReg))
        // PrevDefs is not in CurUses.
        if (!std::includes(CurUses.begin(), CurUses.end() ,
                           PrevDefs.begin(), PrevDefs.end())) {
          assert(!PrevDefs.empty());
          spill(CurOper, SpilledRegisters);
          continue;
        }

      // For each instruction inbetween the current and previous.
      MachineBasicBlock::reverse_iterator II =
        next(MachineBasicBlock::reverse_iterator(Curr)),
        IE = MachineBasicBlock::reverse_iterator(PrevInstr);
      for (; II != IE; ++II) {
        MachineInstr *InstInBetween = &(*II);
        const MCInstrDesc &MCID = TM->getInstrInfo()->get(InstInBetween->getOpcode());

        std::set<unsigned> IUses = getUses(InstInBetween);
        std::set<unsigned> IDefs = getDefs(InstInBetween);

        // IDefs is not in CurUses.
        if (!std::includes(CurUses.begin(), CurUses.end(),
                           IDefs.begin(), IDefs.end())) {
          assert(!IDefs.empty());
          spill(CurOper, SpilledRegisters);
          break;
        }

        // CurReg is in PrevDefs.
        if (PrevDefs.count(CurReg)) {
          // IUses is not in PrevDefs.
          if (!std::includes(PrevDefs.begin(), PrevDefs.end(),
                             IUses.begin(), IUses.end())) {
            assert(!IUses.empty());
            spill(CurOper, SpilledRegisters);
            break;
          }
        } else {
          spill(CurOper, SpilledRegisters);
          break;
        }

        // CurReg is a register that gets killed across a function call.
        if ( MCID.isCall() && CallsKillRegs ) {
          // The ILOC compiler put explicit split around function calls.
          // In LLVM, this is not done.  Call instr have a physical register operands
          // for each register that is killed across the call, but not the virtual
          // registers live across it.
          assert(false && "death on two legs");
          spill(CurOper, SpilledRegisters);
          break;
        }
      }
    }
  }

  // Emit spill code.
  if (!DoComputeCost) {
    emitSpills();
  }
#if 1
  if ( DebugSpills && InfiniteSpillCostRegs ) {
    dumpInfiniteCost("values after processing basic bock");
  }
#endif
}

/// getStackSpaceFor - Allocates space for the specified virtual register
/// to be held on the stack, or returns a previously allocated slot.
int SpillHelper::getStackSpaceFor(unsigned VirtReg,
                                  const TargetRegisterClass *RC) {

  DenseMap<unsigned, int>::iterator I = StackSlotForVirtReg.find(VirtReg);

  // There already is a stack slot.
  if (I != StackSlotForVirtReg.end())
    return I->second;

  // Allocate a new stack object for this spill location...
  int FrameIdx = MF->getFrameInfo()->CreateStackObject(RC->getSize(),
                                                       RC->getAlignment(),
                                                       true);
  // Assign the slot.
  StackSlotForVirtReg[VirtReg] = FrameIdx;

  return FrameIdx;
}

void SpillHelper::spill(MachineOperand &Oper,
                        DenseSet<unsigned> &SpilledRegisters) {
  SpilledRegisters.insert(Oper.getReg());

  // Compute the cost.
  if (DoComputeCost) {
    MachineBasicBlock *BB = Oper.getParent()->getParent();
    unsigned Reg = Oper.getReg();
    unsigned LoopDepth = LoopInfo->getLoopDepth(BB);

    // Get node.
    InterferenceGraphNode *Node =
      Builder->getInterferenceGraph()->getNodeIfInGraph(Reg);
    assert(Node != NULL && "Node must exists");

    // Set cost.
     // Lower the cost for rematerializable operands.
    if (IsAvailDefBasedRematerializationEnabled && Oper.isUse() &&
        !Node->getSingleDef() && // Handled by "simple/old" remat code.
        Remat.isRematerializable(Oper))
      Node->Refs += (pow(LW_CmdLine, LoopDepth) / 2);
    else
      Node->Refs += pow(LW_CmdLine, LoopDepth);

    // Mark spilled uses as not infinite.
    if (Oper.isUse())
      InfiniteSpillCostRegs->erase(Reg);

    return;
  }

  // Store spill location.
  Spills.push_back(&Oper);
}

void SpillHelper::emitSpills() {
  // Mark loads that only feed one instruction. They can be folded into the
  // instruction.
  analyseDeferredSpills();

  for (SmallVector<MachineOperand *, 32>::iterator SI = Spills.begin(),
       SE = Spills.end(); SI != SE; ++SI) {
    MachineOperand &Oper = **SI;
    emitSpillCode(Oper, SI, SE);
  }

}

void SpillHelper::analyseDeferredSpills() {
  // Copy spills to a lookup set.
  DenseSet<MachineOperand *> SpillSet;
  for (unsigned I = 0, E = Spills.size(); I != E; ++I)
    SpillSet.insert(Spills[I]);

  // Registers whose load is deferred.
  DenseSet<unsigned> DeferredRegs;

  // Traverse basic block backwards.
  for (MachineBasicBlock::reverse_iterator BI = Block->rbegin(),
       BE = Block->rend(); BI != BE;++BI) {
    DenseSet<unsigned> RegistersHandled;
    for (unsigned I = 0, E = BI->getNumOperands(); I != E; ++I) {
      MachineOperand &Oper = BI->getOperand(I);

      // Only consider registers that are to be spilled (loaded) from memory.
      if (!Oper.isReg() || !Oper.getReg() || !Oper.isUse() ||
          RegistersHandled.count(Oper.getReg()) ||
          !VRsToSpill->count(Oper.getReg()))
        continue;

      unsigned Reg = Oper.getReg();
      // If we encounter a spill, check if any deferred loads are pending.
      if (SpillSet.count(&Oper)) {
        // No deferred loads pending.
        if (DeferredRegs.find(Reg) == DeferredRegs.end()) {
          NoDeferredLoadsPending.insert(&Oper);
          // We need to mark all other uses in the current instruction too.
          for (unsigned OtherI = I + 1; OtherI != E; ++OtherI) {
            MachineOperand &Other = BI->getOperand(OtherI);
            if (Other.isReg() && Other.isUse() && Other.getReg() == Reg)
              NoDeferredLoadsPending.insert(&Other);
          }
        }
        DeferredRegs.erase(Reg);
      } else // Otherwise, the load is deferred.
        DeferredRegs.insert(Reg);

      RegistersHandled.insert(Reg);
    }
  }
}

static unsigned getOperandIndex(MachineOperand &Oper) {
  MachineInstr *MI = Oper.getParent();
  for (unsigned Idx = 0, E = MI->getNumOperands(); Idx != E; ++Idx)
    if (&Oper == &MI->getOperand(Idx))
      return Idx;

  assert(false && "Operand not found");
  return 0;
}

/// updateSpillsWithNewInstruction - Updates the vector with operands pointers
/// that point to the new instruction. Also updates the NoDeferredLoadsPending
/// set.
static void updateSpillsWithNewInstruction(MachineInstr *OldInstr,
                                MachineInstr *NewInstr,
                                SmallVector<MachineOperand*, 32>::iterator SI,
                                SmallVector<MachineOperand*, 32>::iterator SE,
                                DenseSet<MachineOperand *> &NoDeferred) {
  // Replace the old operand with the operand of the new instruction.
  for (++SI; SI != SE && (*SI)->getParent() == OldInstr; ++SI) {
    unsigned Index = getOperandIndex(**SI);
    *SI = &NewInstr->getOperand(Index);
  }
  // Update the NoDeferred set.
  for (DenseSet<MachineOperand *>::iterator DI = NoDeferred.begin(),
       DE = NoDeferred.end(); DI != DE; ++DI) {
    if ((*DI)->getParent() == OldInstr) {
      MachineOperand &Old = **DI;
      unsigned Index = getOperandIndex(Old);
      MachineOperand &New = NewInstr->getOperand(Index);
      NoDeferred.erase(&Old);
      NoDeferred.insert(&New);
    }
  }
}

void SpillHelper::emitSpillCode(MachineOperand &Oper,
                                SmallVector<MachineOperand*, 32>::iterator SI,
                                SmallVector<MachineOperand*, 32>::iterator SE) {
  unsigned OrigReg = Oper.getReg();
  const TargetRegisterClass *RC = MF->getRegInfo().getRegClass(OrigReg);
  int FrameIndex = getStackSpaceFor(OrigReg, RC);
  MachineBasicBlock::iterator InstrIter(Oper.getParent());

  // Fold memory operand if possible. It is possible if the operand is a
  // definition or if the load 'services' only this instruction.
  unsigned OperandIndex = getOperandIndex(Oper);
  SmallVector<unsigned,1 > FoldOps(1, OperandIndex);
  MachineInstr *FoldedInst = NULL;

  if (shouldFoldOperand(InstrIter, OperandIndex, Oper.isDef()) &&
      (Oper.isDef() || NoDeferredLoadsPending.count(&Oper)) &&
      (FoldedInst = TM->getInstrInfo()->foldMemoryOperand(InstrIter, FoldOps,
                                                          FrameIndex))) {
    // Delete the operand from the white list.
    NoDeferredLoadsPending.erase(&Oper);
    ++OperandsFolded;
    // After we created a new instruction we must update the spilled operands of
    // the old instruction in the vector of outstanding spills to refer to the
    // spilled operands of the new instruction.
    updateSpillsWithNewInstruction(InstrIter, FoldedInst, SI, SE,
                                   NoDeferredLoadsPending);

    // Delete the old instruction.
    InstrIter->removeFromParent();
    MF->DeleteMachineInstr(InstrIter);
    return;
  }

  if (Oper.isDef()) { // Spill to memory.
    // If the live range is rematerializable (at all uses) we don't need to
    // spill.
    if (isRematerializable(OrigReg))
      return;
    ++StoresAdded;
    // Emit code for spilling Oper to memory.
    TM->getInstrInfo()->storeRegToStackSlot(*Block, next(InstrIter),
                                            OrigReg, false,
                                            FrameIndex, RC, TRI);
  } else { // Spill from memory.
    assert(Oper.isUse());
    unsigned NewSpillReg =  MF->getRegInfo().createVirtualRegister(RC);

    // Emit code for spilling Oper from memory or rematerialize.
    MachineInstr *OrigInstr = isRematerializable(OrigReg);
    if (OrigInstr) { // Rematerialize.
      TM->getInstrInfo()->reMaterialize(*Block, InstrIter, NewSpillReg,
                                        0, OrigInstr, *TRI);
      ++NumRematerializedInsts;
    } else if (IsAvailDefBasedRematerializationEnabled &&
               Remat.isRematerializable(Oper)) { // Rematerialize.
      Remat.rematerialize(*Block, InstrIter, NewSpillReg, Oper, *TRI);
      ++NumAvailDefRematerializedInsts;
    } else { // Insert load from stack.
      ++LoadsAdded;
      TM->getInstrInfo()->loadRegFromStackSlot(*Block, InstrIter, NewSpillReg,
                                               FrameIndex, RC, TRI);
    }
    // Replace register name by new name.
    // This is buckin inefficient! We should'nt be scanning to the end.
#define LIMIT_REPLACEMENT 0
    for (MachineBasicBlock::iterator It = InstrIter, ItEnd = Block->end();
         It != ItEnd; ++It) {
      MachineInstr *Inst = It;
      bool SawUse = false;
      bool SawDef = false;
      // Process uses.
      for (unsigned Idx = 0, E = Inst->getNumOperands(); Idx != E; ++Idx) {
        MachineOperand &Op = Inst->getOperand(Idx);
        if (Op.isReg() && Op.getReg() == OrigReg && Op.isUse()) {
          SawUse = true;
          Op.setReg(NewSpillReg);
        }
      }
      // Process defs.
      for (unsigned Idx = 0, E = Inst->getNumOperands(); Idx != E; ++Idx) {
        MachineOperand &Op = Inst->getOperand(Idx);
        if (Op.isReg() && Op.getReg() == OrigReg && Op.isDef()) {
          SawDef = true;
#if LIMIT_REPLACEMENT
          // Stop at a new def that has no use of the same register.
          if (!SawUse) break;
#endif
          Op.setReg(NewSpillReg);
        }
      }
#if LIMIT_REPLACEMENT
      // Stop at a new def that has no use of the same register.
      if (SawDef && !SawUse) break;
#endif
    }
    assert(Oper.getReg() == NewSpillReg);
  }
}



void RAChaitinBriggs::spillVRs(std::set<unsigned>& VRsToSpill) {
  // Insert spill code.
  SpillHelper Spiller(Remat, &VRsToSpill, TRI, Builder);

  for (MachineFunction::iterator BI = MF->begin(), BE = MF->end(); BI != BE;
       ++BI) {
    // Set current basic block.
    Spiller.startProcessingBlock(&(*BI));
    // Spill.
    Spiller.spillRegisters();
  }
}

///
/// Allocate a physical register for each virtual register
///
void RAChaitinBriggs::mapVirtToPhysRegs() {
  // Now that the colors have been assigned:
  // For each VR, iterate through the alias set and choose a register
  // that fits
  // Insert that into the map
  for (InterferenceGraphNodeIterator graphIter = FunctionGraph->begin();
       graphIter != FunctionGraph->end();
       ++graphIter) {
    if ( ! FunctionGraph->isAPhysicalReg(*graphIter) ) {
      int AllocatedColor = (*graphIter)->Color;
      unsigned VirtReg = (*graphIter)->index;
      unsigned RepRegForColor = getPhysicalRegForColor(AllocatedColor);
      
      RegisterSet RegSetForRepReg = Builder->getRegistersOfRepresentative(RepRegForColor);
      
      //*** only consider virtual register here?
      const TargetRegisterClass *RegClass = MF->getRegInfo().getRegClass(VirtReg);
      
      //
      // Commenting out this and replacing with RegClass->begin() and RegClass->end()
      // This is necessary because if biased coloring decides to color a VR with
      // BP, SP, EBP or ESP, then the VR should be allocated to SP/ESP even if the function
      // dedicates EBP and ESP as frame/stack pointers. This is okay since the VR will
      // just be used and not redefined.
      // IMPORTANT: This assumes that PRs are not coalesced with VRs. If they are, then
      // the VR which is allocated to ESP/EBP may be redefined leading to wrong code
      // Consequently, in coalesceSSACopies and copyCoalesce, BP, SP, EBP, and ESP cannot
      // be coalesced with a VR!!
      //
      /*
       TargetRegisterClass::iterator RegClassIter = RegClass->allocation_order_begin(*MF);
       TargetRegisterClass::iterator RegClassEnd = RegClass->allocation_order_end(*MF);
       */
      
      TargetRegisterClass::iterator RegClassIter = RegClass->begin();
      TargetRegisterClass::iterator RegClassEnd = RegClass->end();
      
      // TODO: Add intersection here instead of iterating through the sets
      //
      
      // Iterate through the register class of the operand
      bool FoundReg = false;
      for (; (RegClassIter != RegClassEnd) && !FoundReg ; ++RegClassIter) {
        
        if ( RegSetForRepReg.find(*RegClassIter) != RegSetForRepReg.end()) {
          // Map the register for the node
          unsigned RepReg = Builder->getRepresentativeNumber(*RegClassIter);
          Virt2PhysRegMap[VirtReg] = RepReg;
          if ( DebugColoring ) {
            dbgs() << "mapping virtual: " << PrintReg(VirtReg) 
            << "   to physical: " << PrintReg(RepReg, TRI) << " (" << PrintReg(RepReg) << ")\n";
          }
          
          // Then map registers for all the nodes coalesced into
          // this one
          for (CoalescedNodesIterator coIter = (*graphIter)->getCoalescedNodesBegin();
               coIter != (*graphIter)->getCoalescedNodesEnd();
               ++coIter) {
            Virt2PhysRegMap[*coIter] = RepReg;
            if ( DebugColoring ) {
              dbgs() << "mapping virtual: " << PrintReg(VirtReg) << "(" << VirtReg << ")"
              << "   to physical: " << PrintReg(RepReg, TRI) << " (" << PrintReg(RepReg) << ")\n";
            }
          }
          
          FoundReg = true;
        }
        
      }
      
      // Debug
      if (! FoundReg) {
        if ( DebugColoring ) {
          dbgs() << "Could not find for register: " << (*graphIter)->index << "\n";
        }
      }
      assert(FoundReg && "Could not find Physical register for the VR");
      
    }
  }
}


///
/// Register assignment -- assign physical registers to virtual registers
///
void RAChaitinBriggs::changeVirtRegs() {
  MachineRegisterInfo &MRI = MF->getRegInfo();
  for (MachineFunction::iterator BlockIter = MF->begin();
       BlockIter != MF->end(); ++BlockIter) {
    
    // Some allocators may need to ignore certain blocks.
    if (blockNotInCurrentRegion(*BlockIter)) {
      continue;
    }
    
    for (MachineBasicBlock::iterator InstIter = BlockIter->begin();
         InstIter != BlockIter->end();
         ++InstIter) {
      
      MachineInstr* MI = InstIter;
      
      // Assign explicit register operands
      for (unsigned i = 0, e = MI->getNumOperands(); i != e; ++i) {
        
        MachineOperand& Oper = MI->getOperand(i);
        
        if ( Oper.isReg() &&
             Oper.getReg() != 0 &&
             TargetRegisterInfo::isVirtualRegister(Oper.getReg()) &&
             Builder->isCorrectRegisterType(Oper.getReg()) &&
             (Oper.getReg() != 0 /*is not NoRegister*/) ) {
          
          unsigned RegisterOp = Oper.getReg();
          unsigned subRegIndex = Oper.getSubReg();
          
          assert( (Virt2PhysRegMap.find(RegisterOp) != Virt2PhysRegMap.end()) &&
                 "Virtual register not allocated");
          
          unsigned AssignedPhysReg = Virt2PhysRegMap[RegisterOp];
          assert(Virt2PhysRegMap[RegisterOp] != 0);
          
#if 0
          // The RET instruction has no operand info defined, so the only
          // thing we can do is use the full register and hope for the best
          // TODO: What about non-X86 targets?

          if (MI->getOpcode() == X86::RET) {
            Oper.setReg(AssignedPhysReg);
            MF->getRegInfo().setPhysRegUsed(AssignedPhysReg);
            continue;
          }
#endif
          
          // Used to be: Pick up register class from instruction description.
          //const TargetInstrDesc& Desc = MF->getTarget().getInstrInfo()->get(MI->getOpcode());
          //const TargetRegisterClass* OpClass = Desc.getRegClass(i,TRI);
          //
          // Now: Use register class assigned when creating the register. This
          // way we handle uses of high 8 bit registers (e.g. AH) correctly.  In
          // the example below (from Benchmarks/Misc-C++/oopack_v1p8) we would
          // assign AL instead of AH because we were looking at the register
          // class of the TEST8ri instruction which is GR8. GR8 contains both AL
          // and AH, with AL comming first.
          //
          //    %vreg84<def> = COPY %vreg83:sub_8bit_hi<kill>; GR8_ABCD_H:%vreg84 GR64_ABCD:%vreg83
          //    TEST8ri %vreg84<kill>, 1, %EFLAGS<imp-def>; GR8_ABCD_H:%vreg84
          //
          // X86-64 A register in all of its glory:
          //
          //    |63..32|31..16|15-8|7-0|
          //                   |AH.|AL.|
          //                   |AX.....|
          //           |EAX............|
          //    |RAX...................|

          const TargetRegisterClass *OpClass = MRI.getRegClass(RegisterOp);
          
#if 0
          // Used to be: When using the instruction descriptor to pick up the
          // register class the following code needs to be renabled.

          if ( OpClass == NULL && MI->isCopy() ) {
            DEBUG(
                  dbgs() << "Assigning virtual register\n"
                  << "\tfor instruction: " << *MI 
                  << "\tfor operand: " << PrintReg(RegisterOp) << "\n"
                  << "\twith sug register index: " << subRegIndex << "\n";
                  );
            OpClass = MF->getRegInfo().getRegClass(RegisterOp);
          }
#endif
          
          // certain LLVM instructions do not have register classes defined
          // (e.g. EXTRACT_SUBREG's first operand can be from multiple classes)
          // We need to handle these cases specially.
          // See the documentation for llvm::TargetOpcode for more details.

          if ( OpClass == NULL ) {
            // Set Operand 2 to the correct subregister, and remove the
            // subregister index. Operands for SUBREG_TO_REG are not defined
            // anywhere, so this may not be correct in all cases
            if (MI->getOpcode() == TargetOpcode::SUBREG_TO_REG) {
              if (i == 2) {
                assert(MI->getNumOperands() > 3 && "Must have a least 4 operands");
                MachineOperand &SubRegOper = MI->getOperand(3);
                assert(SubRegOper.isImm() && "Must be an immediate value");
                subRegIndex = (unsigned)SubRegOper.getImm();
                assert(subRegIndex != 0 && "Must have a valid subregister index");
                AssignedPhysReg = TRI->getSubReg(AssignedPhysReg, subRegIndex);
                Oper.setSubReg(0);
              }
            }
            // Operand 0 should use Operand 2 as the subregister index.
            else if (MI->getOpcode() == TargetOpcode::EXTRACT_SUBREG) {
              if (i == 0) {
                AssignedPhysReg = TRI->getSubReg(AssignedPhysReg, MI->getOperand(2).getImm());
              }
            }
            // Operand 2 should use Operand 3 as the subregister index.
            else if (MI->getOpcode() == TargetOpcode::INSERT_SUBREG) {
              if (i == 2) {
                AssignedPhysReg = TRI->getSubReg(AssignedPhysReg, MI->getOperand(3).getImm());
              }
            }
            else assert(false && "Unhandled special instruction");
            
            Oper.setReg(AssignedPhysReg);
            MF->getRegInfo().setPhysRegUsed(AssignedPhysReg);
            
            continue;
          }
          
          // if we have a subregister index, use it to get the right register
          if (subRegIndex != 0) {
            AssignedPhysReg = TRI->getSubReg(AssignedPhysReg, subRegIndex);
            assert(AssignedPhysReg != 0);
#if 0
            DEBUG(dbgs() << "machine instruction was:\n\t" << *MI;);
#endif

            Oper.setReg(AssignedPhysReg);
            Oper.setSubReg(0);
            MF->getRegInfo().setPhysRegUsed(AssignedPhysReg);

#if 0
            DEBUG(dbgs() << "machine instruction becomes:\n\t" << *MI << "\n";);
#endif
          }
          else {
            // otherwise just loop through the subregisters and get the one
            // in the desired register class

            bool found = false;
            if ( OpClass->contains(AssignedPhysReg) ) {
              found = true;
            }
            else {
              // AssignedPhysReg now has a Representative Register which is a placeholder 
              // for a set of registers of varying sizes.
              //
              // OpClass is the target register class of the current operend.
              //
              // Need to get the right sized register from the Representative Register's associated set.
              //
              // HACK:
              // There could be multiple registers to match.  Chose the register the register with the
              // the lowest SubRegisterIndex verus the first match.  Needed this going from 2.9 to 3.0,
              // since the order of the Alias Set changed.  For example, given the color RAX and the
              // register class GR8, we want to chose AL over AH.  We use the SubRegisterIndex to sort
              // the two out.  THIS IS FRAGILE CODE.
              
              unsigned SmallestAlias;
              unsigned SmallestAliasIndex;

              for (const unsigned* AI = TRI->getAliasSet(AssignedPhysReg);
                   AI && *AI; ++AI) {
                unsigned AliasReg = *AI;
                bool ContainsAlias = OpClass->contains(AliasReg);

                if ( ContainsAlias ) {
                  if ( !found ) {
                    SmallestAlias = AliasReg;
                    SmallestAliasIndex = TRI->getSubRegIndex(AssignedPhysReg, AliasReg);
                    found = true;
                  }
                  else {
                    if ( SmallestAliasIndex > TRI->getSubRegIndex(AssignedPhysReg, AliasReg) ) {
                      SmallestAlias = AliasReg;
                      SmallestAliasIndex = TRI->getSubRegIndex(AssignedPhysReg, AliasReg);
                    }
                  }
                }
              }

              AssignedPhysReg = SmallestAlias;
            }
            assert(found);
            
            assert(AssignedPhysReg != 0);
            Oper.setReg(AssignedPhysReg);
            MF->getRegInfo().setPhysRegUsed(AssignedPhysReg);
          }
#if 0
          if ( DebugColoring ) {
            unsigned RepReg = Virt2PhysRegMap[RegisterOp];
            dbgs() << "Assigning virtual register " << PrintReg(RegisterOp) << "(" << RegisterOp << ")\n"
            << "\twith representative register " << PrintReg(RepReg) << "("
            << TRI->getName(RepReg) << ")\n"
            << "\tto    physical      register " << PrintReg(AssignedPhysReg) << "("
            << TRI->getName(AssignedPhysReg) << ")\n\n";
          }
#endif
        }
      }
    }
  }
}


///
/// spillNodes -- Mark interference nodes for spilling
/// Pre: all the nodes must contain virtual registers
///
void RAChaitinBriggs::spillNodes(std::set<InterferenceGraphNode*>& spillNodes) {
  std::set<unsigned> VRsToSpill;

  if ( DebugSpills ) {
    dbgs() << "SPILLING " << spillNodes.size() << " virtual registers\n";
  }

  for (std::set<InterferenceGraphNode*>::iterator snIter = spillNodes.begin();
       snIter != spillNodes.end();
       ++snIter) {
    
    InterferenceGraphNode *ignode = *snIter;
    assert(TargetRegisterInfo::isVirtualRegister(ignode->index) &&
           "Node contains Physical Register and trying to spill that");
    
    VRsToSpill.insert(ignode->index);
    if ( DebugSpills ) {
      dbgs() << "SPILLING virtual register: " << PrintReg(ignode->index) << "\n";
    }
    
    ++RegsSpilled;

    if (ignode->getSingleDef())
      ++NumRematerializableRegs;

    ProcRegsSpilled.increment(MF->getFunction()->getName());
    
  }

  spillVRs(VRsToSpill);
}



///
/// conduct copy coalescing
//
// Returns if any coalescing has occured
bool RAChaitinBriggs::copyCoalesce(bool ConservativeCoalescing /*= false*/) {

  //
  // Note on Conservative Colascing : Preston introduced conservative coalescing
  // with his agressive live range splitting.  Copies were used to split register
  // live ranges.  The original colaescing was agressive and would coalesce all of
  // these all copies, thus defeating the the splitting pass.  Preston made the
  // colaescer conservative so as to leave the splits needed for spilling.  Preston
  // found that some of the copies that were added did not coalesce or lead to spills.
  // Preston was not completely satusfued with his aggressive live range splitting
  // and did not recommend its implementation.  Thus, conservative coalescing is not 
  // recommended (with or without live range splitting).
  //

  if ( DebugCoalesce ) {
    dbgs() << "Before copy coalescing ...\n";
    FunctionGraph->dump(TRI, MF);
  }
  
  bool coalesced = false;
  
  bool AggressiveCoalescing = ( ! ConservativeCoalescing );
  
  bool CanCoalesceCopy;
  
  // Track pairs that were coalesced so we can update liveness.
  std::set<std::pair<unsigned, unsigned> > CoalescedPairs;
  
  for (MachineFunction::iterator BlockIter = MF->begin();
       BlockIter != MF->end(); ++BlockIter) {
    MachineBasicBlock *MBB = BlockIter;

    bool EraseThisInst = false;
    
    MachineBasicBlock::iterator InstIter    = MBB->begin();
    MachineBasicBlock::iterator InstIterEnd = MBB->end();
    while( InstIter != InstIterEnd ) {
      
      MachineInstr* MI = InstIter;
      
      // If it's a copy instruction
      InterferenceGraphNode* srcNode;
      InterferenceGraphNode* destNode;
      
      unsigned destReg;
      unsigned srcReg;
      
      unsigned SrcSubReg, DstSubReg;
      bool IsMove, CoalesceThisMove;
      
      CoalesceThisMove = false;

      // If its a move instr, ensure that both regs are VRs
      // and also that they are the correct type
      IsMove = isMoveInstr(TRI, MI, srcReg, destReg, SrcSubReg, DstSubReg);

      if ( IsMove &&
           Builder->isCorrectRegisterType(srcReg) &&
           Builder->isCorrectRegisterType(destReg) ) {

        srcNode  = FunctionGraph->getNode(Builder->getRepresentativeNumber(srcReg));
        destNode = FunctionGraph->getNode(Builder->getRepresentativeNumber(destReg));
        
        // One or both of these regs might have been coalesced.
        // Get their coalesced register numbers
        srcReg  = srcNode->index;
        destReg = destNode->index;
        
        bool MoveChangesPrecision;
        MoveChangesPrecision =  ( MI->isSubregToReg()) || ((MI->isCopy() && SrcSubReg != DstSubReg) );
        // Disable coalescing of different register classes. We don't want to
        // coalesce GR64_NOREX to GR64.
        //      %vreg12 = copy %vreg10 ; GR64_NOREX:%vreg12,GR64:%vreg10
        MachineRegisterInfo &MRI = MF->getRegInfo();
        bool IsMoveBetweenDifferentRegClasses =
          MI->isCopy() &&
          TargetRegisterInfo::isVirtualRegister(srcReg) &&
          TargetRegisterInfo::isVirtualRegister(destReg) &&
          MRI.getRegClass(srcReg) != MRI.getRegClass(destReg);

        if ( (! MoveChangesPrecision) && 
             (! IsMoveBetweenDifferentRegClasses) &&
            /*TODO: We disabled subregtoreg cast coalescing.
             * I believe we can coalesce the two registers but we need to make
             * sure that the coalesced registers are rewritten correctly
             * (subregisters should stay in their subregister class). 
             * We still attempt to bias color the Subreg and Reg. */
            (srcReg != destReg) &&
            TargetRegisterInfo::isVirtualRegister(srcReg) &&
            TargetRegisterInfo::isVirtualRegister(destReg)) {
          
          if ( DebugCoalesce ) {
            bool di = FunctionGraph->doesInterfere(destReg, srcReg);
            bool ws = FunctionGraph->willSpillCoalescedNodes(destReg, srcReg);
            dbgs() << "  Does Interfere: " << di
            << "  Aggressive: " << AggressiveCoalescing
            << "\t Will Spill: " << ws
            << "  srcReg = " << PrintReg(srcReg)
            << "  destReg = " << PrintReg(destReg)
            << "\n";
          }

          CanCoalesceCopy = false;
          if ( ( ! FunctionGraph->doesInterfere(destReg, srcReg) ) ) {
            // Source and Destination of Copy do not interfere.
            if ( AggressiveCoalescing ) {
              // coalesce without futher conditions, some may say" wreckless".
              CanCoalesceCopy = true;
            } else if ( ConservativeCoalescing &&
                        ( ! FunctionGraph->willSpillCoalescedNodes(destReg, srcReg) ) ) {
              // coalesce only when the new coalesce register will have < k neighbors
              CanCoalesceCopy = true;
            }
          }

          if ( CanCoalesceCopy ) {
                
            coalesced = true;
            CoalesceThisMove = true;
                
            if ( DebugCoalesce ) {
              dbgs() << "Coalescing dest(" << PrintReg(destReg) << ") with src(" << PrintReg(srcReg) << ")\n"; 
            }
            
            CoalescedPairs.insert(std::make_pair(destReg, FunctionGraph->getNode(srcReg)->index));
            
            FunctionGraph->coalesceNodes(destReg, srcReg);
            
            // Remove the redundant move instruction
            EraseThisInst = true;
            
            // If the registers have been coalesced, then should no
            // longer be in InfiniteSpillCostRegister
            InfiniteSpillCostRegisters.erase(destReg);
            InfiniteSpillCostRegisters.erase(srcReg);
          }
          else {
            // have move/copy instruction, no type conversion, two virtual registers, 
            // but alas interference
            if ( DebugCoalesce ) {
              dbgs() << "Can not coalesce " << PrintReg(destReg) << " with " << PrintReg(srcReg) << "\n";
            }
          }
        }
        else {
          // have move/copy instruction, but either the move/type is a type cast or
          // one of the registers is a physical regsiter
          if ( DebugCoalesce ) {
            dbgs() << "Can not coalesce ";
            if ( TargetRegisterInfo::isPhysicalRegister(destReg) )
              dbgs() << PrintReg(destReg) << "(" << PrintReg(destReg, TRI) << ")";
            else
              dbgs() << PrintReg(destReg);
            dbgs() << " with ";
            if ( TargetRegisterInfo::isPhysicalRegister(srcReg) )
              dbgs() << PrintReg(srcReg) << "(" << PrintReg(srcReg, TRI) << ")";
            else
              dbgs() << PrintReg(srcReg);
            dbgs() << "\n";
          }
        }
        
        srcNode  = FunctionGraph->getNode(Builder->getRepresentativeNumber(srcReg));
        destNode = FunctionGraph->getNode(Builder->getRepresentativeNumber(destReg));
        
        if ( srcReg != destReg && ! CoalesceThisMove ) { 
        
          // Update partner nodes. Placing this code here ensures that
          // copies of the form mov PhysicalReg -> VirtualReg are
          // also assigned to be partners

          srcNode->addPartnerNode(destNode);
          destNode->addPartnerNode(srcNode);
          if ( DebugCoalesce ) {
            dbgs() << "Making partners: ";
            if ( TargetRegisterInfo::isPhysicalRegister(destReg) )
              dbgs() << PrintReg(destReg) << "(" << PrintReg(destReg, TRI) << ")";
            else
              dbgs() << PrintReg(destReg);
            dbgs() << " and ";
            if ( TargetRegisterInfo::isPhysicalRegister(srcReg) )
              dbgs() << PrintReg(srcReg) << "(" << PrintReg(srcReg, TRI) << ")";
            else
              dbgs() << PrintReg(srcReg);
            dbgs() << "\n";
          }

          if (DebugCoalesce) {
            dbgs() << "\n";
          }
        }
        else {
          if ( DebugCoalesce ) {
            if ( CoalesceThisMove ) {
              dbgs() << "Can not partner after nodes are coalesced.\n\n";
            }
            else if ( srcReg == destReg ) {
              dbgs() << "Can not partner with itself: " << PrintReg(srcNode->index) << "\n\n";
            }
          }
        }
      }
      
      ++InstIter;
      
      if (EraseThisInst) {
        BlockIter->erase(prior(InstIter));
        
        EraseThisInst = false;
      }
    }
  }
  
  // Update the liveness information.
  if ( coalesced ) {
    for (MachineFunction::iterator BlockIter = MF->begin();
         BlockIter != MF->end(); ++BlockIter) {
      for (std::set<std::pair<unsigned, unsigned> >::iterator
           PI = CoalescedPairs.begin(), PE = CoalescedPairs.end();
           PI != PE; ++PI) {
        unsigned dest = PI->first;
        unsigned src  = PI->second;
        
        ValueSetType& livein  = Builder->getLiveBuilder()->GetLiveInSet()[BlockIter];
        ValueSetType& liveout = Builder->getLiveBuilder()->GetLiveOutSet()[BlockIter];
        
        if (livein.find(src) != livein.end()) {
          livein.erase(src);
          livein.insert(FunctionGraph->getNode(dest)->index);
        }
        if (liveout.find(src) != liveout.end()) {
          liveout.erase(src);
          liveout.insert(FunctionGraph->getNode(dest)->index);
        }
      } // end foreach pair
    } // end foreach BB
  }
  
  if ( DebugCoalesce ) {
    dbgs() << "After copy coalescing ...\n";
    if ( ! coalesced ) {
      dbgs() << "No copies coalesced (but partners could have been addd)\n";
    }
    else {
      FunctionGraph->dump(TRI, MF);
    }
  }

  return coalesced;
}


///
/// rewriteCoalescedRegs -- rewrite registers that have been coalesced to their
/// coalesced value.  Also, remove copies of the form mov x -> x This is
/// essential to minimize redundant spill code in case x is spilled
///
bool RAChaitinBriggs::rewriteCoalescedRegs() {
  
  bool changed = false;
  
  //const TargetInstrInfo& tii = *(MF->getTarget().getInstrInfo());
  
  unsigned srcReg, destReg;
  
  for (MachineFunction::iterator BlockIter = MF->begin();
       BlockIter != MF->end(); ++BlockIter) {
    
    bool EraseThisInst = false;
    
    MachineBasicBlock::iterator InstIter = BlockIter->begin();
    while (InstIter != BlockIter->end()) {
      MachineInstr *MI = &*InstIter;
      
      for (unsigned i = 0, e = InstIter->getNumOperands(); i != e; ++i) {
        
        MachineOperand& Oper = MI->getOperand(i);
        if (Oper.isReg() && TargetRegisterInfo::isVirtualRegister(Oper.getReg())) {
          
          // If the node is in the interference graph. The node might not be
          // in the interference graph if it is not the correct type for instance
          // (i.e. a floating point register in an integer interference graph)
          InterferenceGraphNode* coalescedParent = FunctionGraph->getNodeOrNull(Oper.getReg());
          
          // Avoid to assign 0 (i.e. NoRegister) to virtual register
          // Check if the coalesced parent is nullVal_
          if (coalescedParent && coalescedParent != FunctionGraph->getIGNullVal()
              && (coalescedParent->index != Oper.getReg())
              && coalescedParent->index != 0) {
            unsigned Reg;
            Reg = coalescedParent->index;
            Oper.setReg(Reg);
            changed = true;
          }
          
        }
      }
      
      unsigned SrcSubReg, DstSubReg;
      if (isMoveInstr(TRI, MI, srcReg, destReg, SrcSubReg, DstSubReg) &&
          TargetRegisterInfo::isVirtualRegister(srcReg) &&
          TargetRegisterInfo::isVirtualRegister(destReg) &&
          (srcReg == destReg)) {
        //	    BlockIter->erase(InstIter);
        EraseThisInst = true;
        changed = true;
      }
      
      ++InstIter;
      if (EraseThisInst) {
        BlockIter->erase(prior(InstIter));
        
        EraseThisInst = false;
      }
      
    }
  }
  
  return changed;
}


bool RAChaitinBriggs::coalesceSSACopies(std::vector<std::pair<unsigned, unsigned> >& SSACopyRegs) {
  
  bool coalesced = false;
  
  // Track pairs that were coalesced so we can update liveness.
  std::list<std::pair<unsigned, unsigned> > CoalescedPairs;
  
  for (std::vector<std::pair<unsigned, unsigned> >::const_iterator copyIter = SSACopyRegs.begin();
       copyIter != SSACopyRegs.end(); ++copyIter) {
    
    unsigned destReg = copyIter->first;
    unsigned srcReg = copyIter->second;
    
    if ( DebugCoalesce ) {
      
      dbgs() << "Looking at SSA copy: " << srcReg << " -> " << destReg << "\n";
      dbgs() <<  "SrcReg exists in graph: " << FunctionGraph->getNodeOrNull(srcReg) << "  "
      <<  "    DestReg? " << FunctionGraph->getNodeOrNull(destReg);
      if (FunctionGraph->getNodeOrNull(srcReg)) {
        dbgs()  <<  " is correct reg type: " << Builder->isCorrectRegisterType(srcReg);
      }
      
      if ( FunctionGraph->getNodeOrNull(srcReg) && FunctionGraph->getNodeOrNull(destReg)) {
        dbgs()
        <<  " is srcReg Virtual Register " << (TargetRegisterInfo::isVirtualRegister(srcReg)) << "   "
        << "  is destReg Virtual Register " << (TargetRegisterInfo::isVirtualRegister(destReg)) << "   "
        << "  does interfere: " << (FunctionGraph->doesInterfere(srcReg, destReg))
        << "\n";
      }
      else {
        dbgs() << "\n";
      }
    }	// DebugCoalesce
    
    
    
    // Since this method is in the build-coalesce-rewrite loop, some of the SSA registers
    // might already have been coalesced, check for that:
    
    if (FunctionGraph->getNodeOrNull(srcReg) && FunctionGraph->getNodeOrNull(destReg) ) {
      // Coalesce the registers if:
      // 1. The registers are the correct type for this interference graph
      // 2. The registers have not already been coalesced
      // 3. The registers are both VRs
      // 4. They do not interfere
      if (Builder->isCorrectRegisterType(srcReg) &&
          Builder->isCorrectRegisterType(destReg) &&
          (FunctionGraph->getNode(srcReg)->index != FunctionGraph->getNode(destReg)->index) &&
          TargetRegisterInfo::isVirtualRegister(srcReg) &&
          TargetRegisterInfo::isVirtualRegister(destReg) &&
          (!FunctionGraph->doesInterfere(srcReg, destReg)) ) {
        
        if ( DebugCoalesce ) {
          dbgs() << "SSA Coalescing: " << destReg
          << " with degree: " << FunctionGraph->getNode(destReg)->getDegree()
          <<" and " << srcReg << " with degree: " << FunctionGraph->getNode(srcReg)->getDegree()
          << "\n";
        } // DebugCoalesce
        
        CoalescedPairs.push_back(std::make_pair(FunctionGraph->getNode(destReg)->index, FunctionGraph->getNode(srcReg)->index));
        
        FunctionGraph->coalesceNodes(destReg, srcReg);
        coalesced = true;
        
        
        // If the registers have been coalesced, then should no
        // longer be in InfiniteSpillCostRegister
        InfiniteSpillCostRegisters.erase(destReg);
        InfiniteSpillCostRegisters.erase(srcReg);
        
        
        if ( DebugCoalesce ) {
          dbgs() << "Node1 was: " << destReg << " is " << FunctionGraph->getNode(destReg)->index
          << "    Node2 was " << srcReg << " is " << FunctionGraph->getNode(srcReg)->index
          << "\n";
          
          dbgs() << " Node1 now has degree: " << FunctionGraph->getNode(destReg)->getDegree()
          << " and Node2: " << FunctionGraph->getNode(srcReg)->getDegree() << "\n";
        } // DebugCoalesce
        
      }
    }
  }
  
  // Update the liveness information.
  if (coalesced) {
    for (MachineFunction::iterator BlockIter = MF->begin();
         BlockIter != MF->end(); ++BlockIter) {
      for (std::list<std::pair<unsigned, unsigned> >::iterator
           PI = CoalescedPairs.begin(), PE = CoalescedPairs.end();
           PI != PE; ++PI) {
        unsigned dest = PI->first;
        unsigned src = PI->second;
        ValueSetType& livein = Builder->getLiveBuilder()->GetLiveInSet()[BlockIter];
        ValueSetType& liveout = Builder->getLiveBuilder()->GetLiveOutSet()[BlockIter];
        if (livein.find(src) != livein.end()) {
          livein.erase(src);
          livein.insert(FunctionGraph->getNode(dest)->index);
        }
        if (liveout.find(src) != liveout.end()) {
          liveout.erase(src);
          liveout.insert(FunctionGraph->getNode(dest)->index);
        }
      } // end foreach pair
    } // end foreach BB
  }
  
  return coalesced;
}


void RAChaitinBriggs::resetAllStructures() {
  Color2PhysRegMap.clear();
  PhysReg2ColorMap.clear();
  assert(ColoringStack.empty() &&
         "Coloring stack needs to be empty at the begining of a function");
  Virt2PhysRegMap.clear();
  InfiniteSpillCostRegisters.clear();
}


void RAChaitinBriggs::CalcInterRegCopies(Statistic& StatsCounter) {
  //const TargetInstrInfo& tii = *(MF->getTarget().getInstrInfo());
  
  for (MachineFunction::iterator BlockIter = MF->begin();
       BlockIter != MF->end(); ++BlockIter) {
    
    for (MachineBasicBlock::iterator InstIter = BlockIter->begin();
         InstIter != BlockIter->end();
         ++InstIter) {
      
      unsigned srcReg, destReg, SrcSubReg, DstSubReg;

      if ( isMoveInstr(TRI, InstIter, srcReg, destReg, SrcSubReg, DstSubReg) ) {
        ++StatsCounter;
      }
    }
  }
}


bool RAChaitinBriggs::runOnMachineFunction(MachineFunction &F) {
  DidRun++;
  
  // 
  // Input: lower LLVM IR to mchine level, with instruction selected, out of SSA
  // 
  // Chaitin et al expect "getting the right number of names" as input as well.
  // This translates into partitioning the live ranges for a given lexical name
  // into disjoint sets and giving each set its own lexical name.  The prior
  // work at Rice accomplished this by tanslating into SSA, by partitioning the live
  // ranges via union-ing the connected phi functions, and by giving each partition
  // a different name when going out-of-SSA.
  //
  // For LLVM, our input is already out-of-SSA, either via regular or strong phi
  // elimination.  (See lib/CodeGen/{StrongPHIElimination,PHIElimination}.cpp)
  // During phi eliminaion, many copy instructions are inserted and no care was
  // given to "getting the right number of names".  Most of the exisitng LLVM
  // register allocators run an additional pre-register allocation pass to coalesce
  // registers, to clean up the copy instructions inserted during phi elimination.
  //
  // Register coalescing is an integral part of a Chaitin Briggs register allocator,
  // so LLVM's register coalescer is not run.  Thus, on input the name space will be 
  // more fractured than was seen during the prior work and the disjoint live ranges 
  // should not share the same names.  It is an open question what impact this will 
  // have on the Chaitin Briggs allocator.
  // 
  
  DEBUG(dbgs() << "********** CHAITIN BRIGGS REGISTER ALLOCATION **********\n"
        << "********** Function: "
        << ((Value*)F.getFunction())->getName() << '\n';
        F.dump(););
  
  MF = &F;
  
  TRI = F.getTarget().getRegisterInfo();

  MRI = &MF->getRegInfo();
  
  RCI.runOnMachineFunction(F);
  
  initPhysRegClasses(TRI);
  
  DEBUG(dbgs() << "Starting register allocator for MachineFunction: "
        << F.getFunction()->getName() << "\n");
  
  std::vector<std::pair<unsigned, unsigned> > SSACopyRegs;
  
  resetAllStructures();
  
  if (CalcIRCopies) {
    CalcInterRegCopies(InterRegCopiesBeforeAllocation);
  }
  
  // Get the loop information
  // CAUTION: This loop information becomes outdated if the CFG is modified
  // The CFG is not modified in this allocator and thus LoopInfo does not have
  // to be recalculated on every iteration
  
  // Collect the initial liveness information for the function.
  InitialLiveBuilder = new BuildLiveSets(F);
  InitialLiveBuilder->build(F);
  
  // Construct register classes
  ConstructRegisterClasses AllRegClasses(TRI, F);
  
  BlockDataMap LO;
  
  // Iterate over the register classes, building a graph for each class
  for (std::vector<std::set<const TargetRegisterClass*> >::iterator
       RCIter = AllRegClasses.RegisterClasses.begin();
       RCIter !=  AllRegClasses.RegisterClasses.end(); ++RCIter) {
    
    Builder = new BuildInterferenceGraph(&F, F.getTarget(), *RCIter, InitialLiveBuilder);
    SetupColors();
    Builder->buildExcludedRepresentativeRegs(Colors, Color2PhysRegMap, RCI);

    allocateAndAssign(F, SSACopyRegs);

    delete Builder; Builder = 0;
    
    InfiniteSpillCostRegisters.clear();
  }
  
  delete InitialLiveBuilder; InitialLiveBuilder = 0;
  
  if (CalcIRCopies) {
    CalcInterRegCopies(InterRegCopiesAfterAllocation);
  }
  
  // populate the LiveIn bit vector for each MachineBasicBlock, 
  // PhysRegs only needless to say.  Post RA LICM, wets itself
  // if LiveIns are not correct.
  populateMBBLiveIns(F);

  // Need to compute correct operand kill information for downstream passes.
  recomputeKilledOperands();

  DEBUG(dbgs() << "********** CHAITIN BRIGGS REGISTER ALLOCATION over **********\n"
        << "********** Function: "
        << ((Value*)F.getFunction())->getName() << '\n';
        F.dump(););

  return true;
}

struct CompareRegSetBySize {
  bool operator() (const std::pair<unsigned, int>& a, const std::pair<unsigned,
                   int>& b) {
    return a.second < b.second;
  }
};

static inline bool isCalleeSavedRegister(unsigned Reg, 
                                         MachineFunction *MF,
                                         const TargetRegisterInfo* TRI,
                                         BuildInterferenceGraph *Builder) {
  unsigned int i;
  const unsigned *CSRegs = TRI->getCalleeSavedRegs(MF);
  for (i=0; CSRegs[i] != 0; i++) {
    if ( Builder->getRepresentativeNumber(CSRegs[i]) == Builder->getRepresentativeNumber(Reg) ) {
      return true;
    }
  }
  return false;
}

struct CompareRegSetByCalleeSave {
  MachineFunction *MF;
  const TargetRegisterInfo* TRI;
  BuildInterferenceGraph *Builder;
  
  CompareRegSetByCalleeSave(MachineFunction *MF_, 
                            const TargetRegisterInfo* TRI_,
                            BuildInterferenceGraph *Builder_) 
  : MF(MF_), TRI(TRI_), Builder(Builder_) {};
  
  bool operator() (const std::pair<unsigned, int>& a, const std::pair<unsigned,
                   int>& b) {
    // is a before b?
    bool isACalleeSavedRegister = isCalleeSavedRegister(a.first, MF, TRI, Builder);
    bool isBCalleeSavedRegister = isCalleeSavedRegister(b.first, MF, TRI, Builder);
    return ( !isACalleeSavedRegister && isBCalleeSavedRegister );
  }
};

/// SetupColors
void RAChaitinBriggs::SetupColors() {
  int CurrentColor;
    
  Colors.clear();
  PhysReg2ColorMap.clear();
  Color2PhysRegMap.clear();
  
  CurrentColor = 0;
  for (RepresentativeNumbersIterator r = Builder->getRepresentativeNumbersBegin();
       r != Builder->getRepresentativeNumbersEnd(); ++r) {
    int RepReg = r->first;
    
    if ( DebugColoring && RCI.isReserved(RepReg) ) {
      dbgs() << PrintReg(RepReg, TRI) << " is reserved register.\n";
    }
    
    Colors.insert(CurrentColor);
    PhysReg2ColorMap[RepReg] = CurrentColor;
    Color2PhysRegMap[CurrentColor] = RepReg;
    ++CurrentColor;
  }
  
  //const TargetRegisterInfo* RegInfo = MF->getTarget().getRegisterInfo();
  // Setup ColorsByClass
  for (TargetRegisterInfo::regclass_iterator  classit = TRI->regclass_begin();
       classit != TRI->regclass_end(); ++classit) {
    
    const TargetRegisterClass *RegClass = *classit;
    
    for (TargetRegisterClass::iterator rit = RegClass->begin(); rit != RegClass->end(); ++rit) {
      
      unsigned Reg = *rit;
      
      if (! Builder->isCorrectRegisterType(Reg)) {
        continue;
      }
      
      unsigned RepReg = Builder->getRepresentativeNumber(Reg);
      ColorsByClass[RegClass].insert( getColorOfPhysicalReg(RepReg) );
      
      break;
    }
  }
  
  // Make an order of colors that is sensitive to the restrictiveness
  // of certain registers, i.e, on X86, choose SI first instead of AX.
  // The former does not have byte-sized sub-registers, whereas the
  // latter does.
  std::vector<std::pair<unsigned, int> > AllocationOrder;
  
  // Need to rebuild RegClassToAllocOrderMap for every MachineFunction
  // since allocation_order_begin() can change for every MachineFunction
  RegClassToAllocOrderMap.clear();
  
  for (TargetRegisterInfo::regclass_iterator classit = TRI->regclass_begin();
       classit != TRI->regclass_end(); ++classit) {
    
    const TargetRegisterClass *CurrRegClass = *classit;

    ArrayRef<unsigned> RawAlocationOrder = CurrRegClass->getRawAllocationOrder(*MF);
    ArrayRef<unsigned>::iterator RI = RawAlocationOrder.begin();
    ArrayRef<unsigned>::iterator RE = RawAlocationOrder.end();

    for (; RI != RE; ++RI) {
      
      if ( ! Builder->isCorrectRegisterType(*RI) ) {
        continue;
      }
      
      unsigned RepReg = Builder->getRepresentativeNumber(*RI);
      RegisterSet RegSetForRepReg = Builder->getRegistersOfRepresentative(RepReg);
      AllocationOrder.push_back(std::make_pair(*RI, RegSetForRepReg.size()));
    }
    
    std::sort(AllocationOrder.begin(), AllocationOrder.end(), CompareRegSetBySize());
    std::sort(AllocationOrder.begin(), AllocationOrder.end(), CompareRegSetByCalleeSave(MF, TRI, Builder));
    
    if ( DebugColoring ) {
      if ( AllocationOrder.size() > 0 ) {
        dbgs() << "Registers in sorted order for register class " << CurrRegClass->getName() << ":\n\t";
        for (std::vector<std::pair<unsigned, int> >::iterator
             I = AllocationOrder.begin(), E = AllocationOrder.end(); I != E; ++I) {
          unsigned RepReg = I->first;
          dbgs() << "(reg=" << PrintReg(RepReg, TRI) << ",size=" << I->second << ",callee=" 
          << (isCalleeSavedRegister(RepReg, MF, TRI, Builder)?"true":"false");
          if ( RCI.isReserved(RepReg) ) {
            dbgs() << ",RESERVED";
          }
          dbgs() << "), ";
        }
        dbgs() << "\n";
      }
    }
    
    for (std::vector<std::pair<unsigned, int> >::iterator
         I = AllocationOrder.begin(), E = AllocationOrder.end(); I != E; ++I) {
      unsigned RepReg = I->first;
      if ( ! RCI.isReserved(RepReg) ) {
        RegClassToAllocOrderMap[CurrRegClass].push_back(RepReg);
      }
    }
    
    AllocationOrder.clear();
    
  }

  return;
}


///
/// Main, driver function for register allocation and assignment
///
void RAChaitinBriggs::allocateAndAssign(MachineFunction& F,
                                        std::vector<std::pair<unsigned, unsigned> >& SSACopyRegs ) {
  
  std::set<InterferenceGraphNode*> nodesToSpill;
  
  Remat.setFunction(&F);

  ////////////////////////////////////////////////////////////////////////////
  //  Main graph coloring loop
  //
  //  The loop's pseudo-code:
  //
  //
  // spill = false
  // do {
  //
  //     if spill (if any node is marked white) then:
  //           insert spill code
  //
  //     Build interference graph
  //
  //     Coalesce nodes
  //
  //     first pre-color the physical registers
  //
  //     Calculate spill costs
  //
  //     remove trivially colorable nodes from the interference graph
  //     remove these nodes from the graph and push on stack
  //
  //     choose a spill candidate and push onto stack. Infinite-spill-cost registers
  //     and physical registers are not considered as spill candidates
  //
  //     Next, remove the infinite-spill-cost nodes and push onto the stack. These will
  //     be at the top of the stack and will thus be easily colorable
  //
  //     when the graph only consists of PR nodes
  //
  //        Pop a node from stack and insert into graph
  //        Try to color the node
  //        If it cannot be colored, then color it white
  //              and spill = true;
  //
  //
  // } while (!spill)
  ////////////////////////////////////////////////////////////////////////////
  
  bool NoRegistersToProcess = false;
  
  RAChaitinBriggs::AllocationIterations = 0;
  InterferenceGraph::SimplifyIterations = 0;
  
  nodesToSpill.clear();
  
  do {
    RAChaitinBriggs::AllocationIterations++;

    //
    // SPILL CODE
    //
    
    //
    // if any nodes marked white then insert spill code
    //
    if ( ! nodesToSpill.empty() ) {
      if ( DebugSpills ) {
        dbgs() << "\nBefore spill: " << "\n";
        dbgs() << "\n";
      }

      spillNodes(nodesToSpill);
      
      if ( DebugSpills ) {
        dbgs() << "\nAfter spill: " << "\n";
        dbgs() << "\n";
      }
    }

    // Clear rematerialization info.
    Remat.clear();

    nodesToSpill.clear();
    
    DEBUG(dbgs() << "\n\nStarting Register Allocation Loop, Iteration #" 
          << RAChaitinBriggs::AllocationIterations << ", function " 
          << F.getFunction()->getName() << " for this register class.\n\n\n";);
    
    //
    // BUILD and COALESCE
    //
    
    //
    // Coalesce-build iteration
    //
    doBuildAndCoalesceLoop(F, SSACopyRegs);

    if ( DebugCoalesce ) {
      dbgs() << "\n";
    }
    
    // first pre-color the physical registers
    if ( DebugColoring ) {
      dbgs() << "\nColoring all of the Physical Registers.\n";
    }
    colorPhysRegs();
    
    //
    // SPILL COSTS
    //
    
    // Deviating from the Chaitin Briggs Allocator Algorithm by calculating
    // spill costs JIT before they are needed and after the first round of
    // removing trivially colorable nodes.
    //
    // The advantages are that
    //
    // 1. Some sort comparators cannot handle nodes with degree = 0.
    //
    // 2. This is an efficiency hack since that there is less spill costs to calculate.
        
    //
    // SIMPLIFY
    //
    
    // Until the graph only consists of PRs and infinite cost nodes.
    // Remove all virtual register nodes with degree < K
    // K varies with the Register class of the node
    //

    unsigned NumVirtualRegiters = 0;
    unsigned NumVirtualRegitersWithBiases = 0;
    unsigned NumBiases = 0;
    unsigned NumOptimisticVirtualRegiters = 0;
    
    if ( DebugColoring ) {
      for (InterferenceGraphNodeIterator GraphNodeIter = FunctionGraph->begin();
           GraphNodeIter != FunctionGraph->end(); ++GraphNodeIter) {
        InterferenceGraphNode *IGNode = *GraphNodeIter;
        if ( TargetRegisterInfo::isVirtualRegister(IGNode->index) ) {
          NumVirtualRegiters++;
          if ( IGNode->getNumPartners() > 0 ) {
            NumVirtualRegitersWithBiases++;
            NumBiases += IGNode->getNumPartners();
          }
        }
        //assert();
        //BuildInterferenceGraph::isCorrectRegisterType()
      }
    } 
    
    bool Finished = false;
    bool FirstTime = true;

    InfiniteSpillCostRegisters.clear();
    
    // Logically virtual registers can live in three places:
    //   1. Interference Graph
    //   2. Coloring Stack
    //   3. Infinite Spill Cost set
    //
    // When a VR is trivilly colorable, the VR is moved from the IG to the Coloring Stack.
    //
    // When spill costs are calculated (first per Simplify loop), VRs which can infinite
    // spill costs are removed from the IG and put in their own set.  Since spill registers
    // are only selected from the IG, infinite spill cost VRs will never be selected for
    // spilling.

    while ( true ) { // till the graph only contains PRs
      
      InterferenceGraph::SimplifyIterations++;
      
      if ( DebugColoring ) {
        dbgs() << "\n\nStarting Simplify Round #" << InterferenceGraph::SimplifyIterations << "\n\n\n";
      }

      // Remove all trivially colorable VRs from IG and InfiniteSpillCostRegisters set
      // and place them on the ColoringStack.

      unsigned NumTriviallyColoringNodes = 0;
      NumTriviallyColoringNodes = FunctionGraph->removeTriviallyColorableNodes(MF, 
                                                                               &InfiniteSpillCostRegisters, 
                                                                               &ColoringStack);

      if ( DebugColoring ) {
        dbgs() << "Round #" << InterferenceGraph::SimplifyIterations 
        << " -- Pushing " << NumTriviallyColoringNodes << " trivially colored nodes.\n";
      }
      
      if ( ! FunctionGraph->isEmpty() ) {

        if ( FirstTime ) {
          // Calculate spill costs and rematerialization here, JIT.
          
          if ( DebugSpills ) {
            dbgs() << "\nCalculating spill costs.\n\n";
          }

          // Compute rematerialization opportunity.

          if ( IsAvailDefBasedRematerializationEnabled ) {
            // TODO: Fugly. Make stack a vector so that we can access is
            // elements.
            std::stack<InterferenceGraphNode *> CopyOfStack(ColoringStack);
            DenseSet<unsigned> GuaranteedColoredRegs;

            // Fill set.
            while(!CopyOfStack.empty()) {
              InterferenceGraphNode* CurrNode = CopyOfStack.top();
              CopyOfStack.pop();
              GuaranteedColoredRegs.insert(CurrNode->index);
            }

            Remat.findOpportunities(GuaranteedColoredRegs);
          }

          // Calculate spill costs.
          {
            // Compute spill costs.
            MachineLoopInfo& LI = getAnalysis<MachineLoopInfo>();
            InfiniteSpillCostRegisters.clear();

            SpillHelper Spiller(Remat, &InfiniteSpillCostRegisters, TRI,
                                Builder, MF, &LI);

            // Process basic blocks.
            for (MachineFunction::iterator BI = MF->begin(), BE = MF->end();
                 BI != BE; ++BI) {
              Spiller.startProcessingBlock(&(*BI));
              Spiller.spillRegisters();
            }
            FunctionGraph->calcSpillCosts(InfiniteSpillCostRegisters);
          }
          
          FirstTime = false;
        }
        
        InterferenceGraphNode* spillCandidate;
        
        // Choose spill candidates and pushing those nodes on the
        // stack.  Infinite-spill-cost registers will NOT be chosen
        
        do {
          spillCandidate = FunctionGraph->getVirtualRegisterNodeWithMostDegree(InfiniteSpillCostRegisters);
        } while ( spillCandidate && FunctionGraph->isRemovedNode(spillCandidate->index) );
        
        if ( spillCandidate ) {
          
          if ( DebugSpills ) {
            dbgs() << "Optimistically coloring node: " << PrintReg(spillCandidate->index) 
            << " and pushing on stack.\n";
            spillCandidate->dump(TRI, MF, FunctionGraph, true);
          }
          
          pushNodeOptimistically(spillCandidate);

          NumOptimisticVirtualRegiters++;
        }
        else {
          // there is no spill candidate
          Finished = true;
        }
        
      }
      else {
        // Interference Graph is empty
      }


      if ( DebugColoring ) {
        dbgs() << "\n\nEnding Simplify Round #" << InterferenceGraph::SimplifyIterations << "\n\n\n";
      }

      // we are done if
      //   were have remove all of the node from the Interfernece Graph or
      //   the Interfernece Graph only contains nodes with inifinte spill costs
      
      if ( FunctionGraph->isEmpty() || Finished ) break;
      
    } // while

#if 1
    // "nodes with infinite spill costs that are not trivially colorable" should never happen.
    // adding an assert here instead.

    // Now the graph contains only PR nodes and nodes with infinite spill costs
    // that are not trivially colorable.
    //
    // Remove the ones with infinite spill cost that are not trivially colorable.
    //
    // Push them on top of the stack -- these will be the first nodes encountered
    // while coloring and hence will be colored.
    //
    // Note that pre-colored nodes will be ignored here.
    for (std::set<unsigned>::iterator infIter = InfiniteSpillCostRegisters.begin();
         infIter != InfiniteSpillCostRegisters.end();
         ++infIter) {
      InterferenceGraphNode* IGNode;
      bool isInfiniteSpillRegInInterferenceGraph;

      IGNode = FunctionGraph->getNodeIfInGraph(*infIter);
      isInfiniteSpillRegInInterferenceGraph = ( IGNode != FunctionGraph->getIGNullVal() );
      if ( isInfiniteSpillRegInInterferenceGraph ) {

        // infinit spill  registers that are not pre-colored (not currently done).
        // below test is always true.
        if  ( IGNode->PreColorIndex == InterferenceGraphNode::NO_PRECOLOR_INDEX ) {
        
          if (DebugSpills) {
            dbgs() << "Pushing node: " << PrintReg(IGNode->index)
            << " (infinite cost)" << "\n";
          }
          
          assert((false) && 
                 "nodes with infinite spill costs that are not trivially colorable have been encountered." );

          ColoringStack.push(IGNode);
          FunctionGraph->removeNode(IGNode);
        }
      }
    }
#endif

    // Now, the interference graph is empty and the coloring stack
    // contains all the nodes in ascending order of
    // spill preference (the deeper in the stack, the more likely it is that
    // the node will be spilled)
    //
    // The PRs are the only nodes left in the graph. They should remain in
    // the graph because:
    //   The PRs are pre-colored and are thus constrained in their choice of color
    //   Any VR node that interferes with a PR (and thus is its neighbor) cannot be
    //   given the color taken by the PR
    
    //
    // SELECT
    //
    
    ColorStats_T ColorStats;
    
    while ( ! ColoringStack.empty() ) {
      InterferenceGraphNode* currentStackNode = ColoringStack.top();
      ColoringStack.pop();
      
      // Add back to the graph
      FunctionGraph->addNodeBackToGraph(currentStackNode);
      
      // If it cannot be assigned a color, then
      // insert into nodesToSpill
      if ( ! colorNode(currentStackNode, &ColorStats) ) {
          nodesToSpill.insert(currentStackNode);
      }
    }
    
    if ( DebugColoring ) {
      dbgs() << "\n\nSELECT -- attempt to assign colors.\n\n";
      if ( NumVirtualRegiters > 0 ) {
        int NumSpilledRegisters = nodesToSpill.size();
        dbgs() << NumVirtualRegiters << " virtual registers.\n";
        dbgs() << NumOptimisticVirtualRegiters << " optimistic virtual registers.\n";
        if ( NumOptimisticVirtualRegiters > 0 ) {
          dbgs() << NumSpilledRegisters << " spilled virtual registers.\n";
          dbgs() << unsigned( ((NumOptimisticVirtualRegiters - NumSpilledRegisters)*100)/NumOptimisticVirtualRegiters ) 
          << "% optimistic virtual registers colored.\n";
        }
        dbgs() << NumVirtualRegitersWithBiases << " virtual registers with biases.\n";
        if ( NumVirtualRegitersWithBiases > 0 ) {
          dbgs() << ColorStats.NumBiasesAssigned << " virtual registers colored with bias.\n";
          dbgs() << unsigned( (ColorStats.NumBiasesAssigned * 100)/NumVirtualRegitersWithBiases )
          << "% biased virtual registers that get biased color.\n";
        }
        dbgs() << NumBiases << " total number of biases.\n\n";
      }
    }
    
  } while (! nodesToSpill.empty() );
  
  
  if ( ! NoRegistersToProcess ) {
    
    if ( DebugColoring ) {
      dbgs() << "\n\nRewrite the virtual registers in the machine code.\n\n";
    }

    // Allocation is over -- now assign
    mapAndChangeVirtRegs();
    
    // Some allocators need to post-process the colored graph.
    summarizeInterferenceGraph(Builder);
  }
}


///
/// Perform graph building and coalescing.
///
void RAChaitinBriggs::doBuildAndCoalesceLoop(MachineFunction& F, 
                                             std::vector<std::pair<unsigned, unsigned> >& SSACopyRegs) {
  //
  // Coalesce-build iteration
  //
  bool coalesced;
  do {

    coalesced = false;

    // Redo live information.
    assert(InitialLiveBuilder != NULL);
    delete InitialLiveBuilder; InitialLiveBuilder = 0;
    InitialLiveBuilder = new BuildLiveSets(*MF);
    InitialLiveBuilder->build(*MF);
    Builder->LiveBuilder = InitialLiveBuilder;

    // Build the interference graph
    Builder->build(F);

    FunctionGraph = Builder->getInterferenceGraph();

    if (SSACoalescing) {
      coalesced = coalesceSSACopies(SSACopyRegs);
      rewriteCoalescedRegs();
    }
    
    // Conduct copy coalescing
    bool couldCopyCoalesce = copyCoalesce();
    
    bool DidRewrite = rewriteCoalescedRegs();
    
    coalesced = couldCopyCoalesce || coalesced;
    if (coalesced == false) {
      assert(DidRewrite == false);
    }
  } while (coalesced);
  
}


///
/// Map virtual regs to physical, and do assignment.
///
void RAChaitinBriggs::mapAndChangeVirtRegs() {
  mapVirtToPhysRegs();
  changeVirtRegs();
}


///
/// Clear state after each MachineFunction
///
void RAChaitinBriggs::releaseMemory() {
  
  Colors.clear();
  Color2PhysRegMap.clear();
  PhysReg2ColorMap.clear();
  Virt2PhysRegMap.clear();
}


void RAChaitinBriggs::SanityCheck() {
  for (InterferenceGraphNodeIterator graphIter = FunctionGraph->begin();
       graphIter != FunctionGraph->end();
       ++graphIter) {
    
    int mycolor = (*graphIter)->Color;
    DEBUG(dbgs() << "Node is: " << (*graphIter)->index << "\n");
    DEBUG(dbgs() << "Color is " << mycolor << "\n");
    unsigned pr1 = getPhysicalRegForColor(mycolor);
    
    for ( InterferenceGraphNodeIterator niter = (*graphIter)->getNeighborsBegin();
         niter != (*graphIter)->getNeighborsEnd();
         ++niter) {
      InterferenceGraphNode *NeighborNode = *niter;

      DEBUG(dbgs() << "nodes " << (*graphIter)->index << "  and  " << NeighborNode->index << "\n");
      
      DEBUG(dbgs() << "pr2 Color is "<< NeighborNode->Color << "\n");
      unsigned pr2 = getPhysicalRegForColor(NeighborNode->Color);
      
      if ( NeighborNode->Color == mycolor) {
        assert(!"Adjacent nodes have the same color");
      }
      
      if (pr1 == pr2) {
        assert(!"Two adjacent nodes assigned the same register\n");
      }
      
    }
    
  }
}


/// recomputeKilledOperands - Computes 'kill' operands. This function relies on
/// up to date liveness information in InitialLiveBuilder (populateMBBLiveIns
/// has just computed it).
/// Also computes dead definitions.
void RAChaitinBriggs::recomputeKilledOperands() {
  assert(InitialLiveBuilder != NULL);
  for (MachineFunction::iterator BB = MF->begin(), BE = MF->end(); BB != BE;
       ++BB) {
    // Every register that is live out is not killed (until the next def of the
    // same register is encountered in the current basic block).
    //   ...
    //   EAX = ...
    //       = EAX <set killed>
    //   EAX = ...
    //       = EAX <dont set killed because of live out>
    //   LiveOut = { EAX, ...}
    std::set<unsigned> NotKilled = InitialLiveBuilder->GetLiveOutSet()[BB];

    for (MachineBasicBlock::reverse_iterator I = BB->rbegin(), E = BB->rend();
         I != E; ++I) {
      MachineInstr *Inst = &*I;
      // Process defs.
      for (unsigned Idx = 0, End = Inst->getNumOperands(); Idx != End; ++Idx) {
        MachineOperand &Oper = Inst->getOperand(Idx);
        // Ignore non-register operands.
        if (!Oper.isReg() || Oper.getReg() == 0) continue;
        unsigned Reg = Oper.getReg();
        if (Oper.isDef() && NotKilled.count(Reg)) {
            NotKilled.erase(Reg);
            for (const unsigned* AI = TRI->getAliasSet(Reg); AI && *AI; ++AI)
              if (NotKilled.count(*AI))
                NotKilled.erase(*AI);
        }
      }
      // Process uses. Last use in a basic block that is not live out is a kill.
      // An instruction could have two uses of a register that is killed by the
      // instruction. That is why we process killed register after we visited
      // all instructions.
      SmallVector<unsigned, 8> KilledByInstr;
      for (unsigned Idx = 0, End = Inst->getNumOperands(); Idx != End; ++Idx) {
        MachineOperand &Oper = Inst->getOperand(Idx);
        // Ignore non-register operands.
        if (!Oper.isReg() || Oper.getReg() == 0 || !Oper.isUse()) continue;

        unsigned Reg = Oper.getReg();
        if (!NotKilled.count(Reg)) {
          Oper.setIsKill();
          KilledByInstr.push_back(Reg);
        } else {
          Oper.setIsKill(false);
        }
      }

      // Add killed registers and aliased registers to NotKilled set.
      for (unsigned Idx = 0, End = KilledByInstr.size(); Idx != End; ++Idx) {
        unsigned Reg = KilledByInstr[Idx];
        NotKilled.insert(Reg);
        for (const unsigned* AI = TRI->getAliasSet(Reg); AI && *AI; ++AI) {
          unsigned AliasedReg = *AI;
          NotKilled.insert(AliasedReg);
        }
      }
    }
  }
  // Compute dead definitions.
  for (MachineFunction::iterator BB = MF->begin(), BE = MF->end(); BB != BE;
       ++BB) {
    std::set<unsigned> UseExists = InitialLiveBuilder->GetLiveOutSet()[BB];
    for (MachineBasicBlock::reverse_iterator II = BB->rbegin(), IE = BB->rend();
         II != IE; ++II) {
      unsigned NumOperands = II->getNumOperands();
      // Handle definitions.
      for (unsigned I = 0; I != NumOperands; ++I) {
        MachineOperand &Oper = II->getOperand(I);
        // Ignore non registers and uses.
        if (!Oper.isReg() || !Oper.getReg() || !Oper.isDef())
          continue;
        unsigned DefReg = Oper.getReg();
        if (UseExists.count(DefReg)) {
          UseExists.erase(DefReg);
          Oper.setIsDead(false);
        } else
          Oper.setIsDead();
      }
      // Handle uses.
      for (unsigned I = 0; I != NumOperands; ++I) {
        MachineOperand &Oper = II->getOperand(I);
        // Ignore non registers and definitions.
        if (!Oper.isReg() || !Oper.getReg() || !Oper.isUse())
          continue;
        UseExists.insert(Oper.getReg());
      }
    }
  }
}


void RAChaitinBriggs::populateMBBLiveIns(MachineFunction &MF) {
  MachineFunction::iterator MBBiter;
  MachineFunction::iterator MBBend;
  MachineBasicBlock *MBB;
  BlockDataMap LiveIn;
  ValueSetType::iterator RegisterIter;
  unsigned PhysReg;
  
  if ( PrintLiveInonExit ) {
    MBBiter = MF.begin();
    MBBend  = MF.end();
    for (; MBBiter != MBBend; ++MBBiter) {
      MBB = MBBiter;
      MachineBasicBlock::livein_iterator LIiter;
      MachineBasicBlock::livein_iterator LIend;
      LIiter = MBB->livein_begin();
      LIend  = MBB->livein_end();
      dbgs() << "Initial LiveIns for basic block: " << MBB->getName() << "\n\t";
      for (;LIiter != LIend; ++LIiter) {
        PhysReg = *LIiter;
        dbgs() << TRI->getName(PhysReg) << "(" << PhysReg << "), ";
      }
      dbgs() << "\n";
    }
  }
  
  //
  // Repopulate the Chaitin Briggs LiveIn and LiveOut objects
  //
  
  delete InitialLiveBuilder; InitialLiveBuilder = 0;
  InitialLiveBuilder = new BuildLiveSets(MF);
  InitialLiveBuilder->build(MF);
  LiveIn = InitialLiveBuilder->GetLiveInSet();
  
  //
  // Chaitin Briggs created thier own internal LiveIn and LiveOut objects,
  // which will be used to populate LLVM's MachineBasicBlock LiveIns field
  //
  
  MBBiter = MF.begin();
  MBBend  = MF.end();
  for (; MBBiter != MBBend; ++MBBiter) {
    MBB = MBBiter;
    ValueSetType BBLiveIn = LiveIn[MBB];
    for (RegisterIter = BBLiveIn.begin(); RegisterIter != BBLiveIn.end(); ++RegisterIter) {
      PhysReg = *RegisterIter;
      if ( ! MBB->isLiveIn(PhysReg) ) {
        MBB->addLiveIn(PhysReg);
      }
    }
  }

  assert(InitialLiveBuilder != NULL &&
         "recomputeKilledOperands depends on current liveness information");

  if ( PrintLiveInonExit ) {
    MBBiter = MF.begin();
    MBBend  = MF.end();
    for (; MBBiter != MBBend; ++MBBiter) {
      MBB = MBBiter;
      MachineBasicBlock::livein_iterator LIiter;
      MachineBasicBlock::livein_iterator LIend;
      LIiter = MBB->livein_begin();
      LIend  = MBB->livein_end();
      dbgs() << "Post Register Allocation LiveIns for basic block: " << MBB->getName() << "\n\t";
      for (;LIiter != LIend; ++LIiter) {
        PhysReg = *LIiter;
        dbgs() << TRI->getName(PhysReg) << "(" << PhysReg << "), ";
      }
      dbgs() << "\n";
    }
  }
}


void RAChaitinBriggs::pushNodeOptimistically(InterferenceGraphNode *IGNode) {
  InterferenceGraphNodeIterator NeighborIterI;
  InterferenceGraphNodeIterator NeighborIterJ;
  InterferenceGraphNode *NeighborI;
  InterferenceGraphNode *NeighborJ;
  
  // Want to bias neighbors that do not interfere - O(n2).
  // This is an attempt to get IGNode's neighbors to use the same color(s)
  // and thus reduce IGNode's degree and make IGNode more colorable,
  // when popped from the coloring stack.

  // when IGNode is optimistically pushed onto the coloring stack, its reamaimning
  // neighbors are still in the IG and not on the stack.  when IGNode is popped from
  // the coloring stack, this set of neighbors has already been popped.  when this set
  // of neighbors are popped, we are trying to assign them colors that give IGNode
  // a beter chance to get colored and NOT get spilled.
  
  if ( BiasedTim ) {
    if ( DebugSpills ) {
      dbgs() << "Attempting to add biases to the optimistically pushed node's neighbors:\n";
    }

    // for I = 1 to N-1
    NeighborIterI = IGNode->getNeighborsBegin();
    for (; NeighborIterI != IGNode->getNeighborsEnd(); ++NeighborIterI) {
      NeighborI = *NeighborIterI;
      if ( NeighborIterI == prior(IGNode->getNeighborsEnd()) ) break;
      
      if ( TargetRegisterInfo::isPhysicalRegister(NeighborI->index) ) continue;
      if ( FunctionGraph->isRemovedNode(NeighborI->index) ) continue;

      // for J = I+1 to N
      NeighborIterJ = NeighborIterI;
      ++NeighborIterJ;
      for (; NeighborIterJ != IGNode->getNeighborsEnd(); ++NeighborIterJ) {
        NeighborJ = *NeighborIterJ;

        if ( TargetRegisterInfo::isPhysicalRegister(NeighborJ->index) ) continue;
        if ( FunctionGraph->isRemovedNode(NeighborJ->index) ) continue;
        if ( NeighborI->index == NeighborJ->index ) continue;
#if 0
        if ( DebugSpills ) {
          dbgs() << "\tDo " << PrintReg(NeighborI->index) << " and " << PrintReg(NeighborJ->index) 
          << " interfer? ";
          if ( FunctionGraph->doesInterfere(NeighborI->index, NeighborJ->index) ) dbgs() << "yes";
          else                                                                    dbgs() << "no";
          dbgs() << "\n";
        }
#endif
        if ( ! FunctionGraph->doesInterfere(NeighborI->index, NeighborJ->index) ) {
          NeighborI->addPartnerNode(NeighborJ);
          NeighborJ->addPartnerNode(NeighborI);
          
          if ( DebugSpills ) {
            dbgs() << "\tMaking " << PrintReg(NeighborI->index) << " and " << PrintReg(NeighborJ->index) 
            << " partners.\n";
          }
        }
      }
    }
  }
  
  // remove from IG and push onto the Color Stack
  
  ColoringStack.push(IGNode);
  IGNode->OptimisticallyPushed = true;
  FunctionGraph->removeNode(IGNode);
}


