//===----------------------------------------------------------------------===//
//This software was produced with support from the Defense Advanced
//Research Projects Agency (DARPA) through AFRL Contract FA8650-09-C-1915. 
//Nothing in this work should be construed as reflecting the official policy or
//position of the Defense Department, the United States government, or
//Rice University.
//
//THIS SOFTWARE HAS BEEN APPROVED FOR PUBLIC RELEASE, UNLIMITED DISTRIBUTION.
//
//===----------------------------------------------------------------------===//

#include "Rematerialization.h"
#include "AvailDefinitions.h"

#include "llvm/CodeGen/MachineBasicBlock.h"
#include "llvm/CodeGen/MachineFunction.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"
#include <cassert>


void Rematerialization::clear() {
  // Clear AvailDef.
  if (AvailDef != NULL) {
    delete AvailDef;
    AvailDef = NULL;
  }
  RematCandidates.clear();
}

Rematerialization::~Rematerialization() {
  if (AvailDef)
    delete AvailDef;
}

void Rematerialization::setFunction(MachineFunction *F) {
  MF = F;
  TII = MF->getTarget().getInstrInfo();
}

bool Rematerialization::isRematerializable(MachineOperand &Oper) {
  assert(Oper.isUse());
  return RematCandidates.find(&Oper) != RematCandidates.end();
}

MachineInstr * Rematerialization::getRematerializableInst(MachineOperand &Op) {
  assert(Op.isUse());
  DenseMap<MachineOperand *, MachineInstr *>::iterator I =
    RematCandidates.find(&Op);
  if (I == RematCandidates.end()) return NULL;

  return I->second;
}

void Rematerialization::rematerialize(MachineBasicBlock &BB,
                                      MachineBasicBlock::iterator &InsertBefore,
                                      unsigned NewReg,
                                      MachineOperand &UseToSpill,
                                      const TargetRegisterInfo &TRI) {
  assert(isRematerializable(UseToSpill));

  MachineInstr *OrigDefiningInst = RematCandidates[&UseToSpill];
  MachineInstr *NewInst = MF->CloneMachineInstr(OrigDefiningInst);
  unsigned SubRegIndex =  NewInst->getOperand(0).getSubReg();
  NewInst->substituteRegister(NewInst->getOperand(0).getReg(), NewReg,
                              SubRegIndex, TRI);
  BB.insert(InsertBefore, NewInst);
}

void Rematerialization::findOpportunities(DenseSet<unsigned> &ColoredRegs) {
  computeAvailDefs();
  for (MachineFunction::iterator BB = MF->begin(), BE = MF->end(); BB != BE;
       ++BB) {
    findOpportunities(BB, ColoredRegs);
  }
#if 0 
  errs() << "\nRematerialization info: \n";
  MF->dump();
  errs() << "Number of opportunities: " << RematCandidates.size() << "\n";
  for (DenseMap<MachineOperand *, MachineInstr *>::iterator RI =
       RematCandidates.begin(), RE = RematCandidates.end(); RI != RE; ++RI) {
    errs() << *RI->first << ": " << *RI->second;
    errs() << "   in " << *RI->first->getParent();
  }
  errs() << "\n";
#endif
}

static bool doesDefineOperand(MachineInstr *Inst, unsigned Reg) {
  for (unsigned I = 0, E = Inst->getNumOperands(); I != E; ++I) {
    MachineOperand &Oper = Inst->getOperand(I);
    if (!Oper.isReg() || !Oper.getReg() || !Oper.isDef()) continue;
    if (Oper.getReg() == Reg)
      return true;
  }
  return false;
}

static bool areUseOperandsAvailable(MachineInstr *Inst,
                                    DenseSet<unsigned> &AvailRegs) {
  for (unsigned I = 0, E = Inst->getNumOperands(); I != E; ++I) {
    MachineOperand &Oper = Inst->getOperand(I);
    if (!Oper.isReg() || !Oper.getReg() || !Oper.isUse()) continue;
    if (!AvailRegs.count(Oper.getReg()))
      return false;
  }
  return true;
}

static bool isLEA(MachineInstr *Inst) {
  std::string OpName(Inst->getDesc().getName());
  if (OpName.find_first_of("LEA") == 0)
    return true;
  return false;
}

static bool isMoveImmediate(MachineInstr *Inst) {
  return Inst->getDesc().isMoveImmediate();
}

bool Rematerialization::isCheaperThanLoadAndSafe(MachineInstr *Inst) {
  if (TII->isTriviallyReMaterializable(Inst))
    return true;
  // Check safety.
  // TODO: Relax, loads from constant pool can be rematerialized.
  if (Inst->getDesc().mayLoad() || Inst->getDesc().mayStore() ||
      Inst->hasUnmodeledSideEffects())
    return false;

  if (Inst->isCopy()) return true;
  if (isLEA(Inst)) return true;
  if (isMoveImmediate(Inst)) return true;
  // There must be more ....

  return false;
}

void Rematerialization::findOpportunities(MachineBasicBlock *BB,
                                          DenseSet<unsigned> &ColoredRegs) {
  for (AvailDefsIterator AI = AvailDef->begin(BB), AE = AvailDef->end(BB);
       AI != AE; ++AI) {
    const std::set<MachineInstr *> &AvailDefs = *AI;
    MachineInstr *Inst = AI.getCurrentInstruction();
    if (AvailDefs.empty()) continue;

    // Check whether any of the current instruction's operands are in the
    // available set of the current instruction.
    for (unsigned I = 0, E = Inst->getNumOperands(); I != E; ++I) {
      MachineOperand &Oper = Inst->getOperand(I);
      if (!Oper.isReg() || !Oper.getReg() || !Oper.isUse()) continue;
      unsigned Reg = Oper.getReg();

      for (std::set<MachineInstr *>::const_iterator SI = AvailDefs.begin(),
           SE = AvailDefs.end(); SI != SE; ++SI) {
        MachineInstr *DefiningInst = *SI;
        // Check whether the current avail definition defines the operand and
        // whether its operands are available (guaranteed to get a color).
        if (doesDefineOperand(DefiningInst, Reg) &&
            areUseOperandsAvailable(DefiningInst, ColoredRegs)) {
          if (isCheaperThanLoadAndSafe(DefiningInst))
            RematCandidates[&Oper] = DefiningInst;
          break;
        }
      }
    }
  }
}

void Rematerialization::computeAvailDefs() {
  assert(AvailDef == NULL);
  // Build avail defs.
  AvailDef = new AvailableDefinitions(MF);
}
