/*
 * This was taken from the MSCP ILOC compiler.
 * 
 * VectorSet.c
 * 
 * Implementation of vector sets.
 * 
 * by Robert Shillingsburg, June, 1992
 * 
 * 7/9/92 Rob Shillingsburg: Replaced VectprSetPts with VectorSets.
 * Improved the set printer for pretty output.  Improved the
 * VectorSet_ForAll iterator macro.  Got rid of NextElt() by
 * incorporating it into ChooseMember.  Various other beautifying and
 * efficiency changes.
 * 
 * 7/16/92 Rob Shillingsburg: Changed VectorSet_Print to VectorSet_Dump,
 * removed the File parameter, and changed VectorSet_Dump and PrintRange
 * to dump to stderr.
 *
 * 2/23/93 Preston Briggs: Made union, intersect, difference, and complement
 * into 3-address operations.  Added difference_union.
 *
 * 3/1/93 Preston Briggs: Added imply operation.
 *
 */
/*===----------------------------------------------------------------------===//
/*This software was produced with support from the Defense Advanced
/*Research Projects Agency (DARPA) through AFRL Contract FA8650-09-C-1915. 
/*Nothing in this work should be construed as reflecting the official policy or
/*position of the Defense Department, the United States government, or
/*Rice University.
/*
/*THIS SOFTWARE HAS BEEN APPROVED FOR PUBLIC RELEASE, UNLIMITED DISTRIBUTION.
/*
/*===----------------------------------------------------------------------===*/


/* 
 * A VectorSet is a set implemented as a bit vector.  Most of the details
 * should be obvious, but I'll give some of them anyway.  The bit vector
 * structure contains a couple of housekeeping fields and a one-element
 * array called word.  Word is really just the beginning of the
 * arbitrary-length bit vector; when a new bit vector set is created,
 * the array is sized large enough to hold the whole universe.  The
 * bookkeeping field universe_size is the size of the universe
 * [0..universe_size-1]; word_count is the total number of words in the
 * word array.
 * 
 * The word array contains enough 32-bit words to contain the universe,
 * and it is rounded up to the next four-byte boundary.  Excess bits in
 * the final word must always be cleared to zero.  This invariant is
 * needed to avoid errors when counting, say, the total number of '1'
 * bits.  The only interface function which has to deal with clearing
 * these bits is VectorSet_Complement, which inverts the entire array,
 * then clears the extra bits in the final word.  By the way,
 * VectorSet_Clear zeroes out the entire array, including any excess
 * trailing bits.
 * 
 * Universe elements are laid out in "natural" order: word[0] contains
 * the first 32 elements [0..31], word[1] has the next 32, and so on.
 * Within a word, the lsb has the lowest-numbered element.  For example,
 * for a universe_size of 32, the word 0x00000001 corresponds to a set
 * containing only element 0.
 * 
 * 
 * 
 */

#include "Shared.h"
#include <memory.h>



/*** Constants: *************************************************************/

/* 32 bits in a word. */
static const Unsigned_Int BITS_PER_WORD = 32;

/* Shift 5 times to divide by 32. */
static const Unsigned_Int LOGBITS_PER_WORD = 5;

/* 4 bytes in a word. */
static const Unsigned_Int BYTES_PER_WORD = 4;



/*** File-scope variables: ***************************************************/


/* BitsInByte is a lookup table which tells the number of bits that
   are in the looked-up number.  It is very useful in VectorSet_Size. */

static Unsigned_Int1 bitsInByte[256] = {
  0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8
};


/*** Internal function declarations: *****************************************/
#ifdef RUNTIME_CHECK
static Void CheckUniverses(VectorSet destSet, VectorSet srcSet);
#endif
static Void PrintRange(Int rangeBottom, Int rangeTop);



/*** Interface function definitions: *****************************************/



/* VectorSet_Create: Creates a new set, returning a pointer to it. */
VectorSet VectorSet_Create(Unsigned_Int universe_size)
{
  VectorSet tmpSet;    /* The pointer we will return to the user. */
  Int word_count;       /* Number of 32-bit words needed for the set. */
  Int size;            /* Total size of the set's structure. */


#ifdef RUNTIME_CHECK
  if (universe_size == 0) {
    /* Error!  Print message and abort. */

    fprintf(stderr, "VectorSet_Create: Zero passed as universe_size.\n");
    abort();
  }
#endif /* RUNTIME_CHECK */


  /* Find number of words needed for universe_size bits. */
  word_count = ((universe_size - 1) >> LOGBITS_PER_WORD) + 1;
  
  /* Calculate size of the structure: structure header minus first word
     plus the size of the bit vector. */
  size = sizeof(struct vector_set) - sizeof(Unsigned_Int4) +
    (sizeof(Unsigned_Int4) * word_count);
  
  /* Allocate memory for the set. */
  /*tmpSet = Arena_GetMem(arena, size);*/
  tmpSet = (VectorSet)malloc(size);
  
  /* Fill in universe size and number of words. */
  tmpSet->universe_size = universe_size;
  tmpSet->word_count = word_count;

  /* Empty out the set. */
  VectorSet_Clear(tmpSet);

  /* Return set to user. */
  return tmpSet;

} /* VectorSet_Create. */






/* VectorSet_Clear: Removes all members from a set. */
Void VectorSet_Clear(VectorSet set)
{

  /* Clear out the bit vector. */
  memset(set->word, 0, set->word_count * BYTES_PER_WORD);
  
} /* VectorSet_Clear. */





/* VectorSet_Insert: Inserts a member into a set. */
Void VectorSet_Insert(VectorSet set, Unsigned_Int elt)
{

#ifdef RUNTIME_CHECK
  if (elt >= set->universe_size) {
    /* Element is not in the universe! */
    
    /* Print error message and abort. */
    fprintf(stderr, "VectorSet_Insert: Element(%d) not in universe(%d).\n",
	    elt, set->universe_size);
    abort();
  }
#endif /* RUNTIME_CHECK */

  /* Set the appropriate bit using fancy bitwise stuff. */
  set->word[elt >> LOGBITS_PER_WORD] |= 1 << (elt & (BITS_PER_WORD - 1));

} /* VectorSet_Insert. */






/* VectorSet_Delete: Removes a member from a set. */
Void VectorSet_Delete(VectorSet set, Unsigned_Int elt)
{

#ifdef RUNTIME_CHECK
  if (elt >= set->universe_size) {
    /* Element is not in the universe! */
    
    /* Print error message and abort. */
    fprintf(stderr, "VectorSet_Delete: Element(%d) not in universe(%d).\n",
	    elt, set->universe_size);
    abort();
  }
#endif /* RUNTIME_CHECK */

  /* Clear the appropriate bit using fancy bitwise stuff. */
  set->word[elt >> LOGBITS_PER_WORD] &= ~(1 << (elt & (BITS_PER_WORD - 1)));  

} /* VectorSet_Delete. */







/* VectorSet_Member: Returns TRUE if elt is a member of set. */
Boolean VectorSet_Member(VectorSet set, Unsigned_Int elt)
{

#ifdef RUNTIME_CHECK
  if (elt >= set->universe_size) {
    /* Element is not in the universe! */
    
    /* Print error message and abort. */
    fprintf(stderr, "VectorSet_Member: Element(%d) not in universe(%d).\n",
	    elt, set->universe_size);
    abort();
  }
#endif /* RUNTIME_CHECK */

  /* Return the appropriate bit using fancy bitwise stuff. */
  return set->word[elt >> LOGBITS_PER_WORD] >> (elt & (BITS_PER_WORD - 1)) & 1;

} /* VectorSet_Member. */





/* VectorSet_Size: Returns the number of members in set. */
Unsigned_Int VectorSet_Size(VectorSet set)
{
  Unsigned_Int4 i;              /* Iterator. */
  Unsigned_Int1 *currByte;      /* Current byte in the array. */
  Unsigned_Int4 size = 0;       /* Cumulative size so far. */

  /* Start currByte at the beginning of the bit vector. */
  currByte = (Unsigned_Int1 *) set->word;

  /* Count bits in all words. */
  for (i = 0; i < set->word_count * BYTES_PER_WORD; i++) {
    /* Still more bytes to process. */

    /* Add bits in current byte to size. */
    size += bitsInByte[*currByte];

    /* Move to next byte. */
    currByte++;
  }

  /* Return total size to user. */
  return size;

} /* VectorSet_Size. */






/* VectorSet_Copy: Performs destSet <= srcSet. DestSet and srcSet must
   be of the same universe size. */
Void VectorSet_Copy(VectorSet destSet, VectorSet srcSet)
{
  
#ifdef RUNTIME_CHECK
  /* Make sure the sets' universes are the same size. */
  CheckUniverses(destSet, srcSet);
#endif /* RUNTIME_CHECK */  

  /* Copy the bit vector. */
  memcpy(destSet->word, srcSet->word, srcSet->word_count * BYTES_PER_WORD);
  
} /* VectorSet_Copy. */






/* VectorSet_Union: Performs x = y | z
 * x, y, and z are expected to be of the same size.
 * The operands may be aliased.
 */

Void VectorSet_Union(VectorSet x, VectorSet y, VectorSet z)
{
  Unsigned_Int4 *xLast;
  Unsigned_Int4 *xp;        /* Current location in x */
  Unsigned_Int4 *yp;        /* Current location in y */
  Unsigned_Int4 *zp;        /* Current location in z */

#ifdef RUNTIME_CHECK
  /* Make sure the sets' universes are the same size. */
  CheckUniverses(x, y);
  CheckUniverses(x, z);
#endif /* RUNTIME_CHECK */  

  /* Initialize the pointers. */
  xp = x->word;
  yp = y->word;
  zp = z->word;
  xLast = x->word + x->word_count;

  /* Union the bit vectors. */
  while (xp < xLast)
    *xp++ = *yp++ | *zp++;

} /* VectorSet_Union. */



/* VectorSet_Intersect: Performs x = y & z
 * x, y, and z are expected to be of the same size.
 * The operands may be aliased.
 */

Void VectorSet_Intersect(VectorSet x, VectorSet y, VectorSet z)
{
  Unsigned_Int4 *xLast;
  Unsigned_Int4 *xp;        /* Current location in x */
  Unsigned_Int4 *yp;        /* Current location in y */
  Unsigned_Int4 *zp;        /* Current location in z */

#ifdef RUNTIME_CHECK
  /* Make sure the sets' universes are the same size. */
  CheckUniverses(x, y);
  CheckUniverses(x, z);
#endif /* RUNTIME_CHECK */  

  /* Initialize the pointers. */
  xp = x->word;
  yp = y->word;
  zp = z->word;
  xLast = x->word + x->word_count;

  /* Union the bit vectors. */
  while (xp < xLast)
    *xp++ = *yp++ & *zp++;

} /* VectorSet_Intersect. */



/* VectorSet_Difference: Performs x = y - z
 * x, y, and z are expected to be of the same size.
 * The operands may be aliased.
 */

Void VectorSet_Difference(VectorSet x, VectorSet y, VectorSet z)
{
  Unsigned_Int4 *xLast;
  Unsigned_Int4 *xp;        /* Current location in x */
  Unsigned_Int4 *yp;        /* Current location in y */
  Unsigned_Int4 *zp;        /* Current location in z */

#ifdef RUNTIME_CHECK
  /* Make sure the sets' universes are the same size. */
  CheckUniverses(x, y);
  CheckUniverses(x, z);
#endif /* RUNTIME_CHECK */  

  /* Initialize the pointers. */
  xp = x->word;
  yp = y->word;
  zp = z->word;
  xLast = x->word + x->word_count;

  /* Union the bit vectors. */
  while (xp < xLast)
    *xp++ = *yp++ & ~(*zp++);

} /* VectorSet_Difference */




/* VectorSet_Imply: Performs x = not(y) | z
 * x, y, and z are expected to be of the same size.
 * The operands may be aliased.
 */

Void VectorSet_Imply(VectorSet x, VectorSet y, VectorSet z)
{
  Unsigned_Int4 *xLast;
  Unsigned_Int4 *xp;        /* Current location in x */
  Unsigned_Int4 *yp;        /* Current location in y */
  Unsigned_Int4 *zp;        /* Current location in z */

#ifdef RUNTIME_CHECK
  /* Make sure the sets' universes are the same size. */
  CheckUniverses(x, y);
  CheckUniverses(x, z);
#endif /* RUNTIME_CHECK */  

  /* Initialize the pointers. */
  xp = x->word;
  yp = y->word;
  zp = z->word;
  xLast = x->word + x->word_count;

  /* Union the bit vectors. */
  while (xp < xLast)
    *xp++ = ~(*yp++) | *zp++;

} /* VectorSet_Imply */




/* VectorSet_DifferenceUnion: Performs w = (x - y) | z
 * w, x, y, and z are expected to be of the same size.
 * The operands may be aliased.
 */

Void VectorSet_DifferenceUnion(VectorSet w, VectorSet x, 
			       VectorSet y, VectorSet z)
{
  Unsigned_Int4 *wLast;
  Unsigned_Int4 *wp;        /* Current location in w */
  Unsigned_Int4 *xp;        /* Current location in x */
  Unsigned_Int4 *yp;        /* Current location in y */
  Unsigned_Int4 *zp;        /* Current location in z */

#ifdef RUNTIME_CHECK
  /* Make sure the sets' universes are the same size. */
  CheckUniverses(w, x);
  CheckUniverses(w, y);
  CheckUniverses(w, z);
#endif /* RUNTIME_CHECK */  

  /* Initialize the pointers. */
  wp = w->word;
  xp = x->word;
  yp = y->word;
  zp = z->word;
  wLast = w->word + w->word_count;

  /* Union the bit vectors. */
  while (wp < wLast)
    *wp++ = (*xp++ & ~(*yp++)) | *zp++;

} /* VectorSet_DifferenceUnion */






/* VectorSet_Complement: x = not y */

Void VectorSet_Complement(VectorSet x, VectorSet y)
{
  Unsigned_Int4 final_bits;
  Unsigned_Int4 *xLast;
  Unsigned_Int4 *xp;        /* Current location in x */
  Unsigned_Int4 *yp;        /* Current location in y */

#ifdef RUNTIME_CHECK
  /* Make sure the sets' universes are the same size. */
  CheckUniverses(x, y);
#endif /* RUNTIME_CHECK */  

  /* Initialize the pointers. */
  xp = x->word;
  yp = y->word;
  xLast = x->word + x->word_count;

  /* Union the bit vectors. */
  while (xp < xLast)
    *xp++ = ~(*yp++);

  /* chop off any trailing bits not belonging to set */

  final_bits = x->universe_size & (BITS_PER_WORD - 1);
  xp--;
  *xp &= ((Unsigned_Int4)~0 >> (BITS_PER_WORD - final_bits));

} /* VectorSet_Complement. */





/* VectorSet_Equal: Returns TRUE if set1 = set2. Set1 and
   Set2 must be of the same universe size. */
Boolean VectorSet_Equal(VectorSet set1, VectorSet set2)
{
  
#ifdef RUNTIME_CHECK
  /* Make sure the sets' universes are the same size. */
  CheckUniverses(set1, set2);
#endif /* RUNTIME_CHECK */
  
  /* Compare the sets and return the results. */
  return !memcmp(set1->word, set2->word, set1->word_count * BYTES_PER_WORD);

} /* VectorSet_Equal. */







/* VectorSet_ChooseMember: Returns a random member of a nonempty set. */
Unsigned_Int VectorSet_ChooseMember(VectorSet set)
{
  Unsigned_Int wordNum;    /* The current word number in the bit vector. */
  Unsigned_Int bitNum;     /* The current bit number. */
  Unsigned_Int4 *wordPtr;  /* Pointer to the current word. */
  Unsigned_Int4 bitMask;   /* Bit-finder mask. */


  /* Start at zeroth word. */
  wordNum = 0;

  /* Initialize wordPtr to be the zeroth word. */
  wordPtr = &(set->word[wordNum]);

  while (wordNum < set->word_count) {
    /* Need to search this word. */
    
    if (*wordPtr) {
      /* There are one or more '1' bits in this word; we'll search it. */
      
      /* Start bitMask at 0x00000001. */
      bitMask = 1;
      
      /* Start bitNum at the zeroth bit. */
      bitNum = 0;
      
      /* Search in the word. */
      while (bitMask != 0x00000000) {
	/* Haven't finished this word yet. */
	
	if (*wordPtr & bitMask) {
	  /* Found a '1' bit. */
	  
	  /* Return the overall bit position of the bit. */
	  return (wordNum * BITS_PER_WORD) + bitNum;
	  
	} else {
	  /* This bit was a zero. */
	  
	  /* Scoot the bitMask to the left one place. */
	  bitMask = bitMask << 1;
	  
	  /* Increment bitNum so we can quickly know which bit we're on. */
	  bitNum++;
	}
      }
    }

    /* Move to next word. */
    wordPtr++;
    wordNum++;

  }

  /* If we made it here, then the set is empty. */
  
  /* Print error message and abort. */
  fprintf(stderr, "VectorSet_ChooseMember: Set is empty.\n");
  abort();
}





/* VectorSet_Sort: Sorts the members of set. */
Void VectorSet_Sort(VectorSet set)
{
  /* Ha!  They're already sorted! */
}
     





/* VectorSet_Dump: Print the contents of a set to stderr. */
Void VectorSet_Dump(VectorSet set)
{
  Unsigned_Int4 elt;    /* An element of the set. */
  Int rangeBottom;      /* Bottom of a dense range in the set. */
  Int rangeTop;         /* Top of a dense range in the set. */
  
  if (!set) {
    /* VectorSet is NULL. */
    fprintf(stderr, "VectorSet_Dump: NULL set.\n");
    return;
  }

  /* Give rangeBottom a meaningless value. */
  rangeBottom = -1;

  /* Start the job with an open brace. */
  fprintf(stderr, "{ ");
  
  VectorSet_ForAll(elt, set) {
    /* For each element in the set in order: */
    
    if (rangeBottom < 0) {
      /* No current range. */
      
      /* Start a new range. */
      rangeBottom = elt;
      rangeTop = elt;
      
    } else if (static_cast<Unsigned_Int4>(rangeTop + 1) != elt) {
      /* This element is starts a new range. */
      
      /* Print previous range. */
      PrintRange(rangeBottom, rangeTop);
      
      /* Start new range. */
      rangeBottom = elt;
      rangeTop = elt;

    } else {
      /* Add this element to the range. */
      rangeTop = elt;
    }
  }

  if (rangeBottom >= 0) {
    /* One final range to print. */

    /* Print it. */
    PrintRange(rangeBottom, rangeTop);
  }

  /* End set and flush out the line buffer. */
  fprintf(stderr, "}\n");
 
}



/*** Internal function definitions: ******************************************/

#ifdef RUNTIME_CHECK
/* CheckUniverses: If destSet and srcSet don't have the same universe size,
   print and error message and abort. */
static Void CheckUniverses(VectorSet destSet, VectorSet srcSet)
{
  if (destSet->universe_size != srcSet->universe_size) {
    /* Universes are not equivalent. */
    
    /* Print error message and abort. */
    fprintf(stderr, "VectorSet: Universes are not equivalent.\n");
    abort();
  }
} /* CheckUniverses. */
#endif


static Void PrintRange(Int rangeBottom, Int rangeTop)
{
  if (rangeTop != rangeBottom) {
    /* Range has more than one element. */
    
    /* Print this range. */
    fprintf(stderr, "%d-%d ", rangeBottom, rangeTop);
    
  } else {
    
    /* Range has only one element. */
    
    /* Print this range. */
    fprintf(stderr, "%d ", rangeBottom);
  }
} /* PrintRange. */
